/*
Compilation of OpenType hints
Copyright (C) 2005, 2013, 2020 Richard Wordingham

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "compile_instr.h"
#include "tokeniser.h"

static struct instr_type defs[] = {
  {"SVTCA",     AXIS,     0x00,  0, 0, 0},
  {"SPVTCA",    AXIS,     0x02,  0, 0, 0},
  {"SFVTCA",    AXIS,     0x04,  0, 0, 0},  
  {"SPVTL",     ROT,      0x06,  2, 0, 0},  
  {"SFVTL",     ROT,      0x08,  2, 0, 0},  
  {"SPVFS",     NO_FLAGS, 0x0a,  2, 0, 0},
  {"SFVFS",     NO_FLAGS, 0x0b,  2, 0, 0},        
  {"GPV",       NO_FLAGS, 0x0c, -2, 0, 0},       
  {"GFV",       NO_FLAGS, 0x0d, -2, 0, 0},        
  {"SFVTPV",    NO_FLAGS, 0x0e,  0, 0, 0},        
  {"ISECT",     NO_FLAGS, 0x0f,  5, 0, 0},        
  {"SRP0",      NO_FLAGS, 0x10,  1, 0, 0},      
  {"SRP1",      NO_FLAGS, 0x11,  1, 0, 0},        
  {"SRP2",      NO_FLAGS, 0x12,  1, 0, 0},        
  {"SZP0",      NO_FLAGS, 0x13,  1, 0, 0}, /* Pops 0 or 1 */        
  {"SZP1",      NO_FLAGS, 0x14,  1, 0, 0},        
  {"SZP2",      NO_FLAGS, 0x15,  1, 0, 0},        
  {"SZPS",      NO_FLAGS, 0x16,  1, 0, 0},        
  {"SLOOP",     NO_FLAGS, 0x17,  1, 0, 0},
  {"RTG",       NO_FLAGS, 0x18,  0, 0, 0},        
  {"RTHG",      NO_FLAGS, 0x19,  0, 0, 0},        
  {"SMD",       NO_FLAGS, 0x1a,  1, 0, 0},
  {"ELSE",      NO_FLAGS, 0x1b,  0, 0, 0},
  {"JMPR",      NO_FLAGS, 0x1c,  1, 0, 0},
  {"SCVTCI",    NO_FLAGS, 0x1d,  1, 0, 0},        
  {"SSWCI",     NO_FLAGS, 0x1e,  1, 0, 0},
  {"SSW",       NO_FLAGS, 0x1f,  1, 0, 0},
  {"DUP",       NO_FLAGS, 0x20, -1, 0, 0}, /* Pops and pushes */
  {"POP",       NO_FLAGS, 0x21,  1, 0, 0},
  {"CLEAR",     NO_FLAGS, 0x22,  0, 0, 0}, /* Need special stack handling! */
  {"SWAP",      NO_FLAGS, 0x23,  0, 0, 0}, /* Pops and pushes */
  {"DEPTH",     NO_FLAGS, 0x24, -1, 0, 0},
  {"CINDEX",    NO_FLAGS, 0x25, -1, 0, 0},
  {"MINDEX",    NO_FLAGS, 0x26,  0, 0, 0},
  {"ALIGNPTS",  NO_FLAGS, 0x27,  2, 0, 0},
  {"UTP",       NO_FLAGS, 0x29,  1, 0, 0},
  {"LOOPCALL",  NO_FLAGS, 0x2a,  2, 0, 0}, /* Called function may affect stack */
  {"CALL",      NO_FLAGS, 0x2b,  1, 0, 0}, /* Called function may affect stack */
  {"FDEF",      NO_FLAGS, 0x2c,  1, 0, 0},
  {"ENDF",      NO_FLAGS, 0x2d,  0, 0, 0},
  {"MDAP",      ROUND,    0x2e,  1, 0, 0},
  {"IUP",       AXIS,     0x30,  0, 0, 0},
  {"SHP",       RPSEL,    0x32,  0, 1, 0}, /* Affected by loop */
  {"SHC",       RPSEL,    0x34,  1, 0, 0},  
  {"SHZ",       RPSEL,    0x36,  1, 0, 0},  
  {"SHPIX",     NO_FLAGS, 0x38,  1, 1, 0}, /* Affected by loop */        
  {"IP",        NO_FLAGS, 0x39,  0, 1, 0}, /* Affected by loop */        
  {"MSIRP",     SETRP0,   0x3a,  2, 0, 0},  
  {"ALIGNRP",   NO_FLAGS, 0x3c,  1, 0, 0},        
  {"RTDG",      NO_FLAGS, 0x3d,  0, 0, 0},
  {"MIAP",      ROUND,    0x3e,  2, 0, 0},
  {"NPUSHB",    LITERAL,  0x40,  0, 0, 0},
  {"NPUSHW",    LITERAL,  0x41,  0, 0, 0},
  {"WS",        NO_FLAGS, 0x42,  1, 0, 0}, /* Pops and pushes */
  {"RS",        NO_FLAGS, 0x43,  0, 0, 0}, /* Pops and pushes */
  {"WCVTP",     NO_FLAGS, 0x44,  2, 0, 0},
  {"RCVT",      NO_FLAGS, 0x45,  0, 0, 0}, /* Pops and pushes */
  {"GC",        WHEN,     0x46,  0, 0, 0}, /* Pops and pushes */
  {"SCFS",      NO_FLAGS, 0x48,  2, 0, 0},        
  {"MD",        WHEN,     0x49,  1, 0, 0},  /* Pops and pushes */
  {"MPPEM",     NO_FLAGS, 0x4b, -1, 0, 0},
  {"MPS",       NO_FLAGS, 0x4c, -1, 0, 0},        
  {"FLIPON",    NO_FLAGS, 0x4d,  0, 0, 0},
  {"FLIPOFF",   NO_FLAGS, 0x4e,  0, 0, 0},
  {"DEBUG",     NO_FLAGS, 0x4f,  1, 0, 0},
  {"LT",        NO_FLAGS, 0x50,  1, 0, 0}, /* Pop and push */
  {"LTEQ",      NO_FLAGS, 0x51,  1, 0, 0},
  {"GT",        NO_FLAGS, 0x52,  1, 0, 0},
  {"GTEQ",      NO_FLAGS, 0x53,  1, 0, 0},
  {"EQ",        NO_FLAGS, 0x54,  1, 0, 0},
  {"NEQ",       NO_FLAGS, 0x55,  1, 0, 0},
  {"ODD",       NO_FLAGS, 0x56,  0, 0, 0}, /* Pops and pushes */
  {"EVEN",      NO_FLAGS, 0x57,  0, 0, 0}, /* Pops and pushes */
  {"IF",        NO_FLAGS, 0x58,  1, 0, 0},        
  {"EIF",       NO_FLAGS, 0x59,  0, 0, 0},
  {"AND",       NO_FLAGS, 0x5a,  1, 0, 0}, /* These three pop and push */
  {"OR",        NO_FLAGS, 0x5b,  1, 0, 0},
  {"NOT",       NO_FLAGS, 0x5c,  0, 0, 0},
  {"DELTAP1",   NO_FLAGS, 0x5d,  1, 0, 2},
  {"SDB",       NO_FLAGS, 0x5e,  1, 0, 0},
  {"SDS",       NO_FLAGS, 0x5f,  1, 0, 0},        
  {"ADD",       NO_FLAGS, 0x60,  1, 0, 0}, /* These pop and push */
  {"SUB",       NO_FLAGS, 0x61,  1, 0, 0},
  {"DIV",       NO_FLAGS, 0x62,  1, 0, 0},
  {"MUL",       NO_FLAGS, 0x63,  1, 0, 0},
  {"ABS",       NO_FLAGS, 0x64,  0, 0, 0},
  {"NEG",       NO_FLAGS, 0x65,  0, 0, 0},
  {"FLOOR",     NO_FLAGS, 0x66,  0, 0, 0},
  {"CEILING",   NO_FLAGS, 0x67,  0, 0, 0},
  {"ROUND",     COLOUR,   0x68,  0, 0, 0},  
  {"NROUND",    COLOUR,   0x6c,  0, 0, 0}, /* End of 'these' */
  {"WCVTF",     NO_FLAGS, 0x70,  2, 0, 0},
  {"DELTAP2",   NO_FLAGS, 0x71,  1, 0, 2},
  {"DELTAP3",   NO_FLAGS, 0x72,  1, 0, 2},
  {"DELTAC1",   NO_FLAGS, 0x73,  1, 0, 2},
  {"DELTAC2",   NO_FLAGS, 0x74,  1, 0, 2},
  {"DELTAC3",   NO_FLAGS, 0x75,  1, 0, 2},
  {"SROUND",    NO_FLAGS, 0x76,  1, 0, 0},         
  {"SROUND45",  NO_FLAGS, 0x77,  1, 0, 0},
  {"JROT",      NO_FLAGS, 0x78,  2, 0, 0},        
  {"JROF",      NO_FLAGS, 0x79,  2, 0, 0},
  {"ROFF",      NO_FLAGS, 0x7a,  0, 0, 0},        
  {"RUTG",      NO_FLAGS, 0x7c,  0, 0, 0},        
  {"RDTG",      NO_FLAGS, 0x7d,  0, 0, 0},        
  {"SANGW",     NO_FLAGS, 0x7e,  1, 0, 0},
  {"FLIPPT",    NO_FLAGS, 0x80,  0, 1, 0}, /* Performed loop times */        
  {"FLIPRGON",  NO_FLAGS, 0x81,  2, 0, 0},
  {"FLIPRGOFF", NO_FLAGS, 0x82,  2, 0, 0},        
  {"SCANCTRL",  NO_FLAGS, 0x85,  1, 0, 0},        
  {"SDPVTL",    ROT,      0x86,  2, 0, 0},  
  {"GETINFO",   NO_FLAGS, 0x88,  0, 0, 0}, /* Pops and pushes */
  {"IDEF",      NO_FLAGS, 0x89,  1, 0, 0},
  {"ROLL",      NO_FLAGS, 0x8a,  0, 0, 0}, /* Pops and pushes */
  {"MAX",       NO_FLAGS, 0x8b,  1, 0, 0}, /* Pops and pushes */
  {"MIN",       NO_FLAGS, 0x8c,  1, 0, 0}, /* Pops and pushes */
  {"SCANTYPE",  NO_FLAGS, 0x8d,  1, 0, 0},        
  {"INSTCTRL",  NO_FLAGS, 0x8e,  2, 0, 0},        
  {"PUSHB",     OCOUNT,   0xb0,  0, 0, 0},
  {"PUSHW",     OCOUNT,   0xb8,  0, 0, 0},
  {"MDRP",      MOVE,     0xc0,  1, 0, 0},
  {"MIRP",      MOVE,     0xe0,  2, 0, 0}
};

void define_opcodes(struct map *the_list) {
  int i_code, n_codes;

  create_map(the_list);
  n_codes = sizeof(defs) / sizeof(defs[0]);
  for (i_code = 0; i_code < n_codes; i_code++) {
    insert_in_map(the_list, defs[i_code].name, (void *) (defs + i_code));
  }
}

extern struct map opcode_def;
extern struct map program_def;

void compile_instr(unsigned char **code, int *length) {
  struct map local_names;
  struct instr_type *opdesc;
  int capacity = 0;
  int flags, as_bytes, n_items, the_item;

  *length = 0;
  *code = 0;
  create_map(&local_names);
  while (END_OF_FILE != current.form) {
    if (AN_IDENT == current.form) {
      if (lookup_in_map(opcode_def, current.sval, (void**) &opdesc)) {
	if ((*length+2) >= capacity) {
	  *code = realloc(*code, (capacity += 256) * sizeof(int));
	  if (!*code) {
	    fprintf(stderr, "Memory exhausted in compile_instr().\n");
	    exit(1);
	  }
	}
#if 0
	fprintf(stderr, "%s flags %d.\n", opdesc->name, opdesc->flags);
#endif
	switch (opdesc->flags) {
        case AXIS:
        case ROT:
        case SETRP0:
        case ROUND:
        case RPSEL:
        case WHEN:
	  advance_token();
	  if (PUNCTUATION != current.form || '[' != current.ival) {
	    show_source();
	    fprintf(stderr, "Expected '['.\n");
	    break;
	  }
	  advance_token();
	  if (AN_INT != current.form
	      || (0 != current.ival && 1 != current.ival)) {
	    show_source();
	    fprintf(stderr, "Expected flag 0 or 1.\n");
	    flags = 0;
	  } else {
	    flags = current.ival;
	  }
	  (*code)[(*length)++] = opdesc->code + flags;
	  advance_token();
	  if (PUNCTUATION != current.form || ']' != current.ival) {
	    show_source();
	    fprintf(stderr, "Expected ']'.\n");
	  }
	  break;
	case NO_FLAGS:
	  (*code)[(*length)++] = opdesc->code;
	  break;
        case COLOUR:
	  advance_token();
	  if (PUNCTUATION != current.form || '[' != current.ival) {
	    show_source();
	    fprintf(stderr, "Expected '['.\n");
	    break;
	  }
	  advance_token();
	  if (AN_INT != current.form
	      || 0 > current.ival || 3 < current.ival) {
	    show_source();
	    fprintf(stderr, "Expected flag 0 or 1.\n");
	    flags = 0;
	  } else {
	    flags = current.ival;
	  }
	  (*code)[(*length)++] = opdesc->code + flags;
	  advance_token();
	  if (PUNCTUATION != current.form || ']' != current.ival) {
	    show_source();
	    fprintf(stderr, "Expected ']'.\n");
	  }
	  break;
        case MOVE:
	  advance_token();
	  if (PUNCTUATION != current.form || '[' != current.ival) {
	    show_source();
	    fprintf(stderr, "Expected '['.\n");
	    break;
	  }
	  advance_token();
	  if (AN_INT != current.form) {
	    show_source();
	    fprintf(stderr, "Expected a bit string.\n");
	    flags = 0;
	  } else {
	    /* Leading 0 will flag value as octal! */
	    switch (current.ival) {
	    case 0: /* 00000 */
	      flags = 0;
	      break;
	    case 1: /* 00001 */
	      flags = 1;
	      break;
	    case 010: /* 00010 */
	      flags = 2;
	      break;
	    case 011: /* 00011 */
	      flags = 3;
	      break;
	    case 0100: /* 00100 */
	      flags = 4;
	      break;
	    case 0101: /* 00101 */
	      flags = 5;
	      break;
	    case 0110: /* 00110 */
	      flags = 6;
	      break;
	    case 0111: /* 00111 */
	      flags = 7;
	      break;
	    case 01000: /* 01000 */
	      flags = 8;
	      break;
	    case 01001: /* 01001 */
	      flags = 9;
	      break;
	    case 01010: /* 00010 */
	      flags = 10;
	      break;
	    case 01011: /* 01011 */
	      flags = 11;
	      break;
	    case 01100: /* 01100 */
	      flags = 12;
	      break;
	    case 01101: /* 01101 */
	      flags = 13;
	      break;
	    case 01110: /* 01110 */
	      flags = 14;
	      break;
	    case 01111: /* 01111 */
	      flags = 15;
	      break;
	    case 10000: /* 10000 */
	      flags = 16;
	      break;
	    case 10001: /* 10001 */
	      flags = 17;
	      break;
	    case 10010: /* 10010 */
	      flags = 18;
	      break;
	    case 10011: /* 10011 */
	      flags = 19;
	      break;
	    case 10100: /* 10100 */
	      flags = 20;
	      break;
	    case 10101: /* 10101 */
	      flags = 21;
	      break;
	    case 10110: /* 10110 */
	      flags = 22;
	      break;
	    case 10111: /* 10111 */
	      flags = 23;
	      break;
	    case 11000: /* 11000 */
	      flags = 24;
	      break;
	    case 11001: /* 11001 */
	      flags = 25;
	      break;
	    case 11010: /* 10010 */
	      flags = 26;
	      break;
	    case 11011: /* 11011 */
	      flags = 27;
	      break;
	    case 11100: /* 11100 */
	      flags = 28;
	      break;
	    case 11101: /* 11101 */
	      flags = 29;
	      break;
	    case 11110: /* 11110 */
	      flags = 30;
	      break;
	    case 11111: /* 11111 */
	      flags = 31;
	      break;
/* Cases with leading 0 omitted */
	    case 10: /* 00010 */
	      flags = 2;
	      break;
	    case 11: /* 00011 */
	      flags = 3;
	      break;
	    case 100: /* 00100 */
	      flags = 4;
	      break;
	    case 101: /* 00101 */
	      flags = 5;
	      break;
	    case 110: /* 00110 */
	      flags = 6;
	      break;
	    case 111: /* 00111 */
	      flags = 7;
	      break;
	    case 1000: /* 01000 */
	      flags = 8;
	      break;
	    case 1001: /* 01001 */
	      flags = 9;
	      break;
	    case 1010: /* 00010 */
	      flags = 10;
	      break;
	    case 1011: /* 01011 */
	      flags = 11;
	      break;
	    case 1100: /* 01100 */
	      flags = 12;
	      break;
	    case 1101: /* 01101 */
	      flags = 13;
	      break;
	    case 1110: /* 01110 */
	      flags = 14;
	      break;
	    case 1111: /* 01111 */
	      flags = 15;
	      break;
	    default:
	      show_source();
	      fprintf(stderr, "Undefined flag setting.\n");
	      flags = 0;
	    }
	    if (3 == (flags & 3)) {
	      show_source();
	      fprintf(stderr, "Undefined colour!\n");
	    }
	  }
	  (*code)[(*length)++] = opdesc->code + flags;
	  advance_token();
	  if (PUNCTUATION != current.form || ']' != current.ival) {
	    show_source();
	    fprintf(stderr, "Expected ']'.\n");
	  }
	  break;
        case OCOUNT:
        case LITERAL:
	  if (OCOUNT == opdesc->flags) {
	    as_bytes = strcmp(current.sval, "PUSHW");
	    advance_token();
	    if (PUNCTUATION != current.form || '[' != current.ival) {
	      show_source();
	      fprintf(stderr, "Expected '['.\n");
	      break;
	    }
	    advance_token();
	    if (AN_INT != current.form) {
	      show_source();
	      fprintf(stderr, "Expected integer or bit flag.\n");
	      flags = -1;
	    } else {
	      switch (current.ival) {
	      case 0:
	      case 1:
	      case 2:
	      case 3:
	      case 4:
	      case 5:
	      case 6:
	      case 7:
		flags = current.ival;
		break;
	      case 10:
	      case 010:
		flags = 2;
		break;
	      case 11:
	      case 011:
		flags = 3;
		break;
	      case 100:
		flags = 4;
		break;
	      case 101:
		flags = 5;
		break;
	      case 110:
		flags = 6;
		break;
	      case 111:
		flags = 7;
		break;
	      default:
		show_source();
		fprintf(stderr, "Offset count must be between 0 and 7.  "
			"%d is not permitted.\n", current.ival);
		flags = 7;
	      }
	      advance_token();
	    }
	    n_items = flags + 1;
	    (*code)[(*length)++] = opdesc->code + flags;
	    if (PUNCTUATION != current.form || ']' != current.ival) {
	      show_source();
	      fprintf(stderr, "Expected ']'.\n");
	    } else {
	      advance_token();
	    }
	  } else {
	    as_bytes = strcmp(current.sval, "NPUSHW");
	    (*code)[(*length)++] = opdesc->code;
	    advance_token();
	    if (AN_INT != current.form) {
	      show_source();
	      fprintf(stderr, "Expected item count.\n");
	      (*code)[(*length)++] = 0;
	      break;
	    } else if (current.ival < 0) {
	      show_source();
	      fprintf(stderr, "Item count must be non-negative.\n");
	      (*code)[(*length)++] = 0;
	      break;
	    } else if (current.ival > 255) {
	      show_source();
	      fprintf(stderr, "Item count cannot exceed 255.\n");
	      (*code)[(*length)++] = n_items = 255;
	    } else {
	      (*code)[(*length)++] = n_items = current.ival;
	    }
	    advance_token();
	  }
	  if ((*length + 2 * n_items) <= capacity) {
	    *code = realloc(*code, (capacity += 512) * sizeof(int));
	    if (!*code) {
	      fprintf(stderr, "Memory exhausted in compile_instr().\n");
	      exit(1);
	    }
	  }
	  while (n_items > 0 && END_OF_FILE != current.form) {
	    if (AN_INT == current.form) {
	      the_item = current.ival;
	    } else if (AN_IDENT == current.form) {
	      if (!lookup_in_map(local_names,current.sval, (void**) &the_item)
		  &&!lookup_in_map(program_def, current.sval, (void**)&the_item)
		  ) {
		show_source();
                if (lookup_in_map(opcode_def, current.sval, (void**)&the_item)
		    || !strcmp("end", current.sval)){
		  fprintf(stderr, "List of values to push is %d short.\n",
			  n_items);
		  break;
		} else {
		  fprintf(stderr, "Identifier %s not recognised.\n",
			  current.sval);
		  the_item = 0;
		}
	      }
	    } else {
	      show_source();
	      fprintf(stderr, "Expected number or identifier.\n");
	      the_item = 0;
	    }
	    if (as_bytes) {
	      (*code)[(*length)++] = the_item;
	    } else {
	      (*code)[(*length)++] = (the_item & 0x0ff00) >> 8;
	      (*code)[(*length)++] = the_item & 0x0ff;
	    }
	    advance_token();
	    n_items--;
	  }
	  backstep_token();
	  break;
	default:
	  fprintf(stderr, "INTERNAL ERROR: %s has unrecognised argument "
		  "code %d.\n", opdesc->name, opdesc->flags);
	}
	advance_token();
      } else if (!strcmp("EQUATE", current.sval)) {
	/* Return to later */
	show_source();
	fprintf(stderr, "EQUATE is not yet supported!\n");
	advance_token;
      } else if (!strcmp("end", current.sval)) {
	advance_token();
	return;
      } else {
	show_source();
	fprintf(stderr, "Expected an opcode.\n");
	got_what(stderr, current);
	advance_token();
      }
    } else {
      show_source();
      fprintf(stderr, "Expected an identifier, e.g. an opcode.\n");
      got_what(stderr, current);
      advance_token();
    }
  }
  delete_map(&local_names, 0);
}

