/*
Date conversion routines
Copyright (C) 2006, 2020 Richard Wordingham

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.
*/
#include <limits.h>
#include <stdio.h>
#include "dates1904.h"

static const int days_in_month[] = {31, 28, 31,
                                    30, 31, 30,
                                    31, 31, 30,
                                    31, 30, 31};
static int days_before[12];
static int INIT = 0;

static void init_dates() {
  int m;
  days_before[0] = 0;
  for (m = 1; m < 12; m++) {
    days_before[m] = days_before[m-1] + days_in_month[m-1];
  }
  INIT = 1;
}

#define NLEAP(year) ((year)/4 - (year)/100 + (year)/400)
#define IS_LEAP(year) ((year)%4?0:(year)%100?1:(year)%400?0:1)
#define BASE_YEAR 1904
typedef unsigned long long int uquad;
/* day   range 1..31
   month range 0..11
   year  AD
   hour 0..23
   min  0..59
   sec  0..60 */
void date1904_to_seconds(int year, int month, int day,
		     int hour, int min,   int sec,
		     uquad *ndays, uquad *secs) {
  int ny, nleap;
  const uquad hours_in_day = 24;
  const uquad days_in_year = 365; /* Leap days handled separately */

  if (!INIT) init_dates();
  nleap = NLEAP(year) - NLEAP(BASE_YEAR-1);
#if 0
  fprintf(stdout, "%4.4d is %sa leap year.  ", year, IS_LEAP(year)?"":"not ");
#endif
  if (IS_LEAP(year) && month < 2) nleap--; /* No leap day before this day */
  ny = year - BASE_YEAR;
  *ndays = days_in_year * ny + days_before[month] + day - 1 + nleap;
  *secs = (((hours_in_day**ndays + hour) * 60 + min) * 60) + sec;
}

int seconds_to_date1904(uquad secs,
	                int *year, int *month, int *day,
		        int *hour, int *min,   int *sec) {
  uquad ndays, extra_days, ny, ndays_sofar, junk_secs;
  int   odd_secs, odd_mins, m;
#define OVERFLOW 1
#define seconds_per_day (24 * 60 * 60)
  if (!INIT) init_dates();

  ndays    = secs / seconds_per_day;
  odd_secs = secs % seconds_per_day;
  *sec     = odd_secs % 60;
  odd_mins = odd_secs / 60;
  *min     = odd_mins % 60;
  *hour    = odd_mins / 60;
/* Days to years is trickier because of leap years! */
/* Set latest possible date in case calculations overflow. */
  *year = LONG_MAX;
  *month = 11;
  *day = 31;

  ny = 0;
  extra_days = ndays;
  while (extra_days > 400) {
    ny += extra_days / 366;
    if (ny > (LONG_MAX - BASE_YEAR)) return OVERFLOW;
    date1904_to_seconds(ny + BASE_YEAR, 0, 1, 0, 0, 0, &ndays_sofar, &junk_secs);
    extra_days = ndays - ndays_sofar;
  }
  if (IS_LEAP(ny + BASE_YEAR)) {
    if (extra_days >= 366) {
      ny++;
      extra_days -= 366;
    }
  } else {
    if (extra_days >= 365) {
      ny++;
      extra_days -= 365;
    }
  }
  *year = ny + BASE_YEAR;
  /* Deal with 29 February. */
  if (IS_LEAP(ny + BASE_YEAR)) {
    if (extra_days > days_before[2]) {
      extra_days--;
    } else if (extra_days == days_before[2]) {
      *month = 1; /* February */
      *day   = 29;
      return 0;
    }
  }
  for (m = 0; m <= 10; m++) {
    if (extra_days < days_before[m+1]) break;
  }
  *month = m;
  *day = extra_days - days_before[m] + 1;
  return 0;
}

#if TEST == 1
int main(int argc, char *argv[]) {
  int y, m;
  uquad ndays, nsecs, old_ndays;

  for (y = 1904, old_ndays = 0; y <= 1906; y++) {
    for (m = 0; m <= 11; m++) {
      date1904_to_seconds(y, m, 1, 0, 0, 0, &ndays, &nsecs);
      fprintf(stdout, "%d/%2.2d/%4d = Day %7d is %+2d from above.\n",
	      1, m+1, y, (int) ndays, (int)(ndays - old_ndays));
      old_ndays = ndays;
      if (y >= 1906) break;
    }
  }
  for (y = 1999, m = 10; y <= 2001; y++) {
    for (; m <= 11; m++) {
      date1904_to_seconds(y, m, 1, 0, 0, 0, &ndays, &nsecs);
      fprintf(stdout, "%d/%2.2d/%4d = Day %7lld is %+2d from above.\n",
	      1, m+1, y, ndays, (int)(ndays - old_ndays));
      old_ndays = ndays;
      if (y >= 2001) break;
    }
    m = 0;
  }
  return 0;
}
#elif TEST == 2
int check_time(int year, int month, int day, int hour, int min, int sec) {
  uquad ndays, nsecs;
  int   y2, mon2, d2, h2, min2, s2;

  date1904_to_seconds(year, month, day, hour, min, sec, &ndays, &nsecs);
  seconds_to_date1904(nsecs, &y2, &mon2, &d2, &h2, &min2, &s2);
  fprintf(stdout, "%2d/%2.2d/%d %2.2d:%2.2d:%2.2d converts to %16.16llx "
	  "and back to %2d/%2.2d/%d %2.2d:%2.2d:%2.2d.",
	  day, month+1, year, hour, min, sec, nsecs,
	  d2, mon2+1, y2, h2, min2, s2);
  if (day != d2 || month != mon2 || year != y2 ||
      hour != h2 || min != min2 || sec != s2) {
    fprintf(stdout, " ** ERROR **\n");
  } else {
    fputc('\n', stdout);
  }
}

int main(int argc, char *argv[]) {
  int year, month, day, hour, min, sec;
  uquad ndays, nsecs;
  enum {JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC};
  check_time(1904, JAN,  1,  0,  0,  0);
  check_time(1904, FEB, 27, 12, 30, 45);
  check_time(1904, FEB, 28, 15, 43, 26);
  check_time(1904, FEB, 29,  6, 15,  7);
  check_time(1904, MAR,  1,  7,  0,  0);
  check_time(1904, APR,  1,  9,  0,  0);
  check_time(1904, APR, 30,  2,  3,  4);
  check_time(1904, DEC, 31, 23, 59, 48);
  check_time(1905, JAN,  1,  0,  0,  0);
  check_time(1905, FEB, 27, 12, 30, 45);
  check_time(1905, FEB, 28, 15, 43, 26);
  check_time(1905, MAR,  1,  7,  0,  0);
  check_time(1905, APR,  1,  9,  0,  0);
  check_time(1905, APR, 30,  2,  3,  4);
  check_time(1905, DEC, 31, 23, 59, 48);
  check_time(2000, JAN,  1,  0,  0,  0);
  check_time(2000, FEB, 27, 12, 30, 45);
  check_time(2000, FEB, 28, 15, 43, 26);
  check_time(2000, FEB, 29,  6, 15,  7);
  check_time(2000, MAR,  1,  7,  0,  0);
  check_time(2000, APR,  1,  9,  0,  0);
  check_time(2000, APR, 30,  2,  3,  4);
  check_time(2000, DEC, 31, 23, 59, 48);
  check_time(2001, JAN,  1,  0,  0,  0);
  check_time(2001, FEB, 27, 12, 30, 45);
  check_time(2001, FEB, 28, 15, 43, 26);
  check_time(2001, FEB, 29,  6, 15,  7);
  check_time(2001, MAR,  1,  7,  0,  0);
  /* Wed May 09 10:56:01 2001  =  00000000B71ED2C1 */
  check_time(2001, MAY,  9, 10, 56,  1);
  check_time(2001, APR,  1,  9,  0,  0);
  check_time(2001, APR, 30,  2,  3,  4);
  check_time(2001, DEC, 31, 23, 59, 48);
  check_time(2006, MAR, 19, 19, 36, 25);
  check_time(9999, FEB, 28,  0, 59,  17);
  check_time(999916, FEB, 29,  11, 15,  49);
  check_time(3456789, MAR, 15,  11, 15,  49);
  check_time(23456789, MAR, 15,  11, 15,  49);
  check_time(123456789, MAR, 19, 19, 36, 25);
  check_time(2123456789, MAR, 19, 19, 36, 25);
}
#endif
  
