/*
Manipulation of sets of integers.
Copyright (C) 2007-2020      Richard Wordingham

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.
*/
#include <stdio.h>
#include <stdlib.h>
#include "int_set.h"

#define not_written {fprintf(stderr, "Code missing at line %d of %s.\n",\
                            __LINE__, __FILE__); exit(1);}

int_set create_int_set(void) {return 0;}

void delete_int_set(int_set *x) {
  int_set next, pair;
  pair = *x;
  while (pair) {
    next = pair->next;
    free(pair);
    pair = next;
  }
  *x = 0;
}

int_set *copy_int_set(int_set from, int_set *to) {
  int_set prev, pair;
/* Check for copying to self */
  if (*to && from == *to) return to;
  delete_int_set(to);
  prev = 0;
  for (; from; from = from->next) {
    pair = malloc(sizeof(*pair));
    if (!pair) {
      return 0;
    } else if (prev) {
      prev->next = pair;
    } else {
      *to = pair;
    }
    pair->start = from->start;
    pair->end   = from->end;
    pair->next  = 0;
    prev = pair;
  }
  return to;
}

int_set *add_range(int start, int end, int_set *set) {
  struct int_range *curr, *prev, *newun, *nextun;

  if (!*set) {
    if (newun = malloc(sizeof(*newun))) {
      newun->start = start;
      newun->end   = end;
      newun->next  = 0;
      *set = newun;
      return set;
    } else {
      return 0;
    }
  } else {
    prev = 0;
    curr = *set;
    while (curr) {
      if (end + 1 < curr->start) { /* Slot in here */
	if (newun = malloc(sizeof(*newun))) {
	  newun->start = start;
	  newun->end   = end;
	  newun->next  = curr;
	  if (prev) {
	    prev->next = newun;
	  } else {
	    *set = newun;
	  }
	  curr = newun;
	  return set;
	} else {
	  return 0;
	}
      } else if (start > curr->end + 1) { /* Effect is beyond current range */
	prev = curr;
	curr = curr->next;
      } else if (end + 1 >= curr->start) {
	if (start < curr->start) curr->start = start;
	if (end <= curr->end) {
	  return set;
	} else if (nextun = curr->next) {
	  if (end + 1 >= curr->next->start) { /* Delete next range */
	    curr->end  = curr->next->end;
	    curr->next = curr->next->next;
	    free(nextun);
	  } else {
	    curr->end = end;
	    return set;
	  }
	} else {
	  if (end > curr->end) curr->end = end;
	  return set;
	}
      }
    }
    if (prev) {
      if (curr = newun = malloc(sizeof(*newun))) {
	newun->start = start;
	newun->end   = end;
	newun->next  = 0;
	prev->next   = newun;
	return set;
      } else {
	return 0;
      }
    } else { /* Error handling! */
      not_written; /* Should be inaccessible! */
    }
  }
}

int_set *add_set(int_set from, int_set *to) {
  struct int_range *ci, *curr, *prev, *newun, *nextun;

  if (!from) {
    return to;
  } else if (!*to) {
    return copy_int_set(from, to);
  } else {
    prev = 0;
    curr = *to;
    for (ci = from; ci; ci = ci->next) {
      while (curr) {
	if (ci->end + 1 < curr->start) { /* Slot in here */
	  if (newun = malloc(sizeof(*newun))) {
	    newun->start = ci->start;
	    newun->end   = ci->end;
	    newun->next  = curr;
	    if (prev) {
	      prev->next = newun;
	    } else {
	      *to = newun;
	    }
	    curr = newun;
	    goto next_ci;
	  } else {
	    return 0;
	  }
	} else if (ci->start > curr->end + 1) { /* Effect is beyond current range */
	  prev = curr;
	  curr = curr->next;
	} else if (ci->end + 1 >= curr->start) {
	  if (ci->start < curr->start) curr->start = ci->start;
	  if (ci->end <= curr->end) {
	    goto next_ci;
	  } else if (nextun = curr->next) {
	    if (ci->end + 1 >= curr->next->start) { /* Delete next range */
	      curr->end  = curr->next->end;
	      curr->next = curr->next->next;
	      free(nextun);
	    } else {
	      curr->end = ci->end;
	      goto next_ci;
	    }
	  } else {
	    if (ci->end > curr->end) curr->end = ci->end;
	    goto next_ci;
	  }
	}
      }
      if (prev) {
	if (curr = newun = malloc(sizeof(*newun))) {
	  newun->start = ci->start;
	  newun->end   = ci->end;
	  newun->next  = 0;
	  prev->next   = newun;
	  goto next_ci;
	} else {
	  return 0;
	}
      } else { /* Error handling! */
	not_written; /* Should be inaccessible! */
      }
    next_ci:;
    }
    return to;
  }
}

int_set *sub_set(int_set of, int_set *from) {
  struct int_range *cs, *curr, *prev, *newun, *nextun;

  if (of && *from) {
    prev = 0;
    curr = *from;
    for (cs = of; cs; cs = cs->next) {
      while (curr) {
	if (cs->end < curr->start) {
	  goto next_cs;
	} else if (cs->start > curr->end) {
	  prev = curr;
	  curr = curr->next;
	} else if (cs->start <= curr->start) {
	  if (cs->end < curr->end) {
	    curr->start = cs->end + 1;
	    goto next_cs;
	  } else {
	    nextun = curr->next;
	    free(curr);
	    if (prev) {
	      prev->next = nextun;
	    } else {
	      *from = nextun;
	    }
	    curr = nextun;
	  }
	} else if (cs->end < curr->end) {
	  if (newun = malloc(sizeof(*newun))) {
	    newun->start = cs->end + 1;
	    newun->end   = curr->end;
	    newun->next  = curr->next;
	    curr->end    = cs->start - 1;
	    curr->next   = newun;
	    prev = curr;
	    curr = newun;
	    goto next_cs;
	  } else {
	    return 0;
	  }
	} else {
	  curr->end = cs->start - 1;
	  prev = curr;
	  curr = curr->next;
	}
      }
    next_cs:;
    }
    return from;
  } else {
    return from;
  }
}

int_set *sub_range(int start, int end, int_set *set) {
  struct int_range *curr, *prev, *newun, *nextun;

  if (start > end) {
    return 0;
  } else if (*set) {
    prev = 0;
    curr = *set;
    while (curr) {
      if (end < curr->start) {
	return set;
      } else if (start > curr->end) {
	prev = curr;
	curr = curr->next;
      } else if (start <= curr->start) {
	if (end < curr->end) {
	  curr->start = end + 1;
	  return set;
	} else {
	  nextun = curr->next;
	  free(curr);
	  if (prev) {
	    prev->next = nextun;
	  } else {
	    *set = nextun;
	  }
	  curr = nextun;
	}
      } else if (end < curr->end) {
	if (newun = malloc(sizeof(*newun))) {
	  newun->start = end + 1;
	  newun->end   = curr->end;
	  newun->next  = curr->next;
	  curr->end    = start - 1;
	  curr->next   = newun;
	  return set;
	} else {
	  return 0;
	}
      } else {
	curr->end = start - 1;
	prev = curr;
	curr = curr->next;
      }
    }
    return set;
  } else {
    return set;
  }
}

void print_int_set(FILE *fp, int_set x) {
  if (!x) {
    fprintf(fp, "none");
  } else {
    while (x) {
      fprintf(fp, "%d", x->start);
      if (x->end != x->start) {
	fprintf(fp, " to %d", x->end);
      }
      if (x->next) fprintf(fp, " and ");
      x = x->next;
    }
  }
}

void printf_int_set(FILE *fp, const char *num, const char *none,
		    const char *to, const char *and, int_set x) {
  if (!x) {
    fputs(none, fp);
  } else {
    while (x) {
      fprintf(fp, num, x->start);
      if (x->end != x->start) {
	fputs(to, fp);
	fprintf(fp, num, x->end);
      }
      if (x->next) fputs(and, fp);
      x = x->next;
    }
  }
}
