/*
Generation of cmap table
Copyright (C) 2005-2017      Richard Wordingham

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.
*/
/*
Algorithm:

1. Determine continuous runs of encoding.

2. Design decision - continuous runs will be encoded as segments - no splitting to reduce table size.

3. For each run, look for an earlier affinely similar part encoding run.

4. Generate cmap accordingly.

*/
#include <stdlib.h>
#include <stdio.h>
#include "manip_structure.h"
#include "oft.h"

struct rawseg {
  const unsigned short int *glyphs; /* Array of glyph numbers */
  unsigned int start, end;    /* Character IDs */ 
  unsigned short int idglyph_index;
  short int idDelta;
  short int extends;
  short int contiguous;
};

#define UW_max 0xffff
void make_cmap_subtable4
  (int n, int cpt[], unsigned short int gid[], struct encoding *subtable) {
  int n_segments;
  struct rawseg *rawseg = 0;
  int n_rawseg, max_rawseg, iseg, neff;
  static const unsigned short int notdef = 0;
  int idglyph_next = 0;
  unsigned short int *idptr;

/* Determine continuous runs of encoding */ 
  max_rawseg = 4;
  rawseg = malloc(4*sizeof(struct rawseg));
  iseg  = 0;
  n_rawseg = 1;
  if (n > 0 && cpt[0] <= 0xfff) {
    int i;
    
    rawseg[0].glyphs = gid;
    rawseg[0].start = cpt[0];
    rawseg[0].contiguous = 1;
    neff = n;
    for (i=1; i < n; i++) {
      if (cpt[i] > UW_max) {
	neff = i;
	rawseg[iseg].end = cpt[i-1];
	break;
      }
      if (cpt[i] != cpt[i-1] + 1) {
	rawseg[iseg].end = cpt[i-1];
	if (n_rawseg >= max_rawseg - 1) {
	  max_rawseg = 1.5 * max_rawseg;
	  rawseg = realloc(rawseg, max_rawseg * sizeof(struct rawseg));
	}
	iseg++; n_rawseg++;
	rawseg[iseg].glyphs = & gid[i];
	rawseg[iseg].start = cpt[i];
        rawseg[iseg].contiguous = 1;
      } else if (gid[i] != gid[i-1] + 1) {
	rawseg[iseg].contiguous = 0;
      }
    }
    if (n == neff) {
      rawseg[iseg].end = cpt[n-1];
    }
  } else {
    rawseg[0].start = UW_max;
    rawseg[0].end = UW_max;
    rawseg[0].glyphs = &notdef;
    rawseg[0].contiguous = 1;
  }
  if (UW_max != rawseg[iseg].end) {
    iseg++; n_rawseg++;
    rawseg[iseg].glyphs = &notdef;
    rawseg[iseg].start = UW_max;
    rawseg[iseg].end   = UW_max;
    rawseg[iseg].contiguous = 1;
  } 
  
/* Look for parallel runs - all or nothing, to simplify generation
   of the glyphID array. */
  for (iseg = 0; iseg < n_rawseg; iseg++) {
    int jseg, longer, offset, k, kj, match, delta, delta_got, new_delta;
    
    if (rawseg[iseg].contiguous) {
      rawseg[iseg].extends = 0;
    } else {
      match = 0;
      for (jseg = 0; jseg < iseg; jseg++) {
	if (!rawseg[jseg].extends) continue;
	longer = (rawseg[jseg].end - rawseg[jseg].start) -
	  (rawseg[iseg].end - rawseg[iseg].start);
	if (longer < 0) continue; /* Too short */
	for (offset = 0; offset <= longer; offset++) {
	  match = 1;
	  delta_got = 0;
	  for (k = 0; k <= rawseg[iseg].end - rawseg[iseg].start; k++) {
	    kj = k + offset;
	    if (0 == rawseg[jseg].glyphs[kj]) {
	      if (rawseg[jseg].glyphs[k]) {
		match = 0;
		break;
	      }
	    } else if (rawseg[iseg].glyphs[k]) {
	      new_delta = rawseg[iseg].glyphs[k] - rawseg[jseg].glyphs[kj];
	      if (delta_got) {
		if (new_delta != delta) {
		  match = 0;
		  break;
		}
	      } else {
		delta = new_delta;
		delta_got = 1;
	      }
	    } else {
	      match = 0;
	      break;
	    }
	  }
	  if (1 == match) break;
	}
	if (1 == match) {
	  rawseg[iseg].extends = 0;
	  rawseg[iseg].idDelta = delta;
	  rawseg[iseg].idglyph_index = rawseg[jseg].idglyph_index + offset;
	  break;
	}
      }
      if (!match) {
	rawseg[iseg].extends = 1;
	rawseg[iseg].idDelta = 0;
	rawseg[iseg].idglyph_index = idglyph_next;
	idglyph_next += rawseg[iseg].end - rawseg[iseg].start + 1;
      }
    }
#if 0
    if (rawseg[iseg].extends) printf("Segment %d extends.\n", iseg);
    if (rawseg[iseg].contiguous) printf("Segment %d is contiguous.\n", iseg);
#endif
  }
  /* Now generate the cmap subtable */
  subtable->format = 4;
  subtable->length = 2*(8 + 4 * n_rawseg + idglyph_next);
  subtable->language = 0;
  subtable->segCountX2 = 2 * n_rawseg;
  subtable->searchRange = 2;
  subtable->entrySelector = 0;
  while (subtable->searchRange <= n_rawseg) {
    subtable->searchRange *= 2;
    subtable->entrySelector++;
  }
  subtable->rangeShift = 2 * n_rawseg - subtable->searchRange;
  subtable->_n_endCount = n_rawseg;
  subtable->endCount = malloc(2*n_rawseg);
  subtable->reservedPad = 0;
  subtable->_n_startCount = n_rawseg;
  subtable->startCount    = malloc(2*n_rawseg);
  subtable->_n_startCount = n_rawseg;
  subtable->idDelta       = malloc(2*n_rawseg);
  subtable->_n_idRangeOffset = n_rawseg;
  subtable->idRangeOffset = malloc(2*n_rawseg);  
  for (iseg = 0; iseg < n_rawseg; iseg++) {
    subtable->startCount[iseg] = rawseg[iseg].start;
    subtable->endCount[iseg]   = rawseg[iseg].end;
    if (rawseg[iseg].contiguous) {
      subtable->idRangeOffset[iseg] = 0;
      subtable->idDelta[iseg] = *rawseg[iseg].glyphs - rawseg[iseg].start;
    } else {
      subtable->idRangeOffset[iseg] =
	2 * rawseg[iseg].idglyph_index + 2 * (n_rawseg - iseg);
      subtable->idDelta[iseg] = rawseg[iseg].idDelta;
    }
  }
  subtable->_n_glyphIdArray = idglyph_next;
  idptr = subtable->glyphIdArray = malloc(2*idglyph_next);
  for (iseg = 0; iseg < n_rawseg; iseg++) {
    int k;
    if (rawseg[iseg].extends) {
#if 0
      printf("Extending glyphIdArray for segment %d.\n", iseg);
#endif
      for (k = 0; k <= rawseg[iseg].end - rawseg[iseg].start; k++) {
	*idptr++ = rawseg[iseg].glyphs[k];
      }
    }
  }
}

void make_cmap_subtable12
  (int n, int cpt[], unsigned short int gid[], struct encoding *subtable) {

  /*
STRUCTURE(format_8_group)
     DATA_UL(startCharCode)
     DATA_UL(endCharCode)
     DATA_UL(startGlyphID)
END_STRUCTURE(format_8_group)
STRUCTURE(encoding)
     DATA_UW(format)
     DATA_UW(reserved)
     DATA_UL(length)
     DATA_UL(language)
     DATA_UL(nGroups)
     ARRAY_VC(groups, here->nGroups, format_8_group)
END_STRUCTURE(encoding)
  */
  int i, max_gp, prev;

  subtable->format = 12;
  subtable->reserved = 0;
  subtable->language = 0;
  subtable->nGroups = 0;
  subtable->groups = 0;
  if (n > 0) {
    max_gp = 0; /* No memory yet for groups */
    for (i = 0; i < n; i++) {
      if (i && gid[i] == gid[i-1] + 1 && cpt[i] == cpt[i-1] + 1) {
	/* Carry on with current group. */
      } else {
	if (i) {
	  subtable->groups[subtable->nGroups - 1].endCharCode = cpt[i-1];
	}
	subtable->nGroups ++;
	if (subtable->nGroups > max_gp) {
	  max_gp = 4 + max_gp + max_gp / 2;
	  if (!(subtable->groups =
		realloc(subtable->groups,
			max_gp * sizeof(*subtable->groups)))) {
	    fprintf(stderr, "Memory exhausted at line %d of %s.\n",
		    __LINE__, __FILE__);
	    exit(1);
	  }
	}
	subtable->groups[subtable->nGroups - 1].startCharCode = cpt[i];
	subtable->groups[subtable->nGroups - 1].startGlyphID  = gid[i];
      }
    }
    subtable->groups[subtable->nGroups - 1].endCharCode = cpt[n-1];
  }
  /*  subtable->_n_groups = subtable->nGroups; */
  subtable->length = 16 + 12 * subtable->nGroups;
  for (i = 0; i < subtable->nGroups; i++) {
    subtable->groups[i]._next = &subtable->groups[i+1];
  }
  subtable->groups[i]._next = 0;
}
