/*
Maps from strings to pointers.
Copyright (C) 2007      Richard Wordingham

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
/* Include common definition. */
#include "map.h"

enum entry_status {FREE, IN_USE, DELETED};
/* Deletion of entries not yet implemented! */

void create_map(struct map *the_map) {
  the_map->capacity = 0;
  the_map->usage = 0;
  the_map->x = 0;
}

void delete_map(struct map *the_map, int free_value) {
  int k;

  for (k = 0; k < the_map->capacity; k++) {
    if (IN_USE == the_map->x[k].status) {
      free(the_map->x[k].key);
      if (free_value) free(the_map->x[k].datum);
    }
  }
  free(the_map->x);
  the_map->capacity = 0;
  the_map->x = 0;
}

static int hash(const char *s) {

  int value = 0;
  while (1) {
#if 0
    if (!*s) return value;
    value += 131 * *s++;
    if (!*s) return value;
    value += 17 * *s++;
    if (!*s) return value;
    value *=2;
    value += 37 * *s++;
#else
    if (!*s) return value;
    value *= 107;
    value += *s++;
#endif
  }
}

int lookup_in_map(struct map the_map, const char *key, void **val) {
/* Returns 1 if found, 0 if not. */
  int locn, k;

  *val = 0;
  if (!the_map.capacity) return 0;
  locn = hash(key) % the_map.capacity;
  if (locn < 0) locn = -locn;
#if 0
  printf("lookup_in_map: %s hashes to %d.\n", key, locn);
#endif
  for (k = locn; k < the_map.capacity; k++) {
    if (IN_USE == the_map.x[k].status) {
      if (!strcmp(key, the_map.x[k].key)) {
	*val = the_map.x[k].datum;
	return 1;
      }
    } else if (FREE == the_map.x[k].status) {
      return 0;
    }
  }
  for (k = 0; k < locn; k++) {
    if (IN_USE == the_map.x[k].status) {
      if (!strcmp(key, the_map.x[k].key)) {
	*val = the_map.x[k].datum;
	return 1;
      }
    } else if (FREE == the_map.x[k].status) {
      return 0;
    }
  }
  return 0;
}

static void put_in_map(struct map *the_map, const char *key, void *val,
		       int may_replace, int free_value) {
  int locn, k, put_here, iter, old_capacity;
  struct map_entry *old_x;

  put_here = 0;
  for (iter = 1; iter <= 2; iter++) {
    if (the_map->capacity) {
      locn = hash(key) % the_map->capacity;
      if (locn < 0) locn = -locn;
    } else {
      locn = 0;
    }
#if 0
    printf("put_in_map: %s hashes to %d.\n", key, locn);
#endif
    if (the_map->usage < the_map->capacity / 2) {
      for (k = locn; k < the_map->capacity; k++) {
	switch (the_map->x[k].status) {
	case FREE:
	case DELETED:
	  put_here = 1;
	  break;
	case IN_USE:
	  if (!strcmp(key, the_map->x[k].key)) {
	    put_here = 1;
	    break;
	  }
	}
	if (put_here) break;
      }
      if (!put_here) {
	for (k = 0; k < locn; k++) {
	  switch (the_map->x[k].status) {
	  case FREE:
	  case DELETED:
	    put_here = 1;
	    break;
	  case IN_USE:
	    if (!strcmp(key, the_map->x[k].key)) {
	      put_here = 1;
	      break;
	    }
	  }
	  if (put_here) break;
	}
      }
    }
    if (put_here) {
      if (IN_USE == the_map->x[k].status) {
	if (may_replace) {
	  free(the_map->x[k].key);
	  if (free_value) free(the_map->x[k].datum);
	} else {
	  fprintf(stderr, "Cannot replace mapping for %s.\n",
		  the_map->x[k].key);
	  exit(1);
	}
      }
      the_map->x[k].key = malloc(strlen(key)+1);
      strcpy(the_map->x[k].key, key);
      if (FREE == the_map->x[k].status) the_map->usage++;
      the_map->x[k].status = IN_USE;
      the_map->x[k].datum = val;
#if 0
      printf("put_in_map: Inserted value for %s.\n", key);
#endif
      break;
    } else if (1 == iter) {
      old_capacity = the_map->capacity;
      old_x = the_map->x;
      the_map->capacity = 2.3 * old_capacity + 10;
      the_map->x = malloc(the_map->capacity * sizeof(struct map_entry));
      for (k = 0; k < the_map->capacity; k++) {
	the_map->x[k].status = FREE;
      }
      the_map->usage = 0;
      for (k = 0; k < old_capacity; k++) {
	if (IN_USE == old_x[k].status) {
	  insert_in_map(the_map, old_x[k].key, old_x[k].datum);
	  free(old_x[k].key);
	  the_map->usage++;
	}
      }
      free(old_x);
    } else {
      fprintf(stderr, "put_in_map: Insertion did not converge!");
      exit(1);
    }
  }
}

int delete_from_map(struct map *the_map, const char *key, int free_value) {
/* Returns 1 if found, 0 if not. */
  int locn, k;

  if (!the_map->capacity) return 0;
  locn = hash(key) % the_map->capacity;
  if (locn < 0) locn = -locn;
#if 0
  printf("lookup_in_map: %s hashes to %d.\n", key, locn);
#endif
  for (k = locn; k < the_map->capacity; k++) {
    if (IN_USE == the_map->x[k].status) {
      if (!strcmp(key, the_map->x[k].key)) {
	if (free_value) free(the_map->x[k].datum);
	the_map->x[k].status = DELETED;
	return 1;
      }
    } else if (FREE == the_map->x[k].status) {
      return 0;
    }
  }
  for (k = 0; k < locn; k++) {
    if (IN_USE == the_map->x[k].status) {
      if (!strcmp(key, the_map->x[k].key)) {
	if (free_value) free(the_map->x[k].datum);
	the_map->x[k].status = DELETED;
	return 1;
      }
    } else if (FREE == the_map->x[k].status) {
      return 0;
    }
  }
  return 0;
}


/* The map takes ownership of *val as the result of a successful call.
   It makes its own copy of *key      */
void replace_in_map(struct map *the_map, const char *key, void *val,
		    int free_value) {
  put_in_map(the_map, key, val, 1, free_value);
}
  
void insert_in_map(struct map *the_map, const char *key, void *val) {
  put_in_map(the_map, key, val, 0, 0);
}

static int compare_index(struct map_entry *left, struct map_entry *right) {
  return strcmp(left->key, right->key);
}
#define ipsort entry_sort
#define DTYPE  struct map_entry*
#include "ipsort.h"
#undef DTYPE
#undef ipsort

void sort_map(struct map the_map, struct map_entry ***entries, int *n_entries){
  int i;
  struct map_entry **latest;

  *n_entries = 0;
  *entries = malloc((the_map.capacity+1)*sizeof(struct map_entry*));
  if (!*entries) {
    fprintf(stderr, "sort_map: memory exhuasted.\n");
    exit(1);
  }
  latest = *entries;
  for (i = 0; i < the_map.capacity; i++) {
    if (IN_USE == the_map.x[i].status) {
#if 0
      fprintf(stdout, "Key = %s\n", the_map.x[i].key);
#endif
      *latest = &the_map.x[i];
      latest++;
      (*n_entries)++;
#if 0
      if (*n_entries >= 3) break;
#endif
    }
  }
  *latest = 0;
  entry_sort(compare_index, *entries, *n_entries);
#if 0
  for (i = 0; i < *n_entries; i++) {
    fprintf(stdout, "Key = %s\n", (*entries)[i]->key);
  }
#endif
}
