/*
OTL portion of OpenType compiler
Copyright (C) 2015-2020      Richard Wordingham

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.
*/
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "int_set.h"
#include "map.h"
#include "tokeniser.h"
#include "parse.h"
/* Do not disturb the order of the next three includes and the redefinition of
   macro STRUCTURE. */
#include "manip_structure.h"
#undef STRUCTURE
#define STRUCTURE(type) struct type;					\
  void put_bine_ ## type (FILE* fp, struct type *here, int*follow, int*pos); \
  void free_ ## type(struct type* here);                                \
  struct type {struct type *_next;
#include "oft.h"
/* Types for easier writing of font structures. */
#include "gpos_io.h"
static int pair_fmt;
    /* Do not disturb the order of the next three includes */
#include "manip_befile.h"
#include "manip_put.h"
#include "gpos_io.h"
    /* Do not disturb the order of the next two includes */
#include "manip_free.h"
#include "gpos_io.h"

/* Desperate debugging aids */
#define DOFREE 1
#if 0
    /* Do not disturb the order of the next three includes */
#include "manip_print.h"
#include "manip_display.h"
#include "oft.h"
#endif


/* Encapsulate access to the maps from strings to pointers */ 
#define encapsulate_map(OBJECT, DATATYPE)		\
core_encapsulate_map(OBJECT)		\
extend_encapsulate_map(OBJECT, DATATYPE)

#define extend_encapsulate_map(OBJECT, DATATYPE)		\
struct OBJECT##_map_entry {\
  char *key;\
  DATATYPE *datum;\
  char status;\
};\
static int find_##OBJECT(struct OBJECT##_map ft, const char *featname,\
			DATATYPE **data) {\
  *data = 0;\
  return lookup_in_map(ft.x, featname, (void**) data);\
}\
static void add_##OBJECT(struct OBJECT##_map *ft, const char *featname,\
			DATATYPE *data) {\
  insert_in_map(&ft->x, featname, data);\
}\
static void sort_##OBJECT##_map(struct OBJECT##_map the_map,\
			     struct OBJECT##_map_entry ***entries,\
			     int *n_entries) {\
  struct map_entry **map_entries;\
\
  sort_map(the_map.x, &map_entries, n_entries);\
  *entries = (struct OBJECT##_map_entry **) map_entries;\
}\
 static void delete_##OBJECT##_map(struct OBJECT##_map *the_map,\
                                   int free_value) {\
  delete_map(&the_map->x, free_value);\
}\
DATATYPE *provide_##OBJECT(struct OBJECT##_map *ft,\
                           const char *featname, int *isnew) {	\
  DATATYPE *featdat;\
\
  if (lookup_in_map(ft->x, featname, (void**) &featdat)) {\
    /* Feature already encountered */\
    if (isnew) *isnew = 0;				\
  } else if (featdat = malloc(sizeof * featdat)) {\
    create_##OBJECT(featdat);\
    insert_in_map(&ft->x, featname, featdat);\
    if (isnew) *isnew = 1;		     \
  } else {\
    fprintf(stderr, "Memory exhausted!\n");\
    exit(1);\
  }\
  return featdat;\
}

#define no_providing(OBJECT, DATATYPE)			\
static void create_##OBJECT(DATATYPE *data) {\
  fprintf(stderr, "Internal error in %s: Function provide_" #OBJECT\
	  "() should never be invoked.\n", __FILE__);\
}
  

#define GSUB_1 GSUB_SINGLE
#define GSUB_9 GSUB_REVERSE /* Last GSUB code */
#define GPOS_1 GPOS_SINGLE
#define GPOS_9 GPOS_EXTENSION /* Last GPOS code */
#define GSUB_ADJ (GSUB_1 - 1)
#define GPOS_ADJ (GPOS_1 - 1)
static const struct locations no_locations = {0, 0, 0};
struct lookup_data {
  struct subtable_map subtables;
  int lookup_index; /* -1 => not yet in lookup list */
  unsigned int own_loc; /* Location of look-up in table */
  struct ptr_spec ref;
  enum lookup_type type;
  enum storedness stored;
  unsigned short int flags;
  unsigned short int n_subtables;
  unsigned short int mark_set;
  unsigned char is_extended;
};
struct subtable_data {
  struct lookup_data *lookup; /* Parent look-up */
  int number; /* Subtable number, from 0 */
  struct ptr_spec ref;
  enum storedness stored;
};

#define unwritten { \
  fprintf(stderr, "Code missing at line %d of %s.\n", __LINE__, __FILE__); \
  exit(1); \
}

void coda_end_syntax(const char *str) {
/* On entry, current token is identifier 'end' or not an identifier, */
  if (AN_IDENT != current.form) {
    show_source();
    fprintf(stderr, "Expected an identifier.\n");
  } else {
    advance_token();
    if (AN_IDENT != current.form || strcmp(str, current.sval)) {
      show_source();
      fprintf(stderr, "Expected identifier '%s'.\n", str);
      got_what(stderr, current);
    }
    advance_token();
  }
}

int right_lookup_type(const char* tt, enum lookup_type it) {
  if (GPOS_1 <= it && it <= GPOS_9) {
    if (strcmp("GPOS", tt)) {
      show_source();
      fprintf(stderr, "Lookup is already assigned to GSUB.\n");
      return 0;
    } else {
      return 1;
    }
  } else if (GSUB_1 <= it && it <= GSUB_9) {
    if (strcmp("GSUB", tt)) {
      show_source();
      fprintf(stderr, "Lookup is already assigned to GPOS.\n");
      return 0;
    } else {
      return 1;
    }
  } else {
    fprintf(stderr, "Internal error tracking lookup assignments.\n");
    exit(1);
  }
}

static void create_lookup(struct lookup_data *ld) {
  create_subtable_map(&ld->subtables);
  ld->lookup_index = -1;
  ld->stored = PENDING;
}

extend_encapsulate_map(lookup, struct lookup_data)

struct feature_data {
  struct lookup_map lookups;
  int number;
  int defined;
};

static void create_feature(struct feature_data *featdat) {
    create_lookup_map(&featdat->lookups);
    featdat->number = -1;
    featdat->defined = 0;
}

encapsulate_map(feature, struct feature_data)

void gsbp_seek(FILE *fp, int loc) {
  if (fseek(fp, loc, SEEK_SET)) {
    fprintf(stderr, "Error seeking within GSUB/GPOS scratch file.\n");
    exit(1);
  }
}

static void gen_lang_sys(FILE *fp, const char* lang,
			 const struct feature_map * const feats, int *pos) {
  struct feature_map_entry **fl2; /* Features as sorted tags and data */
  struct lang_sys fl3;    /* Open type feature list */
  int ift, fno, n_feats, junk;
  struct feature_data *featdat;
  unsigned short int *index;

  sort_feature_map(*feats, &fl2, &n_feats);
  memset(&fl3, 0, sizeof fl3);
  fl3.lookupOrder = 0;
  fl3.reqFeatureIndex = 0xFFFF;
  fl3.featureCount = n_feats;
  if (n_feats) {
    index = fl3.featureIndex = malloc(n_feats * sizeof * fl3.featureIndex);
    if (!fl3.featureIndex) {
      fprintf(stderr, "Memory exhausted.\n");
      exit(1);
    }
    for (ift = 0; ift < (n_feats - 1); ift++) {
      char *f1, *f2;
      f1 = fl2[ift]->key;
      f2 = fl2[ift+1]->key;
      if (!strncmp(f1, f2, 4)) {
	fprintf(stderr, "Warning: Features %s and %s of language %s have "
		"same tag, namely %.4s.\n", f1, f2, lang, f1);
      }
    }
    for (ift = 0; ift < n_feats; ift++) {
      featdat = fl2[ift]->datum;
      fl3.featureIndex[ift] = featdat->number;
#if 0
      fprintf(stdout, "Recorded number %d for feature %s in data at %p"
	      " for language %s.\n",
	      ift, fl2[ift]->key, fl2[ift]->datum, lang);
#endif
    }
  } else {
    fl3.featureIndex = 0;
  }
  put_bine_lang_sys(fp, &fl3, &junk, pos);
  /* It is tempting to call free_lang_sys(&fl3), but this does not work
     if an array has bee nallocated in a single call! */
  if (DOFREE) free(fl3.featureIndex);
  if (DOFREE) free(fl2);
}

#define create_language create_feature_map
encapsulate_map(language, struct feature_map)
#define create_script create_language_map
encapsulate_map(script, struct language_map)

static void gen_per_script_table(FILE *gsbp, const char *script,
				 const struct language_map *const langs,
				 int *pos) {
  /* Think of as Language List (ll) */
  struct language_map_entry **ll1all, **ll1nd; /* ll1all includes default, if any */
  struct script_table ll2;
  struct lang_sys_record **lsr_nexloc;
  int il1, il2, l_count;
  int loc, script_table_loc;
  int junk;

  sort_language_map(*langs, &ll1all, &l_count);
  memset(&ll2, 0, sizeof ll2);
  if (!l_count) {
    fprintf(stderr, "Warning: No languages for script %.4s.\n", script);
    ll1nd = 0;
  } else if (*ll1all[0]->key) { /* No default language */
    ll1nd = ll1all;
    ll2.langSysCount = l_count;
    if (!strcmp("DFLT", script)) {
      fprintf(stderr, "Default script, DFLT, should have a default language, "
	      "but does not.\n");
    }
  } else { /* Have default language */
    ll1nd = ll1all + 1;
    ll2.langSysCount = l_count - 1;
  }
  if (ll2.langSysCount) {
    ll2.langSysRecord =
      malloc(ll2.langSysCount * sizeof * ll2.langSysRecord);
    if (!ll2.langSysRecord) {
      fprintf(stdout, "Memory exhausted!\n");
      exit(1);
    }
    lsr_nexloc = &ll2.langSysRecord;
    for (il2 = 0; il2 < ll2.langSysCount; il2++) {
      *lsr_nexloc = &ll2.langSysRecord[il2];
      memcpy(ll2.langSysRecord[il2].tag, ll1nd[il2]->key, 4);
      ll2.langSysRecord[il2].langSys = 0;
      lsr_nexloc = &ll2.langSysRecord[il2]._next;
      *lsr_nexloc = 0;
    }
  }
  script_table_loc = *pos;
  put_bine_script_table(gsbp, &ll2, &junk, pos);
/* Feature list for default language system */
  if (*ll1all && !*ll1all[0]->key) {
    ll2.defaultLangSys = *pos - script_table_loc;
    gen_lang_sys(gsbp, "(default)", ll1all[0]->datum, pos);
  }
  for (il2 = 0; il2 < ll2.langSysCount; il2++) {
    ll2.langSysRecord[il2].langSys = *pos - script_table_loc;
    gen_lang_sys(gsbp, ll1nd[il2]->key, ll1nd[il2]->datum, pos);
  }
/* Now rewrite updated script table */
  gsbp_seek(gsbp, loc = script_table_loc);
  put_bine_script_table(gsbp, &ll2, &junk, &loc);
  gsbp_seek(gsbp, *pos);
/* Release heap storage */
  free(ll2.langSysRecord);
  free(ll1all);
}

static void gen_script_table(FILE *gsbp,
			     const struct script_map scripts,
			     int *pos) {
  int isc, count, junk;
  int loc, sl_loc;
  struct script_map_entry **sl1;
  struct script_list sl2;
  struct script_record **sr_nexloc;

  sort_script_map(scripts, &sl1, &count);
  memset(&sl2, 0, sizeof sl2);
  sl2.scriptCount = count; /* Should check in bounds */
  sl2.scriptRecord = malloc(sl2.scriptCount * sizeof * sl2.scriptRecord);
  if (!sl2.scriptRecord) {fprintf(stderr, "Memory exhausted!\n");  exit(1);}
  sr_nexloc = &sl2.scriptRecord;
  for (isc = 0; isc < sl2.scriptCount; isc++) {
    *sr_nexloc = &sl2.scriptRecord[isc];
    memcpy(sl2.scriptRecord[isc].tag, sl1[isc]->key, 4);
    sl2.scriptRecord[isc].script = 0;
    sr_nexloc = &sl2.scriptRecord[isc]._next;
    *sr_nexloc = 0;
  }
  sl_loc = *pos;
  put_bine_script_list(gsbp, &sl2, &junk, pos);
  for (isc = 0; isc < sl2.scriptCount; isc++) {
    sl2.scriptRecord[isc].script = *pos - sl_loc;
    gen_per_script_table(gsbp, sl1[isc]->key, sl1[isc]->datum, pos);
  }
/* Now rewrite updated script list table */
  gsbp_seek(gsbp, loc = sl_loc);
  put_bine_script_list(gsbp, &sl2, &junk, &loc);
  gsbp_seek(gsbp, *pos);
  free(sl2.scriptRecord);
  free(sl1);
}

static void gen_per_feature_table(
  FILE *gsbp, const char *name, const struct feature_data *featdat, int *pos) {
  struct feature_table ftbl;
  int_set lookup_nos;
  struct int_range *irge;
  int k, count, lno, n_ok;
  struct lookup_map_entry **lkl;
  struct lookup_data *ld;
  unsigned short int *lli;
  int junk;

  ftbl.featureParams = 0;
  if (featdat->defined) {
    sort_lookup_map(featdat->lookups, &lkl, &count);
    n_ok = 0;
    lookup_nos = create_int_set();
    for (k = 0; k < count; k++) {
      ld = lkl[k]->datum;
      lno = ld->lookup_index;
      if (lno < 0) {
	fprintf(stderr, "Lookup %s discarded from feature %s because "
		"undefined.\n", lkl[k]->key, name);
      } else {
	(void) add_range(lno, lno, &lookup_nos);
	n_ok++;
      }
    }
    lli = malloc(n_ok * sizeof * lli);
    if (n_ok && !lli) {fprintf(stderr, "Memory exhausted!\n"); exit(1);}
    ftbl.lookupListIndex = lli;
    for (irge = lookup_nos; irge; irge = irge->next) {
      for (k = irge->start; k <= irge->end; k++) {
	*lli++ = k;
      }
    }
    ftbl.lookupCount = n_ok;
  } else {
    ftbl.lookupCount = 0;
    ftbl.lookupListIndex = 0;
  }
  put_bine_feature_table(gsbp, &ftbl, &junk, pos);
  free(ftbl.lookupListIndex);
}

static void gen_feature_table(FILE *gsbp, struct feature_map_entry **fl1,
			      int n_features, int *pos) {
  const int fl_loc = *pos;
  struct feature_list fl2;
  struct feature_record **fr_next;
  int ift, junk;
  int loc;
  struct feature_data *featdat;

  fl2.featureCount = n_features;
  fl2.featureRecord = malloc(n_features * sizeof * fl2.featureRecord);
  if (!fl2.featureRecord) {fprintf(stderr, "Memory exhausted!.\n"); exit(1);}
  fr_next = &fl2.featureRecord;
  for (ift = 0; ift < n_features; ift++) {
    *fr_next = &fl2.featureRecord[ift];
    fr_next = &fl2.featureRecord[ift]._next;
    *fr_next = 0;
    memcpy(fl2.featureRecord[ift].tag, fl1[ift]->key, 4);
    fl2.featureRecord[ift].feature = 0;
  }
  put_bine_feature_list(gsbp, &fl2, &junk, pos);
  for (ift = 0; ift < n_features; ift++) {
    fl2.featureRecord[ift].feature = *pos - fl_loc;
    gen_per_feature_table(gsbp, fl1[ift]->key, fl1[ift]->datum, pos);
  }
  gsbp_seek(gsbp, loc = fl_loc);
  put_bine_feature_list(gsbp, &fl2, &junk, &loc);
  gsbp_seek(gsbp, *pos);
  free(fl2.featureRecord);
}

no_providing(subtable, struct subtable_data)
extend_encapsulate_map(subtable, struct subtable_data)

static int subtable_offset_locn(int start, int index) {
  return start + 6 + 2 * index;
}

void store_lookup(const char *name, struct lookup_data *data) {
  int *posn, lkploc, loc;
  FILE *fp;
  int adjust, junk, isub, n_subs, offset;
  struct lookup_table table;
  size_t size;
  struct subtable_map_entry **subs;
  struct subtable_data *subdata;
  enum lookup_type extend;

  if (WRITTEN == data->stored) return; /* Nothing to do */
  if (GSUB_1 <= data->type && data->type <= GSUB_9) {
    posn = &gsub_pos;
    fp = gsub;
    adjust = GSUB_ADJ;
    extend = GSUB_EXTENSION;
  } else if (GPOS_1 <= data->type && data->type <= GPOS_9) {
    posn = &gpos_pos;
    fp = gpos;
    adjust = GPOS_ADJ;
    extend = GPOS_EXTENSION;
  } else {
    fprintf(stderr, "Internal error in store_lookup(): "
	    "unexpected unified lookup type %d encountered for lookup %s.\n",
	    data->type, name);
    exit(1);
  }
/* Write lookup table, with offsets to subtables zeroed (not yet known) */
  if (data->is_extended) {
    table.lookupType = extend - adjust;
  } else {
    table.lookupType = data->type - adjust;
  }
  table.lookupFlag = data->flags;
  table.subTableCount = data->n_subtables;
  size = data->n_subtables * sizeof * table.subTable;
  if (table.subTable = malloc(size)) {
    memset(table.subTable, 0, size);
  } else {
    fprintf(stderr, "Memory exhausted!\n"); exit(1);
  }
  table.markFilteringSet = data->mark_set;
  lkploc = *posn;
  put_bine_lookup_table(fp, &table, &junk, posn);
/* Check formula for location of offsets to subtables. */
  if (data->flags & 16) {
    assert(subtable_offset_locn(lkploc, data->n_subtables) + 2 == *posn);
  } else {
    assert(subtable_offset_locn(lkploc, data->n_subtables) == *posn);
  }
/* Record the offset to the lookup */
  offset = lkploc - data->ref.offset_ref;
  gsbp_seek(fp, loc = data->ref.offset_loc);
  if (0 < offset && 4 == data->ref.offset_size) {
    /* No problem */
  } else if (0 < offset && offset < (1<<16)) {
    /* No problem */
  } else {
    if (offset <= 0) {
      fprintf(stderr, "Offset calculation for lookup %s seems corrupt:\n",
	      name);
    } else {
      fprintf(stderr, "Attempted to store lookup %s too far "
	      "from lookuplist:\n", name);
    }
    fprintf(stderr, "  Lookuplist location:       %12d\n"
	            "  Offset storage location:   %12d\n"
	            "  Lookup %3d table location: %12d\n",
	    data->ref.offset_ref, data->ref.offset_loc, data->lookup_index,
	    lkploc);
    offset = 0; /* Might avoid massive damage */
  }
  if (4 == data->ref.offset_size) {
    unwritten; /* Don't think this is actually possible! */
  } else {
    struct UW_item be_offset;
    be_offset.val = offset;
    put_bine_UW_item(fp, &be_offset, &junk, &loc); 
  }
  gsbp_seek(fp, *posn);
  data->stored = WRITTEN;
/* Record data for calculation and storage of offsets once the location is
   known. */
  sort_subtable_map(data->subtables, &subs, &n_subs);
  if (n_subs != data->n_subtables) {
    fprintf(stderr, "Internal error: number of subtables of lookup %s "
	    "inconsistently held as %d and %d:\n",
	    name, n_subs, data->n_subtables);
    for (isub = 0; isub < n_subs; isub++) {
      subdata = subs[isub]->datum;
      fprintf(stderr, "%4d: %s\n", subdata->number, subs[isub]->key);
    }
    exit(1);
  }
  for (isub = 0; isub < n_subs; isub++) {
    subdata = subs[isub]->datum;
    subdata->ref.offset_ref  = lkploc;
    subdata->ref.offset_loc  = subtable_offset_locn(lkploc, subdata->number);
    subdata->ref.offset_size = 2;
    subdata->stored = DEFERRED;
  }
/* Free heap storage */
  free(table.subTable);
  free(subs);
}

void store_all_lookups() {
  struct lookup_map_entry **ll;
  struct lookup_data *ld;
  int i, n;

  sort_lookup_map(lookups, &ll, &n);
  for (i = 0; i < n; i++) {
    ld = ll[i]->datum;
    if (DEFERRED == ld->stored) store_lookup(ll[i]->key, ld);
  }
  free(ll);
}

static int lookup_offset_locn(int start, int index) {
  return start + 2 * (1 + index);
}

static void gen_lookup_table(FILE *fp, enum lookup_type specimen,
			     int n_lookups, int *pos) {
  const int lk_loc = *pos;
  int ilk, junk, n_all, n_met;
  int loc;
  struct lookup_map_entry **ll1; /* List with names and properties */
  struct lookup_list ll2; /* OTF-style list */
  size_t size;
  enum lookup_type bound1, bound9;
  struct lookup_data *data;

  sort_lookup_map(lookups, &ll1, &n_all);
  ll2.lookupCount = n_lookups;
  size = n_lookups * sizeof * ll2.lookup;
  if (ll2.lookup = malloc(size)) {
    memset(ll2.lookup, 0, size);
  } else {
    fprintf(stderr, "Memory exhausted!\n"); exit(1);
  }
  put_bine_lookup_list(fp, &ll2, &junk, pos);
  assert(lookup_offset_locn(lk_loc, n_lookups) == *pos);
  if (GSUB_1 == specimen) {
    bound1 = GSUB_1;
    bound9 = GSUB_9;
  } else if (GPOS_1 == specimen) {
    bound1 = GPOS_1;
    bound9 = GPOS_9;
  } else {
    fprintf(stderr, "Internal error: unexpected unified lookup type code %d.\n",
	    (int) specimen);
    exit(1);
  }
  n_met = 0;
  for (ilk = 0; ilk < n_all; ilk++) {
    data = ll1[ilk]->datum;
    if (data->type < bound1) continue;
    if (data->type > bound9) continue;
    if (data->lookup_index < 0) {
      fprintf(stderr, "Lookup %s is not defined.\n", ll1[ilk]->key);
      continue;
    } else if (WRITTEN == data->stored) {
      continue; /* From aborted GPOS or GDEF */
    } else if (data->lookup_index >= n_lookups) {
      fprintf(stderr, "Internal glitch!  Lookup %s has invalid look-up "
	      "index %d.  Discarded.\n", ll1[ilk]->key, data->lookup_index);
    }
    n_met++;
    data->ref.offset_ref  = lk_loc;
    data->ref.offset_loc  = lookup_offset_locn(lk_loc, data->lookup_index);
    data->ref.offset_size = 2;
    if (PENDING == data->stored) {
      int old_pos = *pos;
      store_lookup(ll1[ilk]->key, data);
#if 0
      fprintf(stdout, "Calling store_lookup() changed *pos from %d to %d.\n",
	      old_pos, *pos);
#endif
    }
  }
  if (n_met != n_lookups) {
    fprintf(stderr, "Internal glitch: Expected %d valid %s lookups but "
	    "found %d of them.", n_lookups,
	    specimen == GSUB_1 ? "GSUB" : "GPOS", n_met);
  }
  free(ll2.lookup);
  free(ll1);
}

#define provide_feature(fn) provide_feature(&features, fn, 0)
static int add_glyph_set(struct coverage_data *cov);

/* On entry to routine, current token is GSUB or GPOS. */
/* On exit from routine is token immediately after definition. */
void process_gsubpos() {
  struct script_map scripts;
  struct feature_map features;
/* Scripts designate languages, languages (per script) designate features,
   features designate lookups, and lookups designate subtables.  The lists are
   stored as maps.
*/
  const char sp4[5] = {' ', ' ', ' ', ' ', 0};
  char for_table[5];
  int n_lookups = 0;
  FILE *gsbp;
  int *pos; /* Next position to write to in the table */
  int *have_obj;
  enum lookup_type specimen; /* Lookup type with stored code 1 */

  create_script_map(&scripts);
  create_feature_map(&features);
  if (!strcmp("GSUB", current.sval)) {
    if (gsub) fclose(gsub);
    gsbp = gsub = tmpfile();
    pos = &gsub_pos;
    have_obj = &have_GSUB;
    specimen = GSUB_1;
  } else if (!strcmp("GPOS", current.sval)) {
    if (gpos) fclose(gpos);
    gsbp = gpos = tmpfile();
    pos = &gpos_pos;
    have_obj = &have_GPOS;
    specimen = GPOS_1;
  } else {
    fprintf(stderr, "Coding error: process_gsubpos called other than "
	    "for GSUB or GPOS.\n");
    exit(1);
  }
  strcpy(for_table, current.sval);
  if (!gsbp) {
    fprintf(stdout, "Could not open scratch file for %s table.\n", for_table);
    exit(1);
  }
  *have_obj = 0; /* Do not generate if parsing fails */
  advance_token();
  while (AN_IDENT == current.form && strcmp("end", current.sval)) {
    if (!strcmp("script", current.sval)) {
      struct language_map *langs;
      char scrtag[5];

      advance_token();
      if (AN_IDENT == current.form || A_STRING == current.form) {
	int taglen = strlen(current.sval);

	if (taglen > 4) {
	  show_source();
	  fprintf(stderr, "Script name '%s' is too long.\n", current.sval);
	  strncpy(scrtag, current.sval, 4);
	  scrtag[4] = 0;
	} else {
	  strcpy(scrtag, sp4);
	  strncpy(scrtag, current.sval, taglen);
	  if (!strcmp("dflt", scrtag)) {
	    show_source();
	    fprintf(stderr, "The default script is dflt, not DFLT.\n");
	    got_what(stderr, current);
	  }
	}
      } else {
	show_source();
	fprintf(stderr, "Expected a script identifier.\n");
	continue;
      }
      langs = provide_script(&scripts, scrtag, 0);
      advance_token();
      while ((AN_IDENT == current.form || A_STRING == current.form) &&
	     strcmp("end", current.sval)) {
	if (!strcmp("language", current.sval)) {
	  char lantag[5];
	  struct feature_map *feats;
	  int taglen;

	  advance_token();
	  if (AN_IDENT == current.form || A_STRING == current.form) {
	    taglen = strlen(current.sval);
	  } else {
	    taglen = 0;
	  }
	  if (0 == taglen) {
	    show_source();
	    fprintf(stderr, "Expected a language identifier.\n");
	    continue;
	  } else if (!strcmp("default", current.sval)) {
	    lantag[0] = 0; /* Force default language into first place */
	  } else if (taglen > 4) {
	    show_source();
	    fprintf(stderr, "Language name '%s' is too long.\n", 
		    current.sval);
	    strncpy(lantag, current.sval, 4);
	    lantag[4] = 0;
	  } else {
	    strcpy(lantag, sp4);
	    strncpy(lantag, current.sval, taglen);
	  }
	  if (lantag[0]) {
	    if (!strcmp("DFLT", scrtag) || !strcmp("dflt", scrtag)) {
	      show_source();
	      fprintf(stderr, "Default script should only have "
		      "a default language.\n");
	      got_what(stderr, current);
	    }
	  }
	  feats = provide_language(langs, lantag, 0);
	  advance_token();
	  while ((AN_IDENT == current.form || A_STRING == current.form) &&
		 strcmp("end", current.sval)) {
	    int taglen = strlen(current.sval);
	    char *featname, featname_buffer[5];
	    struct feature_data *featdat, *fd2;

#if 0 /* Old code, with 'char featname[5]' */
	    if (taglen > 4) {
	      show_source();
	      fprintf(stderr, "Feature name '%s' is too long.\n", current.sval);
	      strncpy(featname, current.sval, 4);
	      featname[4] = 0;
	    } else {
	      strcpy(featname, sp4);
	      strncpy(featname, current.sval, taglen);
	    }
#else
	    if (taglen >= 4) {
	      featname = current.sval;
	    } else {
	      strcpy(featname_buffer, sp4);
	      strncpy(featname_buffer, current.sval, taglen);
	      featname = featname_buffer;
	    }
#endif
	    featdat = provide_feature(featname);
	    if (!find_feature(*feats, featname, &fd2)) {
	      add_feature(feats, featname, featdat);
	    }
	    advance_token();
	  }
	  coda_end_syntax("language");
	} else {
	  show_source();
	  fprintf(stderr, "Expected keyword language.\n");
	  got_what(stderr, current);
	  advance_token();
	}
      }
      coda_end_syntax("script");
    } else if (!strcmp("feature", current.sval)) {
      struct feature_data *featdat;

      advance_token();
      { /* bound scope of featname */
	int taglen;
	char *featname, featname_buffer[5];

	if (AN_IDENT == current.form || A_STRING == current.form) {
	  taglen = strlen(current.sval);
#if 0 /* Old code, with 'char featname[5]' */
	  if (taglen > 4) {
	    show_source();
	    fprintf(stderr, "Feature name '%s' is too long.\n", current.sval);
	    strncpy(featname, current.sval, 4);
	    featname[4] = 0;
	  } else {
	    strcpy(featname, sp4);
	    strncpy(featname, current.sval, taglen);
	  }
#else
	  if (taglen >= 4) {
	    featname = current.sval;
	  } else {
	    strcpy(featname_buffer, sp4);
	    strncpy(featname_buffer, current.sval, taglen);
	    featname = featname_buffer;
	  }
#endif
	} else {
	  show_source();
	  fprintf(stderr, "Expected a feature name.\n");
	  got_what(stderr, current);
	  continue;
	}
	featdat = provide_feature(featname);
      }
      featdat->defined = 1;
      advance_token();
      if (AN_IDENT == current.form && !strcmp("includes", current.sval)) {
	int isok = 1, advance = 1;
	struct feature_data *featsrc;

	advance_token();
	if (AN_IDENT != current.form && A_STRING != current.form) {
	  show_source();
	  fprintf(stderr, "Expected a feature name.\n");
	  got_what(stderr, current);
	  isok = 0;
	  advance = 0;
	}
	if (!isok) {
	} else if (find_feature(features, current.sval, &featsrc)) {
	  if (!featsrc->defined) {
	    show_source();
	    fprintf(stderr, "Feature %s is not yet defined.\n", current.sval);
	    isok = 0;
	  }
	} else {
	  show_source();
	  fprintf(stderr, "%s is not a feature.\n", current.sval);
	  isok = 0;
	}
	if (isok) {
	  struct lookup_map_entry **fl; /* List of lookup names */
	  int ilkp, nlkp;
	  struct lookup_data *ld;
          sort_lookup_map(featsrc->lookups, &fl, &nlkp);
	  for (ilkp = 0; ilkp < nlkp; ilkp++) {
	    if (!find_lookup(featdat->lookups, fl[ilkp]->key, &ld)) {
	      add_lookup(&featdat->lookups, fl[ilkp]->key, fl[ilkp]->datum);
	    }
	  }
	  free(fl);
	}
	if (advance) advance_token();
      } else {
	while ((AN_IDENT == current.form || A_STRING == current.form) &&
	       strcmp("end", current.sval)) {
	  struct lookup_data *ld, *ld2;
	  int isnew;

	  ld = provide_lookup(&lookups, current.sval, &isnew);
	  if (isnew) {
	    ld->type = strcmp(for_table, "GSUB") ? GPOS_1 : GSUB_1;
	  } else if (!right_lookup_type(for_table, ld->type)) {
	    continue;
	  }
	  if (!find_lookup(featdat->lookups, current.sval, &ld2)) {
	    add_lookup(&featdat->lookups, current.sval, ld);
	  }
	  advance_token();
	}
	coda_end_syntax("feature");
      }
    } else if (!strcmp("lookup", current.sval)) {
      struct lookup_data *ld, ld_junk;
      int isnew;
/* Consider adding deferral option */
      advance_token();
      if (AN_IDENT == current.form || A_STRING == current.form) {
	ld = provide_lookup(&lookups, current.sval, &isnew);
	if (isnew) {
	  ld->type = strcmp(for_table, "GSUB") ? GPOS_1 : GSUB_1;
	} else if (!right_lookup_type(for_table, ld->type)) {
	  ld_junk.lookup_index = -1;
	  ld = &ld_junk;
	}
      } else {
	show_source();
	fprintf(stderr, "Expected an identifier.\n");
	got_what(stderr, current);
	continue;
      }
      if (ld->lookup_index < 0) {
	create_subtable_map(&ld->subtables);
	ld->lookup_index = n_lookups++;
	ld->own_loc = 0;
	ld->ref.offset_loc  = 0;
	ld->ref.offset_ref  = 0;
	ld->ref.offset_size = 2;
	ld->stored = PENDING;
	ld->flags = 0;
	ld->n_subtables = 0;
	ld->is_extended = 0;
      }
      advance_token();
      while ((AN_IDENT == current.form || A_STRING == current.form) &&
	     strcmp("end", current.sval)) {
	if (!strcmp("subtable", current.sval)) {
	  advance_token();
	  if (AN_IDENT == current.form || A_STRING == current.form) {
	    struct subtable_data *std;
	    int isnew;
	    if (find_subtable(subtables, current.sval, &std)) {
	      show_source();
	      fprintf(stderr, "Subtable is already declared - ignored.\n");
	    } else if (std = malloc(sizeof * std)) {
	      std->number = ld->n_subtables++;
	      std->lookup = ld;
	      std->ref.offset_loc  = 0;
	      std->ref.offset_ref  = 0;
	      std->ref.offset_size = 2;
	      std->stored = DEFERRED;
	      add_subtable(&subtables, current.sval, std);
	      add_subtable(&ld->subtables, current.sval, std);
	    } else {
	      fprintf(stderr, "Memory exhausted.\n");
	      exit(1);
	    }
	  } else {
	    show_source();
	    fprintf(stderr, "Expected an identifier or string.\n");
	    got_what(stderr, current);
	  }
	  advance_token();
	} else if (!strcmp("type", current.sval)) {
	  advance_token();
	  if (AN_IDENT == current.form && !strcmp("extension", current.sval)) {
	    ld->is_extended = 1;
	    advance_token();
	  }
	  if (AN_IDENT == current.form || A_STRING == current.form) {
	    if (strcmp(for_table, "GPOS")) { /* GSUB codes */
	      if (!strcmp("single", current.sval)) {
		ld->type = GSUB_SINGLE;
	      } else if (!strcmp("multiple", current.sval)) {
		ld->type = GSUB_MULTIPLE;
	      } else if (!strcmp("alternate", current.sval)) {
		ld->type = GSUB_ALTERNATE;
	      } else if (!strcmp("ligature", current.sval)) {
		ld->type = GSUB_LIGATURE;
	      } else if (!strcmp("context", current.sval)) {
		ld->type = GSUB_CONTEXT;
	      } else if (!strcmp("chained", current.sval)) {
		ld->type = GSUB_CHAINED;
	      } else if (!strcmp("reverse", current.sval)) {
		ld->type = GSUB_REVERSE;
	      } else {
		show_source();
		fprintf(stderr, "'%s' is not a GSUB subtable type.\n",
			current.sval);
		fprintf(stderr, "GSUB subtable types are: single, multiple, "
			"alternate, ligature, context, chained and reverse.\n");
	      }
	    } else { /* GPOS codes */
	      if (!strcmp("single", current.sval)) {
		ld->type = GPOS_SINGLE;
	      } else if (!strcmp("pair", current.sval)) {
		ld->type = GPOS_PAIR;
	      } else if (!strcmp("cursive", current.sval)) {
		ld->type = GPOS_CURSIVE;
	      } else if (!strcmp("mark2base", current.sval)) {
		ld->type = GPOS_MARK2BASE;
	      } else if (!strcmp("mark2ligature", current.sval)) {
		ld->type = GPOS_MARK2LIGATURE;
	      } else if (!strcmp("mark2mark", current.sval)) {
		ld->type = GPOS_MARK2MARK;
	      } else if (!strcmp("context", current.sval)) {
		ld->type = GPOS_CONTEXT;
	      } else if (!strcmp("chained", current.sval)) {
		ld->type = GPOS_CHAINED;
	      } else {
		show_source();
		fprintf(stderr, "'%s' is not a GPOS subtable type.\n",
			current.sval);
		fprintf(stderr, "GPOS subtable types are: single, pair, "
			"cursive, mark2base, mark2ligature, mark2mark, "
			"context, chained\n");
	      }
	    }
	  } else {
	    show_source();
	    fprintf(stderr, "Expected an identifier or string.\n");
	    got_what(stderr, current);
	  }
	  advance_token();
	} else if (!strcmp("deferred", current.sval)) {
	  ld->stored = DEFERRED;
	  advance_token();
	} else if (!strcmp("skip", current.sval)) {
	  advance_token();
	  if (AN_IDENT == current.form || A_STRING == current.form) {
	    if (!strcmp("base", current.sval)) {
	      ld->flags |= 2;
	    } else if (!strcmp("ligature", current.sval)) {
	      ld->flags |= 4;
	    } else if (!strcmp("mark", current.sval)) {
	      ld->flags |= 8;
	    } else {
	      show_source();
	      fprintf(stderr, "Expected 'base', 'ligature' or 'mark'.\n");
	    }
	  } else {
	    show_source();
	    fprintf(stderr, "Expected an identifier or string.\n");
	    got_what(stderr, current);
	  }
	  advance_token();
	} else if (!strcmp("setbit", current.sval)) {
/* Deprecated bit flag interface */
	  advance_token();
	  if (AN_INT == current.form &&
	      0 <= current.ival && current.ival <= 15) {
	    ld->flags |= (1 << current.ival);
	  } else {
	    show_source();
	    fprintf(stderr, "Not a valid bit number.\n");
	    got_what(stderr, current);
	  }
	  advance_token();
	} else if (!strcmp("skipexcept", current.sval)) {
	  advance_token();
	  if (AN_IDENT == current.form || A_STRING == current.form) {
	    struct glyph_group *gg;
	    if (lookup_in_map(glyph_names, current.sval, (void**)&gg)) {
	      switch (gg->grouping) {
	      case SOLO:
		show_source();
		fprintf(stderr, "Must specify a mark attachment class or a "
			"coverage to skip, but %s is a single glyph.\n",
			current.sval);
		break;
	      case COVERAGE:
                ld->mark_set = add_glyph_set(gg->x.coverage_data);
                ld->flags |= 16;
		break;
	      case CLASS:
		if (gg->x.class_data->classification == attachment_classes) {
		  ld->flags &= ~0xFF00;
		  ld->flags |= (gg->x.class_data->number << 8);
		} else {
		  show_source();
		  fprintf(stderr, "Class %s is not a mark attachment class.\n",
			  current.sval);
		}
		break;
	      default:
		fprintf(stderr, "Internal error at line %d of %s.\n",
			__LINE__, __FILE__);
		exit(1);
	      }
	    } else {
	      show_source();
	      fprintf(stderr, "%s does not define a glyph group.\n",
		      current.sval);
	    }
	  } else if (AN_INT == current.form &&
		     0 <= current.ival && current.ival <= 0xFF) {
	    ld->flags &= ~0xFF00;
	    ld->flags |= (current.ival << 8);
	  } else {
	    show_source();
	    fprintf(stderr, "Not a valid mark attachment class.\n");
	    got_what(stderr, current);
	  }
	  advance_token();
	} else if (!strcmp("anchor", current.sval)) {
	  advance_token();
	  if (AN_IDENT == current.form || A_STRING == current.form) {
/* The bit sttings must be checked - they look odd. */
	    if (!strcmp("first", current.sval)) {
	      ld->flags &=~ 1;
	    } else if (!strcmp("last", current.sval)) {
	      ld->flags |= 1;
	    } else {
	      show_source();
	      fprintf(stderr, "Expected 'first' or 'last'.\n");
	    }
	  } else {
	    show_source();
	    fprintf(stderr, "Expected an identifier or string.\n");
	    got_what(stderr, current);
	  }
	  advance_token();
	} else {
          show_source();
	  fprintf(stdout, "Subkeyword %s not recognised.\n", current.sval);
	  advance_token();
	}
      }
      coda_end_syntax("lookup");
    } else {
      unwritten;
    }
  }

/* Start GSUB/GPOS table */
  {
    struct gpos_header hdr;
    int junk, loc;
    struct feature_map_entry **fl1;
    int ift, n_features;

    *pos = 0;
    /* Version 1.0 of the GSUB and GPOS headers are identical. */
    memset(&hdr, 0, sizeof hdr);
    hdr.version = 1 << 16;
    put_bine_gpos_header(gsbp, &hdr, &junk, pos);

/* Sort the features */
    sort_feature_map(features, &fl1, &n_features);
    for (ift = 0; ift < n_features; ift++) {
      struct feature_data *featdat;
      featdat = fl1[ift]->datum;
      featdat->number = ift;
#if 0
      fprintf(stdout, "Assigned number %d to feature %s in data at %p.\n",
	      ift, fl1[ift]->key, fl1[ift]->datum);
#endif
      if (!featdat->defined) {
	fprintf(stderr, "Feature %s is not defined - "
		"assumed to have no lookups.\n", fl1[ift]->key);
      }
    }
/* Generate the script table */    
    hdr.scriptList = *pos;
    gen_script_table(gsbp, scripts, pos);

/* Generate the feature table */
    gsbp_seek(gsbp, *pos);
    hdr.featureList = *pos;
    gen_feature_table(gsbp, fl1, n_features, pos);
/* Generate the lookup table - but not its subtables */
    hdr.lookupList = *pos;
    gen_lookup_table(gsbp, specimen, n_lookups, pos);
/* Update the header. */
    gsbp_seek(gsbp, loc = 0);
    put_bine_gpos_header(gsbp, &hdr, &junk, &loc);
    gsbp_seek(gsbp, *pos);
  }
  *have_obj = 1;
/* Clean up local maps */
#if 0
  unwritten;
#endif
  coda_end_syntax(for_table);
}

static void skip_to_end(const char *keyword) {
  while (current.form != END_OF_FILE &&
	 (current.form != AN_IDENT || previous.form != AN_IDENT ||
	  strcmp(current.sval, keyword) || strcmp(previous.sval, "end"))) {
    advance_token();
  }
  advance_token();
}

no_providing(gclass, struct class_data)
extend_encapsulate_map(gclass, struct class_data)

void locate_class(const char *name, struct classification_data *cd,
		  struct class_data **cldat) {
  struct glyph_group *gg;

  if (find_gclass(cd->classes, name, cldat)) {
    /* Normal case, nothing to do */
  } else if (lookup_in_map(glyph_names, name, (void**)&gg)) {
    show_source();
    fprintf(stderr, "Glyph group %s is not in this classification.\n",
	    current.sval);
    *cldat = 0;
  } else if ((*cldat = malloc(sizeof * * cldat)) &&
	     (gg = malloc(sizeof * gg))) { 
    (*cldat)->members = create_int_set();
    (*cldat)->number = cd->n_classes++;
    (*cldat)->classification = cd;
    add_gclass(&cd->classes, name, *cldat);
    gg->x.class_data = *cldat;
    gg->grouping = CLASS;
    insert_in_map(&glyph_names, name, gg);
  } else {
    fprintf(stderr, "Memory exhausted.\n");
    exit(1);
  }
}

int check_ident() {
  if (AN_IDENT == current.form) {
    return 1;
  } else {
    show_source();
    fprintf(stderr, "Expected an identifier.\n");
    got_what(stderr, current);
    advance_token();
    return 0;
  }
}

no_providing(gclassification, struct classification_data)
extend_encapsulate_map(gclassification, struct classification_data)

/* Note that a classification of this name must not already exist - if it
   does, the program will exit with an error message. */
struct classification_data *create_classification(const char *name) {
  struct classification_data *cd;

  if (cd = malloc(sizeof * cd)) {
    create_gclass_map(&cd->classes);
    cd->n_classes = 0;
    cd->lox = no_locations;
    add_gclassification(&classifications, name, cd);
  } else {
    fprintf(stderr, "Memory exhausted.\n");
    exit(1);
  }
  return cd;
}

void process_classification() {
  struct classification_data *cd;

  if (AN_IDENT != current.form || strcmp("classification", current.sval)) {
    fprintf(stderr, "Coding error: process_classification called other than "
	    "for classification.\n");
    exit(1);
  }
  advance_token();
  if (AN_IDENT != current.form && A_STRING != current.form) {
    show_source();
    fprintf(stderr, "Expected an identifier.\n");
    got_what(stderr, current);
    return;
  }
  if (find_gclassification(classifications, current.sval, &cd)) {
    show_source();
    fprintf(stderr, "Classification %s is already defined.\n", current.sval); 
    skip_to_end("classification");
    return;
  }
/* Note that there is parallel code in function get_void_classification() */
  cd = create_classification(current.sval);
  advance_token();
  while (AN_IDENT == current.form && strcmp(current.sval, "end")) {
    struct class_data *cldat;
    struct glyph_group *gg1, *gg2;
    int igl1, igl2;
    if (!strcmp(current.sval, "class")) {
      advance_token();
      if (!check_ident()) continue;
      locate_class(current.sval, cd, &cldat);
      advance_token();
      if (!check_ident()) continue;
      if (lookup_in_map(glyph_names, current.sval, (void**)&gg1)) {
	switch (gg1->grouping) {
	case SOLO:
	  igl1 = gg1->x.number;
	  (void) add_range(igl1, igl1, &cldat->members);
	  break;
	case COVERAGE:
	  igl1 = -1;
	  (void) add_set(gg1->x.coverage_data->members, &cldat->members);
	  break;
	case CLASS:
	  igl1 = -1;
	  (void) add_set(gg1->x.class_data->members, &cldat->members);
	  break;
	default:
	  fprintf(stderr, "Internal error at line %d of %s.\n",
		  __LINE__, __FILE__);
	  exit(1);
	}
      } else {
	show_source();
	fprintf(stderr, "%s does not identify a group of glyphs.\n",
		current.sval);
	igl1 = -1;
	gg1 = 0;
      }
      advance_token();
      if (!check_ident()) continue;
      if (!strcmp(current.sval, "to")) {
	if (gg1 && igl1 < 0) {
	  show_source();
	  fprintf(stderr, "First identifier for a range must identify "
		  "a single glyph.\n");
	}
	advance_token();
	if (!check_ident()) continue;
	if (lookup_in_map(glyph_names, current.sval, (void**)&gg2)) {
	  if (SOLO == gg2->grouping) {
	    igl2 = gg2->x.number;
	    if (igl1 < 0) {
	      /* Do nothing */
	    } else if (igl2 < igl1) {
              show_source();
	      fprintf(stderr, "Identifiers defining a range must be in "
		      "increasing order.\n");
	    } else {
	      (void) add_range(igl1, igl2, &cldat->members);
	    }
	  } else {
            show_source();
	    fprintf(stderr, "Second identifier for a range must identify "
		    "a single glyph.\n");
	  }
	} else {
	  show_source();
	  fprintf(stderr, "%s does not identify a group of glyphs.\n",
		  current.sval);
	}
	advance_token();
      }
    } else if (!strcmp(current.sval, "empty")) {
      advance_token();
      if (!check_ident()) continue;
      locate_class(current.sval, cd, &cldat);
      advance_token();
    } else if (!strcmp(current.sval, "remove")) {
      struct gclass_map_entry **cl = 0;
      int ic, nc;

      advance_token();
      if (!check_ident()) continue;
      sort_gclass_map(cd->classes, &cl, &nc);
      if (lookup_in_map(glyph_names, current.sval, (void**)&gg1)) {
	switch (gg1->grouping) {
	case SOLO:
	  igl1 = gg1->x.number;
	  for (ic = 0; ic < nc; ic++) {
	    cldat = (struct class_data*) cl[ic]->datum;
	    (void) sub_range(igl1, igl1, &cldat->members);
	  }
	  break;
	case COVERAGE:
	  igl1 = -1;
	  for (ic = 0; ic < nc; ic++) {
	    cldat = (struct class_data*) cl[ic]->datum;
	    (void) sub_set(gg1->x.coverage_data->members, &cldat->members);
	  }
	  break;
	case CLASS:
	  igl1 = -1;
	  for (ic = 0; ic < nc; ic++) {
	    cldat = (struct class_data*) cl[ic]->datum;
	    (void) sub_set(gg1->x.class_data->members, &cldat->members);
	  }
	  break;
	default:
	  fprintf(stderr, "Internal error at line %d of %s.\n",
		  __LINE__, __FILE__);
	  exit(1);
	}
      } else {
	show_source();
	fprintf(stderr, "%s does not identify a group of glyphs.\n",
		current.sval);
	igl1 = -1;
	gg1 = 0;
      }
      advance_token();
      if (!check_ident()) {
	free(cl);
	continue;
      }
      if (!strcmp(current.sval, "to")) {
	if (gg1 && igl1 < 0) {
	  show_source();
	  fprintf(stderr, "First identifier for a range must identify "
		  "a single glyph.\n");
	}
	advance_token();
	if (!check_ident()) {
	  free(cl);
	  continue;
	}
	if (lookup_in_map(glyph_names, current.sval, (void**)&gg2)) {
	  if (SOLO == gg2->grouping) {
	    igl2 = gg2->x.number;
	    if (igl1 < 0) {
	      /* Do nothing */
	    } else if (igl2 < igl1) {
	      fprintf(stderr, "Identifiers defining a range must be in "
		      "increasing order.\n");
	    } else {
	      for (ic = 0; ic < nc; ic++) {
		cldat = (struct class_data*) cl[ic]->datum;
		(void) sub_range(igl1, igl2, &cldat->members);
	      }
	    }
	  } else {
	    fprintf(stderr, "Second identifier for a range must identify "
		    "a single glyph.\n");
	  }
	} else {
	  fprintf(stderr, "%s does not identify a group of glyphs.\n",
		  current.sval);
	}
	free(cl);
	advance_token();
      }
    } else {
      show_source();
      fprintf(stderr, "Keyword %s is not recognised.\n", current.sval);
      advance_token();
    }
  }
  coda_end_syntax("classification");
}

void process_coverage() {
  struct glyph_group *coverage_group, *gg1, *gg2;
  struct coverage_data *cd;
  int gl1, gl2;

  if (AN_IDENT != current.form || strcmp("coverage", current.sval)) {
    fprintf(stderr, "Coding error: process_coverage called other than "
	    "for 'coverage'.\n");
    exit(1);
  }
  advance_token();
  if (!check_ident()) return;
  advance_token();
  if (!check_ident()) return;
  if (strcmp("is", current.sval) && strcmp("also", current.sval)) {
    show_source();
    fprintf(stderr, "Expected 'is' or 'also'.\n");
    got_what(stderr, current);
    return;
  }
  if (lookup_in_map(glyph_names, previous.sval, (void**)&coverage_group)) {
    if (!strcmp("is", current.sval)) {
      show_source();
      fprintf(stderr, "Glyph group %s is already defined.\n", previous.sval);
      cd = 0;
    } else if (COVERAGE == coverage_group->grouping) {
      cd = coverage_group->x.coverage_data;
    } else {
      show_source();
      fprintf(stderr, "Glyph group %s is not a coverage.\n", previous.sval);
      cd = 0;
    }
  } else {
    if (!strcmp("is", current.sval)) {
      cd = malloc(sizeof * cd);
      coverage_group = malloc(sizeof * coverage_group);
      if (cd && coverage_group) {
	cd->members = create_int_set();
	cd->lox = no_locations;
	coverage_group->grouping = COVERAGE;
	coverage_group->x.coverage_data = cd;
	insert_in_map(&glyph_names, previous.sval, coverage_group);
      } else {
	fprintf(stderr, "Memory exhausted.\n");
	exit(1);
      }
    } else {
      show_source();
      fprintf(stderr, "Glyph group %s does not exist.\n", previous.sval);
      cd = 0;
    }
  }
  advance_token();
  if (!check_ident()) return;
  gl1 = -1;
  if (lookup_in_map(glyph_names, current.sval, (void**)&gg1)) {
    switch (gg1->grouping) {
    case SOLO:
      gl1 = gg1->x.number;
      if (cd) (void) add_range(gl1, gl1, &cd->members);
      break;
    case COVERAGE:
      if (cd) (void) add_set(gg1->x.coverage_data->members, &cd->members);
      break;
    case CLASS:
      if (cd) (void) add_set(gg1->x.class_data->members, &cd->members);
      break;
    default:
      fprintf(stderr, "Internal error - unhandled branch at line %d of "
	      "file %s.\n", __LINE__, __FILE__);
      exit(1);
    }
  } else {
    gg1 = 0;
    gl1 = -1;
    show_source();
    fprintf(stderr, "'%s' does not identify a glyph group.\n", current.sval);
  }
  advance_token();
  if (AN_IDENT == current.form && !strcmp("to", current.sval)) {
    if (gg1 && gl1 < 0) {
      show_source();
      fprintf(stderr, "'%s' does not identify just a single glyph.\n",
	      previous.sval);
    } 
    advance_token();
    if (!check_ident()) return;
    gl2 = -1;
    if (lookup_in_map(glyph_names, current.sval, (void**)&gg2)) {
      if (SOLO == gg2->grouping) {
	gl2 = gg2->x.number;
	if (gl1 < 0) {
	  /* Do nothing */
	} else if (gl1 <= gl2) {
	  (void) add_range(gl1, gl2, &cd->members);
	} else {
	  show_source();
	  fprintf(stderr, "Error: Glyph identifiers are out of order.\n");
	}
      } else {
	show_source();
	fprintf(stderr, "'%s' does not identify just a single glyph.\n",
		current.sval);
      }
    } else {
      show_source();
      fprintf(stderr, "'%s' does not identify any glyphs.\n", current.sval);
    }
    advance_token();
  }
}

void store_coverage(const char *name, /* For error messages */
		    enum lookup_type table, /* Table in which to store */
		    struct coverage_data * coverage) {
  FILE *fp;
  int *posn, covpos, loc;
  int nwrit, nunwrit, iloc;
  enum lookup_type bound1, bound9, table_code;
  int junk;

  if (GSUB_1 <= table && table <= GSUB_9) {
    posn = &gsub_pos;
    fp = gsub;
    bound1 = GSUB_1;
    bound9 = GSUB_9;
  } else if (GPOS_1 <= table && table <= GPOS_9) {
    posn = &gpos_pos;
    fp = gpos;
    bound1 = GPOS_1;
    bound9 = GPOS_9;
  } else if (GDEF_TGT <= table && table <= GDEF_TGT) {
    posn = out_pos;
    fp = outfp;
    bound1 = GDEF_TGT;
    bound9 = GDEF_TGT;
  } else {
    fprintf(stderr, "Internal error in store_coverage(): "
	    "unexpected unified lookup type %d encountered for coverage %s.\n",
	    table, name);
    exit(1);
  }
  covpos = *posn;
  if (!coverage->lox.loc) return; /* Nothing to do */
  nwrit = 0;
  nunwrit = 0;
  for (iloc = 0; iloc < coverage->lox.nloc; iloc++) {
    table_code = coverage->lox.loc[iloc].table;
    if (bound1 <= table_code && table_code <= bound9) {
      int offset;

      offset = covpos - coverage->lox.loc[iloc].offset_ref;
      if (4 == coverage->lox.loc[iloc].offset_size) {
	struct UL_item the_offset;
	gsbp_seek(fp, loc = coverage->lox.loc[iloc].offset_loc);
	the_offset.val = offset;
	put_bine_UL_item(fp, &the_offset, &junk, &loc);
	nwrit++;
      } else if (offset <= 0xFFFF) {
	struct UW_item the_offset;
	gsbp_seek(fp, loc = coverage->lox.loc[iloc].offset_loc);
	the_offset.val = offset;
	put_bine_UW_item(fp, &the_offset, &junk, &loc);
	nwrit++;
      } else {
	fprintf(stderr, "Coverage %s stored too late to be reached by "
		 "16-bit offset.\n", name);
      }
    } else {
      coverage->lox.loc[nunwrit++] = coverage->lox.loc[iloc];
    }
  }
  coverage->lox.nloc = nunwrit;
  if (nwrit) {
    int nrge = 0, npt = 0;
    struct int_range *irge;
    struct coverage_table ct;
    int fmt1;

    gsbp_seek(fp, *posn);
    for (irge = coverage->members; irge; irge = irge->next) {
      nrge++;
      npt += irge->end - irge->start + 1;
    }
    memset(&ct, 0, sizeof ct);
    fmt1 = npt <= 3 * nrge;
    if (fmt1) {
      unsigned short int *uptr, igl;

      ct.coverageFormat = 1;
      ct.glyphCount = npt;
      if (!(ct.glyphArray = malloc(npt * sizeof * ct.glyphArray))) {
	fprintf(stderr, "Memory exhausted!\n");
	exit(1);
      }
      uptr = ct.glyphArray;
      for (irge = coverage->members; irge; irge = irge->next) {
	for (igl = irge->start; igl <= irge->end; igl++) *uptr++ = igl;
      }
    } else {
      int npt2 = 0;
      struct range_record *frge, **nexptr;

      ct.coverageFormat = 2;
      ct.rangeCount = nrge;
      if (!(ct.rangeRecord = malloc(nrge * sizeof * ct.rangeRecord))) {
	fprintf(stderr, "Memory exhausted.\n");
	exit(1);
      }
      nexptr = &ct.rangeRecord;
      for (irge = coverage->members, frge = ct.rangeRecord;
	   irge; irge = irge->next, frge++) {
	*nexptr = frge;
	frge->start = irge->start;
	frge->end = irge->end;
	frge->startCoverageIndex = npt2;
	nexptr = &frge->_next;
	*nexptr = 0;
	npt2 += irge->end - irge->start + 1;
      }
      assert(npt == npt2);
    }
    put_bine_coverage_table(fp, &ct, &junk, posn);
    free(ct.glyphArray);
    free(ct.rangeRecord);
  }
}

void store_all_used_coverages() {
  struct map_entry **cl;
  struct glyph_group *gg;
  int i, n;

  sort_map(glyph_names, &cl, &n);
  for (i = 0; i < n; i++) {
    gg = cl[i]->datum;
    if (COVERAGE != gg->grouping) continue;
    store_coverage(cl[i]->key, GSUB_1, gg->x.coverage_data);
    store_coverage(cl[i]->key, GPOS_1, gg->x.coverage_data);
  }
  free(cl);
}

void store_classification(const char *name, /* For error messages */
		    enum lookup_type table, /* Table in which to store */
		    struct classification_data * cldat) {
  FILE *fp;
  int *posn, clpos, loc;
  int nwrit, nunwrit, iloc;
  enum lookup_type bound1, bound9, table_code;
  int junk;

  if (GSUB_1 <= table && table <= GSUB_9) {
    posn = &gsub_pos;
    fp = gsub;
    bound1 = GSUB_1;
    bound9 = GSUB_9;
  } else if (GPOS_1 <= table && table <= GPOS_9) {
    posn = &gpos_pos;
    fp = gpos;
    bound1 = GPOS_1;
    bound9 = GPOS_9;
  } else if (GDEF_TGT == table) {
    posn = out_pos;
    fp = outfp;
    bound1 = GDEF_TGT;
    bound9 = GDEF_TGT;
  } else {
    fprintf(stderr, "Internal error in store_classification(): unexpected "
	    "unified lookup type %d encountered for classification %s.\n",
	    table, name);
    exit(1);
  }
  clpos = *posn;
  if (!cldat->lox.nloc) return; /* Nothing to do */
  nwrit = 0;
  nunwrit = 0;
  for (iloc = 0; iloc < cldat->lox.nloc; iloc++) {
    table_code = cldat->lox.loc[iloc].table;
    if (bound1 <= table_code && table_code <= bound9) {
      int offset;

      offset = clpos - cldat->lox.loc[iloc].offset_ref;
      if (4 == cldat->lox.loc[iloc].offset_size) {
	unwritten; /* Unreachable */
      } else if (offset <= 0xFFFF) {
	struct UW_item the_offset;
	gsbp_seek(fp, loc = cldat->lox.loc[iloc].offset_loc);
	the_offset.val = offset;
	put_bine_UW_item(fp, &the_offset, &junk, &loc);
	nwrit++;
      } else {
	fprintf(stderr, "Classification %s stored too late to be reached by "
		 "16-bit offset.\n", name);
      }
    } else {
      cldat->lox.loc[nunwrit++] = cldat->lox.loc[iloc];
    }
  }
  cldat->lox.nloc = nunwrit;
  if (nwrit) {
    int nrge = 0, npt = 0;
    struct int_range *irge, **nextrge;
    struct class_definition_table cdt;
    int fmt1;
    int_set allglyphs;
    struct gclass_map_entry **cll;
    int ic, nc2, icc;
    unsigned short int *uptr;
    struct class_range_record *frge;
    int prev_class, prev_glyph;

    gsbp_seek(fp, *posn);
    allglyphs = create_int_set();
    sort_gclass_map(cldat->classes, &cll, &nc2);
    if (nc2 != cldat->n_classes && table != GDEF_TGT) {
      unwritten;
    }
    if (!(nextrge = malloc(nc2 * sizeof * nextrge))) {
      fprintf(stderr, "Memory exhausted.\n");
      exit(1);
    }
    for (ic = 0; ic < nc2; ic++) {
      struct class_data *cld = cll[ic]->datum;
      nextrge[ic] = cld->members;
      (void) add_set(cld->members, &allglyphs);
      for (irge = cld->members; irge; irge = irge->next) {
	nrge++;
	npt += irge->end - irge->start + 1;
      }
    }
    memset(&cdt, 0, sizeof cdt);
    fmt1 = ((npt + 1) <= 3 * nrge) && allglyphs && !allglyphs->next;
    if (fmt1) {
      cdt.classFormat = 1;
      cdt.startGlyph = allglyphs->start;
      cdt.glyphCount = npt;
      if (!(cdt.classValueArray = malloc(npt * sizeof * cdt.classValueArray))) {
	fprintf(stderr, "Memory exhausted!\n");
	exit(1);
      }
      uptr = cdt.classValueArray;
    } else {
      cdt.classFormat = 2;
      cdt.classRangeCount = nrge;
      if (cdt.classRangeRecord = malloc(nrge * sizeof * cdt.classRangeRecord)) {
	int iel;
	for (iel = 0; iel < nrge; iel++) {
	  cdt.classRangeRecord[iel]._next = cdt.classRangeRecord + iel + 1;
	}
	cdt.classRangeRecord[nrge-1]._next = 0;
	frge = cdt.classRangeRecord;
      } else {
	fprintf(stderr, "Memory exhausted.\n");
	exit(1);
      }
    }
    prev_glyph = -1; /* Smaller than all glyph numbers */
    prev_class = 0;  /* Value should be irrrelevant */
    while (1) {
      int classno;

      irge = 0;
      for (ic = 0; ic < nc2; ic++) {
	if (nextrge[ic]) {
	  if (irge && nextrge[ic]->start < irge->start || !irge) {
	    irge = nextrge[ic];
	    icc = ic;
	  }
	}
      }
      if (!irge) break;
      nextrge[icc] = nextrge[icc]->next;
      {
	struct class_data *cld = cll[icc]->datum;
	classno = cld->number;
      }
      if (irge->start <= prev_glyph) {
	show_source();
	fprintf(stderr, "Classification %s is corrupt.  Glyph %d (%s) "
		"is in both class %s and class %s.\n",
		name, irge->start, get_glyph_name(irge->start),
		cll[prev_class]->key, cll[icc]->key);
      }
      if (fmt1) {
	int igl;
	for (igl = irge->start; igl <= irge->end; igl++) *uptr++ = classno;
      } else {
	frge->start = irge->start;
	frge->end   = irge->end;
	frge->class = classno;
	frge = frge + 1;
      }
      prev_glyph = irge->end;
      prev_class = icc;
    }
    put_bine_class_definition_table(fp, &cdt, &junk, posn);
    free(cdt.classValueArray);
    free(cdt.classRangeRecord);
    free(cll);
    free(allglyphs);
  }
}

const char *get_lookup_name(const struct lookup_data *ld) {
  struct lookup_map_entry **ll;
  int i, n;
  const char *res = 0;

  sort_lookup_map(lookups, &ll, &n);
/* Could use binary search to speed code up. */
  for (i = 0; i < n; i++) {
    if (ld == ll[i]->datum) {
      res = ll[i]->key;
      break;
    }
  }
  free(ll);
  return res;
}
  
const char *get_subtable_name(const struct subtable_data *ld) {
  struct subtable_map_entry **ll;
  int i, n;
  const char *res = 0;

  sort_subtable_map(subtables, &ll, &n);
/* Could use binary search to speed code up. */
  for (i = 0; i < n; i++) {
    if (ld == ll[i]->datum) {
      res = ll[i]->key;
      break;
    }
  }
  free(ll);
  return res;
}

void store_pointer(struct subtable_data *std) {
  if (!std) return;
  switch (std->stored) {
  case PENDING:
    unwritten; /* Unreachable, I think */
    break;
  case DEFERRED:
    {
      struct lookup_data *ld = std->lookup;
      int *posn, stloc, loc, adjust;
      FILE *fp;
      struct extension_pos_table table;
      int offset, junk = 0;

      if (!ld->is_extended) return;
      /* Reference forces storage of lookup */
      store_lookup(get_lookup_name(ld), ld);
      if (GSUB_1 <= ld->type && ld->type <= GSUB_9) {
	posn = &gsub_pos;
	fp = gsub;
	adjust   = GSUB_ADJ;
      } else if (GPOS_1 <= ld->type && ld->type <= GPOS_9) {
	posn = &gpos_pos;
	fp = gpos;
	adjust   = GPOS_ADJ;
      } else {
	show_source();
	fprintf(stderr, "Internal error in store_pointer(): unexpected "
		"unified lookup type %d encountered for subtable %s.\n",
		ld->type, get_subtable_name(std));
	exit(1);
      }
      if (3 & *posn) {
	gsbp_seek(fp, *posn);
	align_file(fp, posn, 4); /* Helpful, but perhaps not needed */
      }
      offset = *posn - std->ref.offset_ref;
      if (4 == std->ref.offset_size) {
	unwritten; /* Assumed to be inaccessible */
      } else if (offset <= 0xFFFF) {
	struct UW_item twobytes;

	gsbp_seek(fp, loc = std->ref.offset_loc);
	twobytes.val = offset;
	put_bine_UW_item(fp, &twobytes, &junk, &loc);
      } else {
	fprintf(stderr, "Pointer to subtable %s is to be stored too far "
		"from lookup.\n", get_subtable_name(std));
      }
      std->ref.offset_ref  = *posn;
      std->ref.offset_loc  = *posn + 4;
      std->ref.offset_size = 4;
      table.posFormat = 1;
      table.extensionLookupType = ld->type - adjust;
      table.extensionOffset = 0;
      gsbp_seek(fp, *posn);
      put_bine_extension_pos_table(fp, &table, &junk, posn);
      std->stored = EXTENDED;
    }
    break;
  case WRITTEN:
  case EXTENDED:
    break;
  default:
    unwritten;
  }
}

void store_something() {
  if (AN_IDENT != current.form || strcmp("store", current.sval)) {
    fprintf(stderr, "coding error: store_something() called other than "
	    "for 'store'.\n");
    exit(1);
  }
  advance_token();
  if (!check_ident()) return;
  if (strcmp("lookup", current.sval) &&
      strcmp("pointer", current.sval) &&
      strcmp("classification", current.sval) &&
      strcmp("coverage", current.sval)) {
    show_source();
    fprintf(stderr, "Expected 'lookup', 'pointer', 'classification' or "
	    "'coverage'\n.");
    got_what(stderr, current);
    return;
  } else {
    void *ptrval;

    advance_token();
    if (!check_ident()) return;
    if (!strcmp("lookup", previous.sval)) {
      struct lookup_data *ld;

      if (find_lookup(lookups, current.sval, &ld)) {
	store_lookup(current.sval, ld);
      } else {
	show_source();
	fprintf(stderr, "Lookup %s is not declared.\n", current.sval);
      }
    } else if (!strcmp("classification", previous.sval)) {
      struct classification_data *cd;
      if (find_gclassification(classifications, current.sval, &cd)) {

	store_classification(current.sval, GSUB_1, cd);
	store_classification(current.sval, GPOS_1, cd);
      } else {
	show_source();
	fprintf(stderr, "Classification %s does not exist.\n", current.sval);
      }
    } else if (!strcmp("coverage", previous.sval)) {
      struct glyph_group *gg;

      if (lookup_in_map(glyph_names, current.sval, &ptrval)) {
	gg = ptrval;
	if (COVERAGE == gg->grouping) {
	  store_coverage(current.sval, GSUB_1, gg->x.coverage_data);
	  store_coverage(current.sval, GPOS_1, gg->x.coverage_data);
	} else {
	  show_source();
	  fprintf(stderr, "Glyph group %s is not a coverage.\n", current.sval);
	}
      } else {
	show_source();
	fprintf(stderr, "Coverage %s does not exist.\n", current.sval);
      }
    } else if (!strcmp("pointer", previous.sval)) {
      struct subtable_data *std;

      if (find_subtable(subtables, current.sval, &std)) {
	if (!std->lookup->is_extended) {
          show_source();
	  fprintf(stderr, "Subtable %s is not part of an extended lookup.\n",
		  current.sval);
	} else {
	  switch (std->stored) {
	  case PENDING:
	    unwritten; /* Unreachable, I think */
	    break;
	  case DEFERRED:
	    store_pointer(std);
	    break;
	  case WRITTEN:
	    show_source();
	    fprintf(stderr, "Subtable %s has already been stored.\n",
		    current.sval);
	    break;
	  case EXTENDED:
	    show_source();
	    fprintf(stderr, "Pointer to subtable %s has already been stored.\n",
		    current.sval);
	    break;
	  default:
	    unwritten;
	  }
	}
      } else {
	fprintf(stderr, "Subtable %s has not been declared.\n", current.sval);
      }
    } else {
      unwritten;
    }
    advance_token();
  }
}

void store_all_used_classifications() {
  struct gclassification_map_entry **cl;
  struct classification_data *cd;
  int i, n;

  sort_gclassification_map(classifications, &cl, &n);
  for (i = 0; i < n; i++) {
    cd = cl[i]->datum;
    store_classification(cl[i]->key, GSUB_1, cd);
    store_classification(cl[i]->key, GPOS_1, cd);
  }
}

void add_location(enum lookup_type the_type,
		  unsigned int ref_pos, unsigned int store_pos,
		  unsigned int size,
		  struct locations *lox) {
  if (lox->nloc >= lox->mloc) {
    lox->mloc += 5;
    lox->loc = realloc(lox->loc, lox->mloc * sizeof * lox->loc);
    if (!lox->loc) {
      fprintf(stderr, "Memory exhausted.\n");
      exit(1);
    }
  }
  lox->loc[lox->nloc].offset_loc  = store_pos;
  lox->loc[lox->nloc].offset_ref  = ref_pos;
  lox->loc[lox->nloc].offset_size = size;
  lox->loc[lox->nloc].table = the_type;
  lox->nloc++;
}

static struct glyph_group *get_single(const char *name) {
  struct glyph_group *retval;
  if (lookup_in_map(glyph_names, name, (void**)&retval)) {
    if (SOLO == retval->grouping) return retval;
    show_source();
    fprintf(stderr, "'%s' denotes a group of glyphs, not a single glyph.\n",
	    name);
  } else {
    show_source();
    fprintf(stderr, "'%s' does not denote any glyphs.\n", name);
  }
  skip_to_end("lookup");
  return 0;
}
  
static struct glyph_group *context_ctype;
static struct map deferral;
#define MAX_P_LEN 20
/* Returns 1 if glyph/coverage/class string is legitimate */
/* Globals are: context_ctype, deferrals */
int chain_glyph_string(struct glyph_group ***s, int *ms, int *ns,
		       struct classification_data **classn) {
  char buffer[MAX_P_LEN + 1]; /* For formatted pointers */

  while (AN_IDENT == current.form &&
	 strcmp("end", current.sval) &&
	 strcmp("context", current.sval)) {
    struct glyph_group *gg;
    if (lookup_in_map(glyph_names, current.sval, (void**)&gg)) {
      if (!context_ctype) {
	context_ctype = gg;
      } else if (context_ctype->grouping != gg->grouping) {
	const char *tystr;
	show_source();
	fprintf(stderr, "'%s' denotes ", current.sval);
	switch (gg->grouping) {
	case SOLO:
	  tystr = "a glyph";
	  break;
	case COVERAGE:
	  tystr = "a coverage";
	  break;
	case CLASS:
	  tystr = "a class";
	  break;
	default:
	  tystr = "?*?*?";
	}
	fprintf(stderr, "%s, but this subtable is for ", tystr);
	switch (context_ctype->grouping) {
	case SOLO:
	  tystr = "single glyphs";
	  break;
	case COVERAGE:
	  tystr = "coverages";
	  break;
	case CLASS:
	  tystr = "classes";
	  break;
	default:
	  tystr = "?*?*?.";
	}
	fprintf(stderr, "%s.\n", tystr);
	skip_to_end("lookup");
	return 0;
      }
    } else {
      show_source();
      fprintf(stderr, "'%s' does not identify any glyphs.\n", current.sval);
      skip_to_end("lookup");
      return 0;
    }
    if (*ns >= *ms) {
      (*ms)++;
      if (!(*s = realloc(*s, *ms * sizeof * *s))) {
	fprintf(stderr, "Memory exhausted.\n");
	exit(1);
      }
    }
    (*s)[(*ns)++] = gg;
    if (context_ctype->grouping != SOLO) {
      void *ptrval;
      int uselen = 0;
	    
      switch (gg->grouping) {
      case COVERAGE:
	uselen = sprintf(buffer, "%p", gg->x.coverage_data);
	break;
      case CLASS:
	if (!*classn) {
	  *classn = gg->x.class_data->classification;
	} else if (*classn != gg->x.class_data->classification) {
	  show_source();
	  fprintf(stderr, "Inconsistent choice of classification "
		  "for this position.\n");
	  skip_to_end("lookup");
	  return 0;
	}
	uselen = sprintf(buffer, "%p", gg->x.class_data->classification);
	break;
      }
      if (uselen > MAX_P_LEN) {
	fprintf(stderr, "Formatted addresses exceed %d characters.\n",
		MAX_P_LEN);
	exit(1);
      }
      if (!lookup_in_map(deferral, buffer, &ptrval)) {
	insert_in_map(&deferral, buffer, 0);
      }
    }
    advance_token();
    if (PUNCTUATION == current.form && ',' == current.ival &&
	SOLO != gg->grouping) {
      advance_token();
      if (AN_IDENT == current.form) {
	if (!strcmp("defer", current.sval)) {
	  replace_in_map(&deferral, buffer, (void*) 1, 0);
	} else {
	  show_source();
	  fprintf(stderr, "Glyph group qualifier \"%s\" not understood.\n",
		  current.sval);
	  skip_to_end("lookup");
	  return 0;
	}
      } else {
	show_source();
	fprintf(stderr, "Expected an identifier.\n");
	got_what(stderr, current);
	skip_to_end("lookup"); return 0;
      }
      advance_token();
    }
  }
  return 1;
}

void mebbe_store_classification(
  const char *name, /* For error messages */
  enum lookup_type table, /* Table in which to store */
  struct classification_data * cldat) {
  void *ptrval;
  char buffer[MAX_P_LEN+1];

  (void) sprintf(buffer, "%p", cldat);
  if (lookup_in_map(deferral, buffer, &ptrval)) {
    if (!ptrval) store_classification(name, table, cldat);
  } else {
    store_classification(name, table, cldat);
  }
}

void mebbe_store_coverage(
  const char *name, /* For error messages */
  enum lookup_type table, /* Table in which to store */
  struct coverage_data * cldat) {
  void *ptrval;
  char buffer[MAX_P_LEN+1];

  (void) sprintf(buffer, "%p", cldat);
  if (lookup_in_map(deferral, buffer, &ptrval)) {
    if (!ptrval) store_coverage(name, table, cldat);
  } else {
    store_coverage(name, table, cldat);
  }
}

static struct classification_data *get_void_classification() {
  struct classification_data *clsn;
  struct class_data *cldat;
  const char void_name[] = " void "; /* Inaccessible from source code */

  if (!(find_gclassification(classifications, void_name, &clsn))) {
    clsn = create_classification(void_name);
/* A unique class name is harder - at least prevent clashes with identifiers */
    locate_class("-234 -034t\ah\xc0\xa0jge9n", clsn, &cldat);
    add_range(0, 0, &cldat->members); /* Do we need any elements? */
  }
  return clsn;
}

int int_set_same(int_set left, int_set right) {
  struct int_range *r1, *r2;

  if (left == right) return 1;
  for (r1 = left, r2 = right; r1 || r2; r1=r1->next, r2 = r2->next) {
    if (r1 && r2) {
      if (r1->start != r2->start) {
	return 0;
      } else if (r1->end != r2->end) {
	return 0;
      }
     } else { /* One is null, and the other isn't */
      return 0;
    }
  }
  return 1;
}

const char *class_name_byno(struct classification_data *classification,
			    int g1) {
  struct gclass_map_entry **els;
  int nels;
  char *name;

  sort_gclass_map(classification->classes, &els, &nels);
  name = els[g1]->key;
  free(els);
  return name;
}

struct int_range *get_glyph_set(const struct glyph_group *gg,
				struct int_range *singleton)
{
  if (gg) {
    switch (gg->grouping) {
    case SOLO:
      singleton->end = singleton->start = gg->x.number;
      singleton->next = 0;
      return singleton;
    case COVERAGE: return gg->x.coverage_data->members;
    case CLASS:    return gg->x.class_data->members;
    default: unwritten; /* internal error */
    }
  } else {
    return 0;
  }
}

no_providing(sequence, struct sequence_table)
encapsulate_map(sequence, struct sequence_table)
static void add_expansion(struct sequence_map *exps, int source,
			  int_set *members,
			  enum lookup_type ldtype, struct sequence_table *seq) {
  char label[6];
  struct sequence_table *unwanted;

  sprintf(label, "%5d", source);
  add_range(source, source, members);
  if (!seq) {
    /* Nothing to do */
  } else if (find_sequence(*exps, label, &unwanted)) {
    switch (ldtype) {
    case GSUB_ALTERNATE:
      /* It would be better to merge the lists! */
    case GSUB_MULTIPLE:
      show_source();
      fprintf(stderr, "Duplicate definition discarded.\n");
      free(seq);
      break;
    default:
      unwritten;
    }
  } else {
    add_sequence(exps, label, seq);
  }
}

struct compiler_ligature_set {
  struct subst_ligature_table *first, *last;
  int n;
  unsigned short int c1;
};

/* Set of ligature substitutions with same initial. */
no_providing(LS, struct compiler_ligature_set)
encapsulate_map(LS, struct compiler_ligature_set)
void add_ligature(struct LS_map *ligs, int start_glyph,
		  struct subst_ligature_table *slt, int *n_max) {
  struct compiler_ligature_set *cls;
  char ascval[6];

  sprintf(ascval, "%5d", start_glyph);
  if (find_LS(*ligs, ascval, &cls)) {
    cls->last->_next = slt;
    cls->last = slt;
    if (++cls->n > *n_max) *n_max = cls->n;
  } else {
    cls = malloc(sizeof * cls);
    if (!cls) {
      fprintf(stderr, "Memory exhausted.\n");
      exit(1);
    }
    cls->first = cls->last = slt;
    cls->n = 1;
    cls->c1 = start_glyph;
    add_LS(ligs, ascval, cls);
  }
}

struct subst_chain_table {
  void *first, *last;      /* Designate struct chain_sub_rule or */
  int n;                   /* chain_sub_class_rule */
  unsigned short int cval; /* First input glyph or class thereof */
};
struct mark_data {
  int class;
  struct anchor_table at;
};

no_providing(chain_set, struct subst_chain_table)
encapsulate_map(chain_set, struct subst_chain_table)
no_providing(values, struct ValueRecord)
encapsulate_map(values, struct ValueRecord)
no_providing(   value_pair, struct hybrid_pair_value_record)
encapsulate_map(value_pair, struct hybrid_pair_value_record)
no_providing   (value_pair_set, struct value_pair_map)
encapsulate_map(value_pair_set, struct value_pair_map)
no_providing   (mark, struct mark_data)
encapsulate_map(mark, struct mark_data)
no_providing   (anchor, struct anchor_table)
encapsulate_map(anchor, struct anchor_table)
#define create_anchor_set create_anchor_map
encapsulate_map(anchor_set, struct anchor_map)

/* process_subtable() handles a subtable of a lookup .*/
void process_subtable() {
  struct subtable_data *std;
  struct lookup_data *ld;
  const char *st_name, *lkp_name;
  FILE *fp;
  int *posn = 0, st_posn;

  if (AN_IDENT != current.form || strcmp("lookup", current.sval)) {
    fprintf(stderr, "process_subtable called other than for lookup.\n");
    exit(1);
  }
  advance_token();
  if (!check_ident()) return;
  if (find_subtable(subtables, current.sval, &std)) {
    if (WRITTEN == std->stored) {
      fprintf(stderr, "Subtable %s has already been defined.\n", current.sval);
      skip_to_end("lookup");
      return;
    }
  } else {
    show_source();
    fprintf(stderr, "Subtable %s has not been declared.\n", current.sval);
    skip_to_end("lookup");
    return;
  }
  ld = std->lookup;
  if (DEFERRED == ld->stored) store_lookup(get_lookup_name(ld), ld);
  if (ld->is_extended) store_pointer(std);
  advance_token();
#if 0
  if (!check_ident()) {
    skip_to_end("lookup");
    return;
  }
#endif
  switch (ld->type) {
  case GSUB_SINGLE:
    fp = gsub;
    posn = &gsub_pos;
    st_posn = *posn;
/* Syntax:
   {g1 > g2}...
 */
    {
      struct coverage_data src;
      struct map dst;
      struct map_entry **dstel;
      struct glyph_group *gg1, *gg2;
      char ascnum[6];
      struct single_subst st;
      int iel, n_subs, junk;
      int loc;
      void *valptr;

      src.members = create_int_set();
      src.lox = no_locations;
      create_map(&dst);
      while (AN_IDENT == current.form && strcmp("end", current.sval)) {
	if (!lookup_in_map(glyph_names, current.sval, (void*)&gg1) ||
	    SOLO != gg1->grouping) {
	  show_source();
	  fprintf(stderr, "'%s' does not identify a single glyph.\n",
		  current.sval);
	  skip_to_end("lookup");
	  return;
	}
	advance_token();
	if (PUNCTUATION != current.form || '>' != current.ival) {
	  show_source();
	  fprintf(stderr, "Expected '>'.\n");
	  got_what(stderr, current);
	  skip_to_end("lookup");
	  return;
	}
	advance_token();
	if (!check_ident()) {
	  skip_to_end("lookup");
	  return;
	}
	if (!lookup_in_map(glyph_names, current.sval, (void*)&gg2) ||
	    SOLO != gg2->grouping) {
	  show_source();
	  fprintf(stderr, "'%s' does not identify a single glyph.\n",
		  current.sval);
	  skip_to_end("lookup");
	  return;
	}
	advance_token();

	sprintf(ascnum, "%5d", gg1->x.number);
	if (lookup_in_map(dst, ascnum, &valptr)) {
	  show_source();
          fprintf(stderr, "More than one substition defined for glyph.\n");
	  skip_to_end("lookup");
	  return;
	}
	insert_in_map(&dst, ascnum, (void*) gg2->x.number);
	(void) add_range(gg1->x.number, gg1->x.number, &src.members);
      }
      st.substFormat = 2;
      st.coverage = 0;
      sort_map(dst, &dstel, &n_subs);
      st.glyphCount = n_subs;
      st.substitute = malloc(n_subs * sizeof * st.substitute);
      if (!st.substitute) {
	fprintf(stderr, "Memory exhausted.\n");
	exit(1);
      }
      for (iel = 0; iel < n_subs; iel++) {
	st.substitute[iel] = (unsigned short int) (int) dstel[iel]->datum;
      }
      add_location(ld->type, *posn, 2 + *posn, 2, &src.lox);
      put_bine_single_subst(fp, &st, &junk, posn);
      store_coverage("(no name)", ld->type, &src);
      std->stored = WRITTEN;
      free(st.substitute);
      free(dstel);
      delete_map(&dst, 0);
/* Must also free the coverage, as there is no way to reference it apart
   from here */
      free(src.lox.loc);
      delete_int_set(&src.members);
    }
    break;
  case GSUB_MULTIPLE:
  case GSUB_ALTERNATE:
    fp = gsub;
    posn = &gsub_pos;
    st_posn = *posn;
/* Syntax is lines of the forms:
   g1 > g2 g3...
   insert_before g1 g2
   insert_after  g1 g2

In "g1 > g2 g3...", each of g1, g2 and g3 must denote a single glyph.
"insert_before g1 g2" is shorthand for g2 > g1 g2
"insert_after g1 g2" is shorthand for  g1 > g1 g2
However, the duplicated glyph may be a class or category - that is the purpose
of these additional syntaxes.  Using a single glyph for the duplicated glyph is
potentially confusing.

The additional syntaxes should probably be completely disabled for
GSUB_ALTERNATE, but the gain would be small and one *might* have a use for it.
 */

/* TODO: Rewrite parsing using backstep_token() on encountering '>'*/

    {
      /* Has keywords "insert_before" and "insert_after". */
      struct sequence_map exps; /* Source glyph in format %5d indexes seqeunce
				   table. */
      struct sequence_table *seq = 0;
      int source, newsource, mels;
      struct glyph_group *gg, *gg1 = 0, *gg2 = 0;
      struct coverage_data local_coverage;
      enum {list = 0, before, after} next_mode = list, mode = list;
      int got_split = 0;

      create_sequence_map(&exps);
      local_coverage.members = create_int_set();
      local_coverage.lox = no_locations;

/* Reading in keyword or source */
      if (!check_ident()) {
	skip_to_end("lookup");
	return;
      }
      if (!strcmp("insert_before", current.sval)) {
	mode = before;
      } else if (!strcmp("insert_after", current.sval)) {
	mode = after;
      } else {
	if (!(gg = get_single(current.sval))) return;
	source = newsource = gg->x.number;
	mode = list;
      }
      advance_token();
      if (list == mode) {
	if (PUNCTUATION != current.form || '>' != current.ival) {
	  show_source();
	  fprintf(stderr, "Required '>'\n");
	  got_what(stderr, current);
	  skip_to_end("lookup");
	  return;
	}
      }
      while (AN_IDENT == current.form ||
	     PUNCTUATION == current.form && '>' == current.ival) {
/* Discovered end of expansion,  or '>' after glyph known to be source */
/* The two cases are distinguished by the value of seq. */
	int hit_end = 0;

	if (PUNCTUATION == current.form) {
	  if (list != mode) {
	    show_source();
	    fprintf(stderr,
		    "Insufficient glyphs for insert_before or insert_after.\n");
	    skip_to_end("lookup");
	    return;
	  } else if (seq && seq->glyphCount <= 1) {
	    show_source();
	    switch (ld->type) {
	    case GSUB_MULTIPLE:
	      fprintf(stderr, "No glyphs for expansion.\n");
	      break;
	    case GSUB_ALTERNATE:
	      fprintf(stderr, "No alternative glyphs.\n");
	      break;
	    default:
	      unwritten;
	    }
	    skip_to_end("lookup");
	    return;
	  }
	  if (list == next_mode && seq) {
	    newsource = seq->substitute[--seq->glyphCount];
	    got_split = 1;
	  }
	} else if (!strcmp("end", current.sval)) {
	  got_split = 1;
	  hit_end = 1;
	}
	if (PUNCTUATION == current.form || got_split) {
	  if (list == mode) {
	    add_expansion(&exps, source, &local_coverage.members,
			  ld->type, seq);
	    seq = 0; /* Prevent confusion */
	    source = newsource;
	    mode = next_mode;
	  } else if (!gg1 || !gg2) {
	    show_source();
	    fprintf(stderr,
		    "Insufficient glyphs for insert_before or insert_after.\n");
	    skip_to_end("lookup");
	    return;
	  } else {
	    struct glyph_group *ggm = (before == mode) ? gg2 : gg1;
	    struct glyph_group *ggs = (before == mode) ? gg1 : gg2;
	    struct int_range sole, *irge = get_glyph_set(ggm, &sole);
	    struct sequence_table *seqi;
	    int fixed = ggs->x.number;
	    int ig;
	    const int melsi = 2;

	    for (; irge; irge = irge->next) {
	      for (ig = irge->start; ig <= irge->end; ig++) {
		if (seqi = malloc(sizeof * seqi)) {
		  seqi->glyphCount = 2;
		  seqi->substitute = 0;
		  seqi->_next = 0;
		} else {
		  fprintf(stderr, "Memory exhausted!\n");
		  exit(1);
		}
		seqi->substitute = malloc(melsi * sizeof * seqi->substitute);
		if (!seqi->substitute) {
		  fprintf(stderr, "Memory exhausted.\n");
		  exit(1);
		}
		seqi->substitute[before == mode] = ig;
		seqi->substitute[before != mode] = fixed;
		add_expansion(&exps, ig, &local_coverage.members,
			      ld->type, seqi);
	      }
	    }
	    gg1 = gg2 = 0;
	  }
	  mode = next_mode;
	  next_mode = list;
	  source = newsource;
	  if (hit_end) break;
	  if (PUNCTUATION == current.form) {
	    if (seq = malloc(sizeof * seq)) {
	      mels = 0;
	      seq->glyphCount = 0;
	      seq->substitute = 0;
	      seq->_next = 0;
	    } else {
	      fprintf(stderr, "Memory exhausted!\n");
	      exit(1);
	    }
	  }
	  got_split = 0;
	}
	if (AN_IDENT == current.form) {/* Process next glyph */
	  if (list == mode || gg2) {
	    if (!strcmp("insert_before", current.sval)) {
	      next_mode = before;
	      got_split = 1;
	    } else if (!strcmp("insert_after", current.sval)) {
	      next_mode = after;
	      got_split = 1;
	    } else {
	      if (!(gg = get_single(current.sval))) return;
	      if (list == mode) {
		if (!seq) {
		  unwritten; /* Missing '>' */
		} else if (mels <= seq->glyphCount) {
		  mels = seq->glyphCount + 2;
		  seq->substitute =
		    realloc(seq->substitute, mels * sizeof * seq->substitute);
		  if (!seq->substitute) {
		    fprintf(stderr, "Memory exhausted.\n");
		    exit(1);
		  }
		}
		seq->substitute[seq->glyphCount++] = gg->x.number;
	      } else {
		got_split = 1;
		newsource = gg->x.number;
	      }
	    }
	  } else {
	    struct glyph_group **ggn = gg1 ? &gg2 : &gg1;

	    if (!!gg1 == (mode == before)) {
	      if (!(lookup_in_map(glyph_names, current.sval, (void**) ggn))) {
		fprintf(stderr, "%s does not identify any glyphs.\n",
			current.sval);
		skip_to_end("lookup");
		return;
	      }
	    } else {
	      if (!(*ggn = get_single(current.sval))) {
		skip_to_end("lookup");
		return;
	      }
	    }
	  }
	}
	advance_token();
      }
      {
	struct sequence_map_entry **stl;
	int ist, nst, junk;
	struct multiple_subst st;
	int stloc, tmploc;
	struct sequence_table *pst;

	assert(sizeof(struct multiple_subst) == sizeof(struct alternate_subst));
	sort_sequence_map(exps, &stl, &nst);
	memset(&st, 0, sizeof st);
	st.substFormat = 1;
	st.coverage = 0;
	st.sequenceCount = nst;
	if (st.sequence = malloc(nst * sizeof * st.sequence)) {
	  memset(st.sequence, 0, nst * sizeof * st.sequence);
	} else {
	  fprintf(stderr, "Memory exhausted.\n");
	  exit(1);
	}
	stloc = *posn;
	add_location(ld->type, *posn, 2 + *posn, 2, &local_coverage.lox);
	put_bine_multiple_subst(fp, &st, &junk, posn);
	for (ist = 0; ist < nst; ist++) {
	  pst = stl[ist]->datum;
	  st.sequence[ist] = *posn - stloc;
	  put_bine_sequence_table(fp, pst, &junk, posn);
	  free(pst);
	}
	gsbp_seek(fp, tmploc = stloc);
	put_bine_multiple_subst(fp, &st, &junk, &tmploc);
	gsbp_seek(fp, *posn);
	store_coverage("(no name)", ld->type, &local_coverage);
	free(stl);
	free(st.sequence);
      }
      delete_int_set(&local_coverage.members);
      free(local_coverage.lox.loc);
      delete_sequence_map(&exps, 0);
    }
    std->stored = WRITTEN;
    break;
  case GSUB_LIGATURE:
    fp = gsub;
    posn = &gsub_pos;
    st_posn = *posn;
/* Syntax: Lines of the forms
   g1 g2... > g3
   delete-before g1 g2
   delete-after  g1 g2

In "g1 g2... > g3", each of g1, g2 and g3 must denote a single glyph.
"delete-before g1 g2" is shorthand for g1 g2 > g2
"delete-after g1 g2" is shorthand for  g1 g2 > g1
However, the duplicated glyph may be a class or category - that is the purpose
of these additional syntaxes.  Using a single glyph for the duplicated glyph is
potentially confusing.
 */
    {
      int n_max;
      struct LS_map ligs;

      create_LS_map(&ligs);
      n_max = 1;
      while (AN_IDENT == current.form && strcmp("end", current.sval)) {
	enum {list, before, after} mode;

/* Initial character of ligature - stored in coverage */
	if (strcmp("delete", current.sval)) {
	  mode = list;
	} else {
	  advance_token();
	  if (PUNCTUATION == current.form && '-' == current.ival) {
	    int gonbad = 0;

	    advance_token();
	    if (AN_IDENT != current.form) {
	      gonbad = 1;
	    } else if (!strcmp("before", current.sval)) {
	      mode = before;
	    } else if (!strcmp("after", current.sval)) {
	      mode = after;
	    } else {
	      gonbad = 1;
	    }
	    if (gonbad) {
	      show_source();
	      fprintf(stderr, "Required identifier \"before\" or \"after\"");
	      got_what(stderr, current);
	      skip_to_end("lookup");
	      return;
	    }
	  } else {
	    mode = list;
	    backstep_token();
	  }
	}
	if (list == mode) {
	  struct glyph_group *ggi, *ggs, *ggl;
	  int start_glyph;
	  int m_comp;
	  struct subst_ligature_table *slt;

	  if (!(ggi = get_single(current.sval))) return;
	  start_glyph = ggi->x.number;
	  if (!(slt = malloc(sizeof * slt))) {
	    fprintf(stderr, "Memory exhausted.\n");
	    exit(1);
	  }
	  slt->_next = 0;
	  slt->compCount = 1;
	  m_comp = 2;
	  if (!(slt->component = malloc(m_comp * sizeof * slt->component))) {
	    fprintf(stdout, "Memory exhausted.\n");
	    exit(1);
	  }
	  advance_token();
/* Subsequent characters of ligature */
	  while (AN_IDENT == current.form && strcmp("end", current.sval)) {
	    ggs = get_single(current.sval);
	    if (!ggs) return;
	    if (slt->compCount > m_comp) {
	      m_comp = slt->compCount;
	      slt->component =
		realloc(slt->component, m_comp * sizeof * slt->component);
	      if (!slt->component) {
		fprintf(stdout, "Memory exhausted.\n");
		exit(1);
	      }
	    }
	    slt->component[slt->compCount - 1] = ggs->x.number;
	    slt->compCount++;
	    advance_token();
	  }
/* 'Ligature' character - not necessarily a ligature! */
	  if (PUNCTUATION != current.form || '>' != current.ival) {
	    show_source();
	    fprintf(stderr, "Expected '>'.\n");
	    got_what(stderr, current);
	    skip_to_end("lookup");
	    return;
	  }
	  advance_token();
	  if (!check_ident()) {
	    skip_to_end("lookup");
	    return;
	  }
	  ggl = get_single(current.sval);
	  if (!ggl) return;
	  slt->ligGlyph = ggl->x.number;
	  add_ligature(&ligs, start_glyph, slt, &n_max);
	} else {
	  struct glyph_group *ggs, *ggm;
	  struct int_range sole, *irge;
	  struct sequence_table *seqi;
	  int fixed, ig;
	  const int melsi = 2;

	  advance_token(); /* to glyph 1 */
	  if (!check_ident()) {
	    skip_to_end("lookup");
	    return;
	  }
	  if (before == mode) {
	    if (!(ggs = get_single(current.sval))) return;
	  } else {
	    if (!lookup_in_map(glyph_names, current.sval, (void**) &ggm)) {
	      show_source();
	      fprintf(stderr, "%s does not identify a group of glyphs.\n",
		      current.sval);
	      skip_to_end("lookup");
	      return;
	    }
	  }
	  advance_token(); /* to glyph 2 */
	  if (!check_ident()) {
	    skip_to_end("lookup");
	    return;
	  }
	  if (after == mode) {
	    if (!(ggs = get_single(current.sval))) return;
	  } else {
	    if (!lookup_in_map(glyph_names, current.sval, (void**)&ggm)) {
	      show_source();
	      fprintf(stderr, "%s does not identify a group of glyphs.\n",
		      current.sval);
	      skip_to_end("lookup");
	      return;
	    }
	  }
	  fixed = ggs->x.number;
	  for (irge = get_glyph_set(ggm, &sole); irge; irge = irge->next) {
	    for (ig = irge->start; ig <= irge->end; ig++) {
	      struct subst_ligature_table *slti;

	      slti = malloc(sizeof * slti);
	      if (slti) slti->component = malloc(sizeof * slti->component);
	      if (!slti || !slti->component) {
		fprintf(stderr, "Memory exhausted.\n");
		exit(1);
	      }
	      slti->_next = 0;
              slti->compCount = 2;
	      slti->ligGlyph = ig;
	      if (before == mode) {
		slti->component[0] = ig; /* *Second* component */
		add_ligature(&ligs, fixed, slti, &n_max);
	      } else {
		slti->component[0] = fixed;
		add_ligature(&ligs, ig, slti, &n_max);
	      }
	    }
	  }
	}
	advance_token(); /* to next ligature definition */
      }
/* Now write the subtable */
      {
	struct LS_map_entry **lig_list;
	struct coverage_data cd;
	struct ligature_subst ls;
	struct ligature_set ligset;
	struct compiler_ligature_set *cligset;
	int loc_ls, loc_set, loc_tab;
	int i_set, n_sets, junk;

	sort_LS_map(ligs, &lig_list, &n_sets);
	ls.substFormat = 1;
	ls.coverage = 0;
	ls.ligSetCount = n_sets;
	ls.ligatureSet = malloc(n_sets * sizeof * ls.ligatureSet);
	ligset.ligature = malloc(n_max * sizeof * ligset.ligature);
	if (!ls.ligatureSet || !ligset.ligature) {
	  fprintf(stderr, "Memory exhausted.\n");
	  exit(1);
	}
	cd.members = create_int_set();
	cd.lox = no_locations;
	add_location(ld->type, *posn, 2 + *posn, 2, &cd.lox);
	loc_ls = *posn;
	put_bine_ligature_subst(fp, &ls, &junk, posn);
	for (i_set = 0; i_set < n_sets; i_set++) {
	  int i, offset;
	  struct subst_ligature_table *iptr, *optr;

	  loc_set = *posn;
	  offset = loc_set - loc_ls;
	  if (offset < 0 || offset > 0xFFFF) {
	    fprintf(stderr, "Offset within subtable %s is too large.\n",
		    get_subtable_name(std));
	    offset = 0;
	  }
	  ls.ligatureSet[i_set] = offset;
	  cligset = lig_list[i_set]->datum;
	  ligset.ligatureCount = cligset->n;
	  (void) add_range(cligset->c1, cligset->c1, &cd.members);
	  put_bine_ligature_set(fp, &ligset, &junk, posn);
	  for (i = 0, iptr = cligset->first, optr = 0;
	       i < cligset->n;
	       i++, iptr = iptr->_next) {
	    free(optr);
	    loc_tab = *posn;
	    offset = loc_tab - loc_set;
	    if (offset < 0 || offset > 0xFFFF) {
	      fprintf(stderr, "Offset within subtable %s is too large.\n",
		      get_subtable_name(std));
	      offset = 0;
	    }
	    ligset.ligature[i] = offset;
	    put_bine_subst_ligature_table(fp, iptr, &junk, posn);
	    free(iptr->component);
	    optr = iptr;
	  }
	  free(optr);
/* Rewrite ligature subset table with update offsets */
	  gsbp_seek(fp, loc_set);
	  put_bine_ligature_set(fp, &ligset, &junk, &loc_set);
	  gsbp_seek(fp, *posn);
	}
/* Rewrite subtable with updated offsets */
	gsbp_seek(fp, loc_ls);
	put_bine_ligature_subst(fp, &ls, &junk, &loc_ls);
	free(ls.ligatureSet);
	free(ligset.ligature);
	gsbp_seek(fp, *posn);
	store_coverage("no_name", GSUB_1, &cd);
	std->stored = WRITTEN;
	free(lig_list);
/* Must also free the coverage, as there is no way to reference it apart
   from here */
	free(cd.lox.loc);
	delete_int_set(&cd.members);
      }
      delete_LS_map(&ligs, 1);
    }
    break;
  case GSUB_CONTEXT:
  case GPOS_CONTEXT:
  case GSUB_CHAINED:
  case GPOS_CHAINED:
    switch (ld->type) {
    case GSUB_CONTEXT:
    case GSUB_CHAINED:
      fp = gsub;
      posn = &gsub_pos;
      break;
    case GPOS_CONTEXT:
    case GPOS_CHAINED:
      fp = gpos;
      posn = &gpos_pos;
      break;
    default:
      unwritten;
    }
    st_posn = *posn;
    {
/* Syntax:
   [context] gb... | gi... | ga...
   { 0 lookup_0 }...
 */
      struct chain_context_subst ccs;
      struct context_subst ucs; /* For GSUB_CONTEXT & GPOS_CONTEXT */
      const int unchained = GSUB_CONTEXT == ld->type ||
	                    GPOS_CONTEXT == ld->type;
/* Start of variables defining context */
      struct glyph_group **sb, **si, **sa; /* *All* assembled in glyph order */
      struct subst_lookup_record *slr;
      int nb, ni, na, n_slr; /* Numbers of elements in use */
      int mb, mi, ma, m_slr; /* Numbers of elements allocated */
      struct classification_data *classn_b = 0, *classn_i = 0, *classn_a = 0;
/* End of variables defining context */
/* Chain sets - not used for coverages (format 3) */
      int max_input1_class = 0; /* Highest start input class */
      struct chain_set_map chain_sets; /* Indexed by glyph or class number, format %5d. */
      /* The values are pointers to struct subst_chain_table. */
#if 0 /* Yanked out of context to use function definitions */
      struct subst_chain_table {
	void *first, *last;      /* Designate struct chain_sub_rule or */
	int n;                   /* chain_sub_class_rule */
	unsigned short int cval; /* First input glyph or class thereof */
      };
#endif
      struct coverage_data g1_coverage; /* First input *glyph* coverage.  Not */
                                        /* for contexts defined by coverages. */
      int max_rules = 0;

      create_map(&deferral);
      create_chain_set_map(&chain_sets);
      g1_coverage.members = create_int_set();
      g1_coverage.lox = no_locations;
      mb = mi = ma = m_slr = 0;
      sb = si = sa = 0;
      slr = 0;
      context_ctype = NULL;

      while (AN_IDENT == current.form && strcmp("end", current.sval) ||
	     PUNCTUATION == current.form && '|' == current.ival) {
	int is, ie;
	struct subst_chain_table *sct;
	char ascval[6];
#if 1
	if (AN_IDENT == current.form && !strcmp("context", current.sval)) {
	  advance_token();
	  if (PUNCTUATION == current.form && '|' == current.ival) {
	    /* New token is OK. */
	  } else if (!check_ident()) {
	    skip_to_end("lookup");
	    return;
	  }
	}
#endif
	nb = ni = na = n_slr = 0;
	if (unchained) {
	  nb = 0;
	} else {
	  if (!chain_glyph_string(&sb, &mb, &nb, &classn_b)) return;
/* Reverse glyphs*/
	  for (is = 0, ie = nb-1; is < ie; is++, ie--) {
	    struct glyph_group *temp = sb[is];
	    sb[is] = sb[ie];
	    sb[ie] = temp;
	  }
	  if (current.form != PUNCTUATION || current.ival != '|') {
	    show_source();
	    fprintf(stderr, "Expected '|' marking off glyphs before.\n");
	    skip_to_end("lookup");
	    return;
	  }
	  advance_token();
	}
	if (!chain_glyph_string(&si, &mi, &ni, &classn_i)) return;
	if (!ni) {
	  show_source();
	  fprintf(stderr, "An input glyph is required.\n");
	  skip_to_end("lookup");
	  return;
	}
	if (unchained) {
	  na = 0;
	} else {
	  if (current.form != PUNCTUATION || current.ival != '|') {
	    show_source();
	    fprintf(stderr, "Expected '|' introducing glyphs after.\n");
	    skip_to_end("lookup");
	    return;
	  }
	  advance_token();
	  if (!chain_glyph_string(&sa, &ma, &na, &classn_a)) return;
	}
/* Now process the indirectly invoked lookups. */
	while (AN_INT == current.form) {
	  struct lookup_data *ld;
	  if (++n_slr > m_slr) {
	    m_slr++;
	    if (!(slr = realloc(slr, m_slr * sizeof * slr))) {
	      fprintf(stderr, "Memory exhausted!\n");
	      exit(1);
	    }
	  }
	  slr[n_slr-1].sequenceIndex = current.ival;
	  advance_token();
	  if (!check_ident()) {
	    skip_to_end("lookup"); return;
	  } else if (find_lookup(lookups, current.sval, &ld)) {
	    slr[n_slr-1].lookupListIndex = ld->lookup_index;
	  } else {
	    show_source();
	    fprintf(stderr, "'%s' is not the name of a look-up.\n",
		    current.sval);
	    skip_to_end("lookup");
	    return;
	  }
	  advance_token();
	}
/* Now store the 'rule' */
	switch (context_ctype->grouping) {
	case SOLO:
	  {
	    int k, number;
	    struct chain_sub_rule *rule;

	    if (rule = malloc(sizeof * rule)) {
	      rule->backtrack =
		nb ? malloc(nb * sizeof * rule->backtrack) : 0;
	      rule->input =
		ni > 1 ? malloc((ni - 1) * sizeof * rule->input) : 0;
	      rule->lookahead =
		na ? malloc(na * sizeof * rule->lookahead) : 0;
	    }
	    if (!rule ||
		nb && !rule->backtrack ||
		ni > 1 && !rule->input ||
		na && !rule->lookahead) {
	      fprintf(stderr, "Memory exhausted.\n");
	      exit(1);
	    }
	    rule->backtrackGlyphCount = nb;
	    for (k = 0; k < nb; k++) {
	      rule->backtrack[k] = sb[k]->x.number;
	    }
	    rule->inputGlyphCount = ni;
	    number = si[0]->x.number;
	    for (k = 1; k < ni; k++) {
	      rule->input[k-1] = si[k]->x.number;
	    }
	    rule->lookaheadGlyphCount = na;
	    for (k = 0; k < na; k++) {
	      rule->lookahead[k] = sa[k]->x.number;
	    }
	    rule->substCount = n_slr;
	    rule->substLookupRecord = slr;
	    slr = 0;
	    m_slr = 0;
	    rule->_next = NULL;
	    sprintf(ascval, "%5d", number);
	    if (find_chain_set(chain_sets, ascval, &sct)) {
	      struct chain_sub_rule *prev = sct->last;

	      prev->_next = rule;
	      sct->last = rule;
	      sct->n++;
	    } else if (!(sct = malloc(sizeof * sct))) {
	      fprintf(stderr, "Memory exhausted.\n");
	      exit(1);
	    } else {
	      sct->first = sct->last = rule;
	      sct->n = 1;
	      sct->cval = number;
	      add_chain_set(&chain_sets, ascval, sct);
	      (void) add_range(number, number, &g1_coverage.members);
	    }
	    if (sct->n > max_rules) max_rules = sct->n;
	  }
	  break;
	case CLASS:
	  {
	    int k, number;
	    struct chain_sub_class_rule *rule;

	    if (rule = malloc(sizeof * rule)) {
	      rule->backtrack =
		nb ? malloc(nb * sizeof * rule->backtrack) : 0;
	      rule->input =
		ni > 1 ? malloc((ni - 1) * sizeof * rule->input) : 0;
	      rule->lookahead =
		na ? malloc(na * sizeof * rule->lookahead) : 0;
	    }
	    if (!rule ||
		nb && !rule->backtrack ||
		ni > 1 && !rule->input ||
		na && !rule->lookahead) {
	      fprintf(stderr, "Memory exhausted.\n");
	      exit(1);
	    }
	    rule->backtrackGlyphCount = nb;
	    for (k = 0; k < nb; k++) {
	      rule->backtrack[k] = sb[k]->x.class_data->number;
	    }
	    rule->inputGlyphCount = ni;
	    number = si[0]->x.class_data->number;
	    if (number > max_input1_class) max_input1_class = number;
	    for (k = 1; k < ni; k++) {
	      rule->input[k-1] = si[k]->x.class_data->number;
	    }
	    rule->lookaheadGlyphCount = na;
	    for (k = 0; k < na; k++) {
	      rule->lookahead[k] = sa[k]->x.class_data->number;
	    }
	    rule->substCount = n_slr;
	    rule->substLookupRecord = slr;
	    slr = 0;
	    m_slr = 0;
	    rule->_next = NULL;
	    sprintf(ascval, "%5d", number);
	    if (find_chain_set(chain_sets, ascval, &sct)) {
	      struct chain_sub_class_rule *prev = sct->last;

	      prev->_next = rule;
	      sct->last = rule;
	      sct->n++;
	    } else if (!(sct = malloc(sizeof * sct))) {
	      fprintf(stderr, "Memory exhausted.\n");
	      exit(1);
	    } else {
	      sct->first = sct->last = rule;
	      sct->n = 1;
	      sct->cval = number;
	      add_chain_set(&chain_sets, ascval, sct);
	      if (number) {
		(void) add_set(si[0]->x.class_data->members,
			       &g1_coverage.members);
	      } else {
/* Reaching here is bad news.  Perhaps explict specification of the coverage
   should be allowed. */
	        int_set cl0 = create_int_set();
		struct gclass_map_entry **cll;
		int icl, ncl;
		struct class_data *cld;

		add_range(0, last_glyph, &cl0);
		sort_gclass_map(si[0]->x.class_data->classification->classes,
				&cll, &ncl);
		for (icl = 0; icl < ncl; icl++) {
		  cld = cll[icl]->datum;
		  if (cld->number) sub_set(cld->members, &cl0);
		}
		if (DOFREE) free(cll);
		add_set(cl0, &g1_coverage.members);
		if (DOFREE) delete_int_set(&cl0);
	      }
	    }
	    if (sct->n > max_rules) max_rules = sct->n;
	  }
	  break;
	case COVERAGE:
	  break; /* No temporary storage, because only one context */
	default:
	  unwritten;
	  break;
	}
	if (COVERAGE == context_ctype->grouping) break;
      }
/* Now write to file */
      memset(&ccs, 0, sizeof ccs);
      if (COVERAGE == context_ctype->grouping) {
	int wloc = 0;
	int k, junk;

	ccs.substFormat = 3; wloc++;
	ccs.backtrackGlyphCount = nb;
	if (!unchained) wloc++;
	if (nb) {
	  if (!(ccs.coverageb = malloc(nb * sizeof * ccs.coverageb))) {
	    fprintf(stderr, "Memory exhausted.\n");
	    exit(1);
	  }
	  for (k = 0; k < nb; k++) {
	    add_location(ld->type, *posn, 2 * wloc + *posn, 2,
			     &sb[k]->x.coverage_data->lox);
		ccs.coverageb[k] = 0; wloc++;
	  }
	}
	ccs.inputGlyphCount = ni; wloc++;
	if (ni) {
	  if (!(ccs.coveragei = malloc(ni * sizeof * ccs.coveragei))) {
	    fprintf(stderr, "Memory exhausted.\n");
	    exit(1);
	  }
	  for (k = 0; k < ni; k++) {
	    add_location(ld->type, *posn, 2 * wloc + *posn, 2,
			 &si[k]->x.coverage_data->lox);
	    ccs.coveragei[k] = 0; wloc++;
	  }
	}
	ccs.lookaheadGlyphCount = na;
	if (!unchained) wloc++;
	if (na) {
	  if (!(ccs.coveragea = malloc(na * sizeof * ccs.coveragea))) {
	    fprintf(stderr, "Memory exhausted.\n");
	    exit(1);
	  }
	  for (k = 0; k < na; k++) {
	    add_location(ld->type, *posn, 2 * wloc + *posn, 2,
			 &sa[k]->x.coverage_data->lox);
	    ccs.coveragea[k] = 0; wloc++;
	  }
	}
	ccs.substCount = n_slr;
	ccs.substLookupRecord = slr;
	for (k = 1; k < n_slr; k++) {
	  ccs.substLookupRecord[k-1]._next = &ccs.substLookupRecord[k];
	}

	if (unchained) {
	  ucs.substFormat = ccs.substFormat;
          ucs.glyphCount  = ccs.inputGlyphCount;
	  ucs.substCount  = ccs.substCount;
          ucs.coverage3   = ccs.coveragei;
          ucs.substLookupRecord = ccs.substLookupRecord;
	  put_bine_context_subst(fp, &ucs, &junk, posn);
	} else {
	  put_bine_chain_context_subst(fp, &ccs, &junk, posn);
	}
	if (DOFREE) free(ccs.substLookupRecord);
	ccs.substLookupRecord = 0;
      } else {
	struct chain_set_map_entry **ruleset;
	struct chain_sub_rule_set *set; /* Also used for classes (TBC) */
	struct chain_sub_rule_set csrs;
	int iset, nset;
	int loc, ccs_loc, set_loc;
	int junk;
	unsigned short int *setoffset;
	int id;
	int offset;
	struct subst_chain_table *sct;

	sort_chain_set_map(chain_sets, &ruleset, &nset);
	if (COVERAGE != context_ctype->grouping) {
	  if (!(csrs.chainSubRule =
		malloc(max_rules * sizeof * csrs.chainSubRule))) {
	    fprintf(stderr, "Memory exhausted.\n");
	    exit(1);
	  }
	}
	gsbp_seek(fp, ccs_loc = *posn);
	switch (context_ctype->grouping) {
	case SOLO:
	  {
	    ccs.substFormat = 1;
	    ccs.coverage = 0;
	    add_location(ld->type, *posn, 2 + *posn, 2, &g1_coverage.lox);
	    ccs.chainSubRuleSetCount = nset;
	    setoffset = ccs.chainSubRuleSet =
	      malloc(nset * sizeof * ccs.chainSubRuleSet);
	    if (ccs.chainSubRuleSet) {
	      int k;
	      for (k = 0; k < nset; k++) ccs.chainSubRuleSet[k] = 0;
	    } else {
	      fprintf(stderr, "Memory exhausted.\n");
	      exit(1);
	    }
	  }
	  break;
	case CLASS:
	  {
	    ccs.substFormat = 2;
	    ccs.coverage = 0;
	    add_location(ld->type, *posn, 2 + *posn, 2, &g1_coverage.lox);
	    ccs.backtrackClassDef = 0;
	    if (!classn_b) classn_b = get_void_classification();
	    if (unchained) {
	      add_location(ld->type, *posn, 4 + *posn, 2, &classn_i->lox);
	    } else {
	      add_location(ld->type, *posn, 4 + *posn, 2, &classn_b->lox);
	      add_location(ld->type, *posn, 6 + *posn, 2, &classn_i->lox);
	    }
	    ccs.inputClassDef = 0;
	    ccs.lookaheadClassDef = 0;
	    if (!classn_a) classn_a = get_void_classification();
	    if (!unchained) {
	      add_location(ld->type, *posn, 8 + *posn, 2, &classn_a->lox);
	    }
	    ccs.chainSubClassSetCnt = max_input1_class + 1;
	    setoffset = ccs.chainSubClassSet =
	      malloc(ccs.chainSubClassSetCnt * sizeof * ccs.chainSubClassSet);
	    if (ccs.chainSubClassSet) {
	      int k;
	      for (k = 0; k < ccs.chainSubClassSetCnt; k++) {
		ccs.chainSubClassSet[k] = 0; 
	      }
	    } else {
	      fprintf(stderr, "Memory exhausted.\n");
	      exit(1);
	    }
	  }
	  break;
	case COVERAGE:
	default:
	  unwritten;
	}
	if (unchained) {
	  ucs.substFormat = ccs.substFormat;
	  ucs.coverage    = ccs.coverage;
	  switch (ccs.substFormat) {
	  case 1:
	    ucs.subRuleSetCount = ccs.chainSubRuleSetCount;
	    ucs.subRuleSet      = ccs.chainSubRuleSet;
	    break;
	  case 2:
	    ucs.classDef       = ccs.inputClassDef;
	    ucs.subClassSetCnt = ccs.chainSubClassSetCnt;
	    ucs.subClassSet    = ccs.chainSubClassSet;
	    break;
	  default:
	    unwritten;
	  }
	  put_bine_context_subst(fp, &ucs, &junk, posn);
	} else {
	  put_bine_chain_context_subst(fp, &ccs, &junk, posn);
	}
	for (iset = 0; iset < nset; iset++) {
	  int k;
	  struct chain_sub_rule *iptr, *optr;
	  struct sub_rule usr;
/* Use same types for by glyphs and by classes */
	  if (CLASS == context_ctype->grouping) {
	    sscanf(ruleset[iset]->key, "%d", &id);
	    if (id < 0) {
	      show_source();
	      fprintf(stdout, "Internal error: Negative class index %d for "
		      "first slot in 'input' glyph string.\n", id);
	      exit(1);
	    } else if (id > max_input1_class) {
	      show_source();
	      fprintf(stdout, "Internal error: Class index %d for first slot "
		      "in 'input' glyph string exceeds recorded maximum %d.\n",
		      id, max_input1_class);
	      exit(1);
	    } else if (0) { /* Correct case */
	      show_source();
	      fprintf(stdout, "Storing rule set for when character in first "
		      "slot of 'input' has class index %d.\n", id);
	    }
	  } else {
	    id = iset;
	  }
	  sct = ruleset[iset]->datum;
	  csrs.chainSubRuleCount = sct->n;
	  for (k = 0; k < sct->n; k++) csrs.chainSubRule[k] = 0;
	  set_loc = *posn;
	  offset = *posn - ccs_loc;
	  if (offset < 0 || offset > 0xFFFF) {
	    fprintf(stderr, "Offset within subtable %s is too large.\n",
		    get_subtable_name(std));
	    offset = 0;
	  }
	  setoffset[id] = offset;
	  put_bine_chain_sub_rule_set(fp, &csrs, &junk, posn);
	  for (k = 0, optr = 0, iptr = sct->first;
	       k < sct->n;
	       optr = iptr, iptr = iptr->_next, k++) {
	    int j;

	    if (DOFREE) free(optr);
	    offset = *posn - set_loc;
	    if (offset < 0 || offset > 0xFFFF) {
	      fprintf(stderr, "Offset within subtable %s is too large.\n",
		      get_subtable_name(std));
		offset = 0;
	    }
	    csrs.chainSubRule[k] = offset;
	    for (j = 1; j < iptr->substCount; j++) {
	      iptr->substLookupRecord[j-1]._next =
		&iptr->substLookupRecord[j];
	    }
	    if (iptr->substCount) {
	      iptr->substLookupRecord[iptr->substCount-1]._next = NULL;
	    }
	    if (unchained) {
	      usr.glyphCount        = iptr->inputGlyphCount;
	      usr.input             = iptr->input;
	      usr.substCount        = iptr->substCount;
              usr.substLookupRecord = iptr->substLookupRecord;
	      put_bine_sub_rule(fp, &usr, &junk, posn);
	    } else {
	      put_bine_chain_sub_rule(fp, iptr, &junk, posn);
	    }
	    if (DOFREE) free(iptr->substLookupRecord);
	    iptr->substLookupRecord = 0;
	  }
	  gsbp_seek(fp, loc = set_loc);
	  put_bine_chain_sub_rule_set(fp, &csrs, &junk, &loc);
	  gsbp_seek(fp, *posn);
	  if (DOFREE) free(optr);
	}
	gsbp_seek(fp, loc = ccs_loc);
	if (unchained) {
	  put_bine_context_subst(fp, &ucs, &junk, &loc);
	} else {
	  put_bine_chain_context_subst(fp, &ccs, &junk, &loc);
	}
	gsbp_seek(fp, loc = *posn);
	if (DOFREE) free(ruleset);
      }
/* Now write undeferred coverages and classifications */
      switch (context_ctype->grouping) {
      case SOLO:
	store_coverage("(implied coverage)", ld->type, &g1_coverage);
	break;
      case CLASS:
	store_coverage("(implied coverage)", ld->type, &g1_coverage);
	mebbe_store_classification("backtrack class def", ld->type, classn_b);
	mebbe_store_classification("input class def", ld->type, classn_i);
	mebbe_store_classification("lookahead class def", ld->type, classn_a);
	break;
      case COVERAGE:
	{
	  int k;
	  for (k = 0; k < nb; k++) {
	    mebbe_store_coverage("backtrack coverage", ld->type,
				 sb[k]->x.coverage_data);
	  }
	  for (k = 0; k < ni; k++) {
	    mebbe_store_coverage("input coverage", ld->type,
				 si[k]->x.coverage_data);
	  }
	  for (k = 0; k < na; k++) {
	    mebbe_store_coverage("lookahead coverage", ld->type,
				 sa[k]->x.coverage_data);
	  }
	}
	break;
      default:
	unwritten; /* Uneliminatable dead code */
      }
      if (DOFREE) free_chain_context_subst(&ccs);
      if (DOFREE) {free(sb); free(sa); free(si);}
      if (DOFREE) {
	delete_map(&deferral, 0);
	delete_chain_set_map(&chain_sets, 0);
	delete_int_set(&g1_coverage.members);
      } else {
	create_map(&deferral);
	create_chain_set_map(&chain_sets);
	g1_coverage.members = create_int_set();
      }
      if (DOFREE) free(g1_coverage.lox.loc);
      std->stored = WRITTEN;
    }
    break;
  case GSUB_REVERSE:
    fp = gsub;
    posn = &gsub_pos;
    fprintf(stderr, "Lookups of this type are not yet supported.\n");
    skip_to_end("lookup");
    /* Set st_posn and update std->stored when implemented! */
    return;
#define XPlacement 0x0001 	
#define YPlacement 0x0002 	
#define XAdvance   0x0004 	
#define YAdvance   0x0008 	
#define XPlaDevice 0x0010 	
#define YPlaDevice 0x0020 	
#define XAdvDevice 0x0040 	
#define YAdvDevice 0x0080 	
  case GPOS_SINGLE:
    fp = gpos;
    posn = &gpos_pos;
    st_posn = *posn;
    {
/* Index by glyph number (%5.5d) and yields a value record. */
      struct values_map values;
      unsigned short int vf0; /* Value format - vf0 for first entered */
      struct ValueRecord vr;
      char gfid[6];
      int gonbad, all_same, flag, junk;
      struct glyph_group *gg;
      /*
	States:

	NO_GID - Before first glyph ID is specified.
	VALUE1 - Entering first value
	NORMAL - Working on second glyph or later.

	Once entered, a value remains until overwritten.

	Parallel:
	all_same - whether all values are the same.  Reaffirming counts as
                   changing.
       */
      enum {NO_GID, VALUE1, NORMAL} state;
      enum {notoken = 0, dx, dy, ax, ay, glyph, end} tokid;
      create_values_map(&values);
      vf0 = 0;
      state = NO_GID;
      all_same = 1;
      gonbad = 0;
      memset(&vr, 0, sizeof vr);

      while (AN_IDENT == current.form) {
	if (0) {
	  /* This tricky code remembers the token in tokid */
	} else if (!strcmp("dx", current.sval) && (tokid = dx) ||
		   !strcmp("dy", current.sval) && (tokid = dy) ||
		   !strcmp("ax", current.sval) && (tokid = ax) ||
		   !strcmp("ay", current.sval) && (tokid = ay)) {
	  advance_token();
	  if (AN_INT != current.form) {
	    show_source();
	    fprintf(stdout, "Expected an integer.\n");
	    gonbad = 1;
	    continue;
	  }
	  if (current.ival < -0x8000 || 0x7FFF < current.ival) {
	    show_source();
	    fprintf(stdout, "Not a valid signed 16-bit number.\n");
	    gonbad = 1;
	    advance_token();
	    continue;
	  }
	  switch (state) {
	  case NO_GID:
	    show_source();
	    fprintf(stdout, "Expected a glyph name, not a number.\n");
	    gonbad = 1;
	  case VALUE1:
	  case NORMAL:
	    switch (tokid) {
	    case dx:
	      vr.xPlacement = current.ival;
	      flag = XPlacement;
	      break;
	    case dy:
	      vr.yPlacement = current.ival;
	      flag = YPlacement;
	      break;
	    case ax:
	      vr.xAdvance = current.ival;
	      flag = XAdvance;
	      break;
	    case ay:
	      vr.yAdvance = current.ival;
	      flag = YAdvance;
	      break;
	    default:
	      unwritten; /* Should be inaccessible */
	    }
	    if (NORMAL == state) {
	      if (flag & ~vf0) {
		show_source();
		fprintf(stderr, "Field was not present for first value - "
			"will be ignored.\n");
	      }
	      all_same = 0;
	    } else {
	      vf0 |= flag;
	    }
	    break;
	  default:
	    unwritten; /* Internal error trap. */
	  }
	  advance_token();
	} else if (!strcmp("glyph", current.sval) && (tokid = glyph) ||
		   !strcmp("end", current.sval)   && (tokid = end)) {
	  if (gg && NO_GID != state) {
	    struct ValueRecord *vrnew, *unwanted;
	    struct int_range *irge, local;

	    switch (gg->grouping) {
	    case SOLO:
	      local.end = local.start = gg->x.number;
	      local.next = 0;
	      irge = &local;
	      break;
	    case CLASS:
	      irge = gg->x.class_data->members;
	      break;
	    case COVERAGE:
	      irge = gg->x.coverage_data->members;
	      break;
	    default:
	      unwritten; /* Internal error */
	    }
	    for (; irge; irge = irge->next) {
	      int ogl;
	      for (ogl = irge->start; ogl <= irge->end; ogl++) {
		sprintf(gfid, "%5.5d", ogl);
		if (find_values(values, gfid, &unwanted)) {
		  if (!all_same) {
		    show_source();
		    fprintf(stderr, "Multiple values for glyph %d.\n",
			    ogl);
		    gonbad = 1;
		  }
		} else {
		  if (!(vrnew = malloc(sizeof * vrnew))) {
		    fprintf(stderr, "Memory exhausted.\n");
		    exit(1);
		  }
		  *vrnew = vr;
		  add_values(&values, gfid, vrnew);
		}
	      }
	    }
	  }
	  if (glyph == tokid) {
	    advance_token();
	    if (!check_ident()) {
	      gonbad = 1;
	      continue;
	    } else if (lookup_in_map(glyph_names, current.sval, (void**)&gg)) {
	      if (NO_GID == state) {
		state = VALUE1;
	      } else {
		state = NORMAL;
	      }
	    } else {
	      fprintf(stderr, "'%s' does not define any glyphs.\n",
		      current.sval);
	      gonbad = 1;
	      gg = 0;
	    }
	    advance_token();
	  } else if (end == tokid) {
	    break;
	  }
	} else {/* Keyword not recognised */
	  show_source();
	  fprintf(stderr, "Keyword %s not recognised.\n", current.sval);
	  gonbad = 1;
	  skip_to_end("lookup");
	  return;
	}
      }
      if (AN_IDENT != current.form) {
	check_ident();
	gonbad = 1;
      }
      if (!gonbad) { /* Store the lookup */
	struct sap subtable;
	struct coverage_data mycov;
	int use_mycov;
	struct values_map_entry **value_list;
	int ivl, nvl;

	mycov.members = create_int_set();
	mycov.lox = no_locations;
	memset(&subtable, 0, sizeof subtable);
	subtable.coverage = 0;
	if (VALUE1 == state && gg && COVERAGE == gg->grouping) {
	  add_location(ld->type, *posn, 2 + *posn, 2,
		       &gg->x.coverage_data->lox);
	  use_mycov = 0;
	} else {
	  add_location(ld->type, *posn, 2 + *posn, 2, &mycov.lox);
	  use_mycov = 1;
	}
	subtable.valueFormat = vf0;
	sort_values_map(values, &value_list, &nvl);
	if (all_same) {
	  subtable.posFormat = 1;
	  subtable.value = vr;
	} else {
	  struct ValueRecord **ptr;
	  subtable.posFormat = 2;
	  subtable.valueCount = nvl;
	  ptr = &subtable.valueArr;
	  for (ivl = 0; ivl < nvl; ivl++) {
	    *ptr = value_list[ivl]->datum;
	    ptr = &(*ptr)->_next;
	  }
	  *ptr = 0;
#if 0
	  print_sap(stdout, &subtable, &junk, "single", 0);
#endif
	}
	if (use_mycov) {
	  for (ivl = 0; ivl < nvl; ivl++) {
	    int gno;
	    sscanf(value_list[ivl]->key, "%d", &gno);
	    add_range(gno, gno, &mycov.members);
	  }
	}
	put_bine_sap(fp, &subtable, &junk, posn);
        if (DOFREE) free_sap(&subtable);
	if (use_mycov) store_coverage("(anonymous)", ld->type, &mycov);
	delete_int_set(&mycov.members);
	if (DOFREE) free(mycov.lox.loc);
	if (DOFREE) free(value_list);
	std->stored = WRITTEN;
      }
      /* Clean up */
      delete_values_map(&values, all_same);
    }
    break;
  case GPOS_PAIR:
    fp = gpos;
    posn = &gpos_pos;
    st_posn = *posn;
    {
      /* TODO: Explicit coverage, and adjust analysis to use it for either
	 format. */
      /*
	States:

	NO_GID1 - Before first glyph ID is specified.
        NO_GID2 - Before second glyph ID entered - this state may be reentered.
	VALUEP1 - Entering first pair
	NORMAL - Working on second glyph or later.

	Once entered, a value remains until overwritten.
       */
      enum {NO_GID1, NO_GID2A, NO_GID2B, VALUEP1, NORMAL} state;
      struct glyph_group *gg1, *gg2;
      struct glyph_group *org = NULL; /* First glyph ID entered */
      int multiple_g1 = 0;     /* More then one glyph group for glyph 1 */
      unsigned short int vf1, vf2;
      struct ValueRecord vr1, vr2;
      enum {dx1 = 1, dy1, ax1, ay1, dx2, dy2, ax2, ay2, g1, g2, end} tokid;
      int gonbad = 0;
/* Index of table is glyph1 or class1, in format %5.5d.
   Value is a map whose index is glyph2 or class2, in format %5.5d.
   Its value is a pointer to struct hybrid_pair_value_record */
      struct value_pair_set_map value_table;
/* Variables for format 2 */
      int max_class_1 = -1, max_class_2 = -1;
      struct glyph_group *limiting_coverage = NULL;
      struct glyph_group *first_gg2 = NULL;
      struct hybrid_Class1Record *c1r0;

      create_value_pair_set_map(&value_table);
      state = NO_GID1;
      vf1 = vf2 = 0;
      vr1._next = vr2._next = 0;
      while (AN_IDENT == current.form) {
	if (0) {
	} else if (!strcmp("dx1", current.sval) && (tokid = dx1) ||
		   !strcmp("dy1", current.sval) && (tokid = dy1) ||
		   !strcmp("ax1", current.sval) && (tokid = ax1) ||
		   !strcmp("ay1", current.sval) && (tokid = ay1) ||
		   !strcmp("dx2", current.sval) && (tokid = dx2) ||
		   !strcmp("dy2", current.sval) && (tokid = dy2) ||
		   !strcmp("ax2", current.sval) && (tokid = ax2) ||
		   !strcmp("ay2", current.sval) && (tokid = ay2)) {
	  int flag1 = 0;
	  int flag2 = 0;

	  advance_token();
	  if (AN_INT != current.form) {
	    show_source();
	    fprintf(stdout, "Expected an integer.\n");
	    gonbad = 1;
	    continue;
	  }
	  if (current.ival < -0x8000 || 0x7FFF < current.ival) {
	    show_source();
	    fprintf(stdout, "Not a valid signed 16-bit number.\n");
	    gonbad = 1;
	    advance_token();
	    continue;
	  }
	  switch (state) {
	  case NO_GID1:
	  case NO_GID2A:
	  case NO_GID2B:
	    show_source();
	    fprintf(stdout, "Expected a glyph name, not a number.\n");
	    gonbad = 1;
	  case VALUEP1:
	  case NORMAL:
	    switch (tokid) {
	    case dx1:
	      vr1.xPlacement = current.ival;
	      flag1 = XPlacement;
	      break;
	    case dy1:
	      vr1.yPlacement = current.ival;
	      flag1 = YPlacement;
	      break;
	    case ax1:
	      vr1.xAdvance = current.ival;
	      flag1 = XAdvance;
	      break;
	    case ay1:
	      vr1.yAdvance = current.ival;
	      flag1 = YAdvance;
	      break;
	    case dx2:
	      vr2.xPlacement = current.ival;
	      flag2 = XPlacement;
	      break;
	    case dy2:
	      vr2.yPlacement = current.ival;
	      flag2 = YPlacement;
	      break;
	    case ax2:
	      vr2.xAdvance = current.ival;
	      flag2 = XAdvance;
	      break;
	    case ay2:
	      vr2.yAdvance = current.ival;
	      flag2 = YAdvance;
	      break;
	    default:
	      unwritten; /* Should be inaccessible */
	    }
	    if (NORMAL == state) {
	      if ((flag1 & ~vf1) || (flag2 & ~vf2)) {
		show_source();
		fprintf(stderr, "Field was not present for first value - "
			"will be ignored.\n");
	      }
	    } else {
	      vf1 |= flag1;
	      vf2 |= flag2;
	    }
	    break;
	  default:
	    unwritten; /* Internal error trap. */
	  }
	  advance_token();
	} else if (!strcmp("glyph1", current.sval) && (tokid = g1) ||
		   !strcmp("glyph2", current.sval) && (tokid = g2) ||
		   !strcmp("end",    current.sval) && (tokid = end)) {
	  switch (state) {
	  case NO_GID1:
	  case NO_GID2A:
	  case NO_GID2B:
	    /* Nothing to do - unless I choose to moan */
	    break;
	  case VALUEP1:
	  case NORMAL:
	    if (!gg1 || !gg2) {
	      assert(gonbad); /* Ignore pair - bad glyph IDs */
	    } else {
	      struct int_range *irge1, *g1set, local1;
	      struct int_range *irge2, *g2set, local2;
	      int i1, i2;
	      char s1[6], s2[6];
	      struct value_pair_map *fg1;
	      struct hybrid_pair_value_record *nvr;

	      if (CLASS == org->grouping) {
		local1.end = local1.start = gg1->x.class_data->number;
		local1.next = 0;
		g1set = &local1;
		local2.end = local2.start = gg2->x.class_data->number;
		local2.next = 0;
		g2set = &local2;
	      } else {
		g1set = get_glyph_set(gg1, &local1);
		g2set = get_glyph_set(gg2, &local2);
	      }
	      for (irge1 = g1set; irge1; irge1 = irge1->next) {
		for (i1 = irge1->start; i1 <= irge1->end; i1++) {
		  sprintf(s1, "%5.5d", i1);
		  if (!find_value_pair_set(value_table, s1, &fg1)) {
		    if (fg1 = malloc(sizeof * fg1)) {
		      create_value_pair_map(fg1);
		      add_value_pair_set(&value_table, s1, fg1);
		    } else {
		      fprintf(stderr, "Memory exhausted.\n");
		      exit(1);
		    }
		  }
		  for (irge2 = g2set; irge2; irge2 = irge2->next) {
		    for (i2 = irge2->start; i2 <= irge2->end; i2++) {
		      sprintf(s2, "%5.5d", i2);
		      if (find_value_pair(*fg1, s2, &nvr)) {
			show_source();
			if (CLASS == org->grouping) {
			  const char *s1, *s2;
			  s1 = class_name_byno
			    (org->x.class_data->classification, i1);
			  s2 = class_name_byno
			    (first_gg2->x.class_data->classification, i2);
			  fprintf(stderr, "Multiple values for glyph classes "
				  "%s and %s.\n", s1, s2);
			} else {
			  fprintf(stderr, "Multiple values for glyphs "
				  "%s and %s.\n", get_glyph_name(i1),
				  get_glyph_name(i2));
			}
			gonbad = 1;
		      } else if (nvr = malloc(sizeof * nvr)) {
			nvr->secondGlyph = i2;
			nvr->value1 = vr1;
			nvr->value2 = vr2;
			add_value_pair(fg1, s2, nvr);
		      } else {
			fprintf(stderr, "Memory exhausted.\n");
			exit(1);
		      }
		    }
		  }
		}
	      }
	    }
	    break;
	  default:
	    unwritten; /* internal error */
	  }
	  if (end == tokid) break;
	  switch (tokid) {
	  case g1:
	    advance_token();
	    if (!(check_ident())) {
	      gonbad = 1;
	      gg1 = 0;
	    } else if (lookup_in_map(glyph_names, current.sval, (void**)&gg1)) {
	      if (NO_GID1 == state) {
		if (limiting_coverage) {
		  if (CLASS == gg1->grouping) {
		    org = gg1;
		    state = NO_GID2B; /* 1st glyph 2 cannot affect format! */
		  } else {
		    show_source();
		    fprintf(stderr, "Limiting coverage implies organisation by "
			    "glyph class.\n");
		    gg1 = 0;
		    gonbad = 1;
		  }
		} else {
		  org = gg1;
		  state = NO_GID2A;
		}
	      } else {
		multiple_g1 = 1;
		state = NO_GID2B;
		if (CLASS == org->grouping) {
		  if (CLASS != gg1->grouping) {
		    show_source();
		    fprintf(stderr, "Lookup is organised by class, but '%s' "
			    "does not denote a class of glyphs.\n",
			    current.sval);
		    gonbad = 1;
		    gg1 = 0;
		  } else if (gg1->x.class_data->classification !=
			     org->x.class_data->classification) {
		    show_source();
		    fprintf(stderr, "'%s' is in the wrong classification.\n",
			    current.sval);
		    gg1 = 0;
		    gonbad = 1;
		  }
		}
	      }
	      if (CLASS == gg1->grouping) {
		int clno = gg1->x.class_data->number;
		if (clno > max_class_1) max_class_1 = clno;
	      }
	    } else {
	      show_source();
	      fprintf(stderr, "'%s' does not denote any glyph group.\n",
		      current.sval);
	      gonbad = 1;
	      gg1 = 0;
	    }
	    advance_token();
	    break;
	  case g2:
	    advance_token();
	    if (!(check_ident())) {
	      gonbad = 1;
	      gg2 = 0;
	    } else if (lookup_in_map(glyph_names, current.sval, (void**)&gg2)) {
	      if (NO_GID2A == state) {
		if (CLASS == org->grouping) {
		  if (CLASS == gg2->grouping) {
		    first_gg2 = gg2;
		  } else {
		    org = gg2;
		  }
		}
		state = NO_GID2B;
	      }
	      if (NO_GID1 != state && CLASS == org->grouping) {
		if (CLASS != gg2->grouping) {
		  show_source();
		  fprintf(stderr, "This subtable is organised by glyph "
			  "class, but '%s' does not designate "
			  "a glyph class.\n", current.sval);
		  gonbad = 1;
		  gg2 = 0;
		} else if (gg2->x.class_data->classification !=
		     first_gg2->x.class_data->classification) {
		  show_source();
		  fprintf(stderr, "'%s' is in the wrong glyph "
			  "classification.\n", current.sval);
		  gg2 = 0;
		  gonbad = 1;
		} else {
		  int clno = gg2->x.class_data->number;
		  if (clno > max_class_2) max_class_2 = clno;
		}
	      }
	      switch (state) {
	      case NO_GID1:
		show_source();
		fprintf(stderr, "Error: glyph1 must preced glyph2.\n");
		break;
	      case NO_GID2B:
		state = VALUEP1;
		break;
	      case VALUEP1:
		state = NORMAL;
		break;
	      case NORMAL:
		/* No change */
		break;
	      default:
		unwritten; /* internal error */
	      }
	    } else {
	      show_source();
	      fprintf(stderr, "'%s' does not denote any glyph group.\n",
		      current.sval);
	      gonbad = 1;
	      gg2 = 0;
	    }
	    advance_token();
	    break;
	  default:
	    unwritten; /* internal error */
	  }
	} else {/* Keyword not recognised */
	  show_source();
	  fprintf(stderr, "Keyword %s not recognised.\n", current.sval);
	  gonbad = 1;
	  skip_to_end("lookup");
	  return;
	}
      }
      if (AN_IDENT != current.form) {
	check_ident();
	gonbad = 1;
      }
      if (!gonbad && CLASS == org->grouping) {
	struct hybrid_Class1Record *c1r, **c1r_loc, *c1r_junk;
	int g1, g2;

	if (!(c1r0  = malloc((max_class_1 + 1) * sizeof * c1r0))) {
	  fprintf(stderr, "Memory exhausted.\n");
	  exit(1);
	}
	c1r_loc = &c1r_junk;
	for (g1 = 0, c1r = c1r0; g1 <= max_class_1; g1++, c1r++) {
	  char s1[6];
	  struct value_pair_map *fg1;
	  struct hybrid_pair_value_record **pvr_loc;

	  sprintf(s1, "%5.5d", g1);
	  if (!find_value_pair_set(value_table, s1, &fg1)) {
	    free(c1r0);
	    show_source();
	    fprintf(stderr, "Error: No pair value for glyph1 class %s.\n",
		    class_name_byno(org->x.class_data->classification, g1));
	    gonbad = 1;
	    break;
	  }
	  *c1r_loc = c1r;
	  c1r->_next = 0;
	  c1r_loc = &c1r->_next;

	  pvr_loc = &c1r->class2Record;
	  for (g2 = 0; g2 <= max_class_2; g2++) {
	    char s2[6];
	    struct hybrid_pair_value_record *pvr;

	    sprintf(s2, "%5.5d", g2);
	    if (!find_value_pair(*fg1, s2, &pvr)) {
	      free(c1r0);
	      show_source();
	      fprintf(stderr, "Error: No pair value for glyph1 class %s "
		      "and glyph2 class %s.\n",
		      class_name_byno(org->x.class_data->classification, g1),
		      class_name_byno(first_gg2->x.class_data->classification,
				      g2));
	      gonbad = 1;
	      break;
	    }
	    *pvr_loc = pvr;
	    pvr->_next = 0;
	    pvr_loc = &pvr->_next;
	  }
	  if (gonbad) break;
	}
      }
      if (!gonbad) { /* Write the record */
	struct hybrid_pairpos_table subtable;
	struct coverage_data *used_coverage;
	int junk;
	struct coverage_data local_coverage;

	memset(&subtable, 0, sizeof subtable);
	subtable.coverage = 0;
	subtable.valueFormat1 = vf1;
	subtable.valueFormat2 = vf2;
	local_coverage.members = create_int_set();
	local_coverage.lox = no_locations;
	if (CLASS == org->grouping) { /* Format 2 */
	  struct gclass_map_entry **cll;
	  int icl, ncl;
	  struct class_data *cld;
	  int_set uncovered;

	  subtable.posFormat = 2;
/* Calculate coverage */
	  uncovered = create_int_set();
	  sort_gclass_map(org->x.class_data->classification->classes,
			  &cll, &ncl);
	  for (icl = 0; icl < ncl; icl++) {
	    cld = cll[icl]->datum;
	    if (cld->number <= max_class_1) {
	      add_set(cld->members, &local_coverage.members);
	    } else {
	      add_set(cld->members, &uncovered);
	    }
	  }
	  free(cll);
#if 1
	  if (limiting_coverage) {
	    int_set dataless, temp1;

/* Calculate set of glyphs for which data is neither provided nor needed. */
	    temp1 = create_int_set();
	    add_set(uncovered, &temp1);
	    sub_set(limiting_coverage->x.coverage_data->members, &temp1);
/* Calculate problem set - not provided, but needed. */
	    dataless = create_int_set();
	    add_set(uncovered, &dataless);
	    sub_set(temp1, &dataless);
	    if (dataless) {
	      show_source();
	      fprintf(stderr, "No data for glyph1 = %s - using calculated "
		      "coverage.\n", get_glyph_name(dataless->start));
	      used_coverage = &local_coverage;
	    } else {
	      used_coverage = limiting_coverage->x.coverage_data;
	    }
	    delete_int_set(&dataless);
	  } else {
	    used_coverage = &local_coverage;
	  }
#endif
	  subtable.classDef1 = 0;
	  add_location(ld->type, *posn, 8 + *posn, 2,
		       &org->x.class_data->classification->lox);
	  subtable.classDef2 = 0;
	  add_location(ld->type, *posn, 10 + * posn, 2,
		       &first_gg2->x.class_data->classification->lox);
	  subtable.class1Count = max_class_1 + 1;
	  subtable.class2Count = max_class_2 + 1;
	  subtable.class1Record = c1r0;
	  put_bine_hybrid_pairpos_table(fp, &subtable, &junk, posn);
	  store_classification("Pair Lookup Classification 1", ld->type,
			       org->x.class_data->classification);
	  store_classification("Pair Lookup Classification 2", ld->type,
			       first_gg2->x.class_data->classification);
	  free(c1r0);
	  subtable.class1Record = 0;
	  free_hybrid_pairpos_table(&subtable);
	} else { /* Format 1 */
	  int ig1, ng1, gno;
	  struct value_pair_set_map_entry **g1l;
	  int stloc;

	  subtable.posFormat = 1;
	  sort_value_pair_set_map(value_table, &g1l, &ng1);
	  subtable.pairSetCount = ng1;
	  if (!(subtable.pairSetOffset =
		malloc(ng1 * sizeof * subtable.pairSetOffset))) {
	    fprintf(stderr, "Memory exhausted.\n");
	    exit(1);
	  }
	  stloc = *posn;
	  put_bine_hybrid_pairpos_table(fp, &subtable, &junk, posn);
	  for (ig1 = 0; ig1 < ng1; ig1++) {
	    int offset;
	    struct hybrid_pairset_table pst;
	    struct value_pair_map_entry **g2l;
	    int ig2, ng2, gno;
	    struct hybrid_pair_value_record **pvr_loc;
	    struct value_pair_map *fg1;

	    memset(&pst, 0, sizeof pst);
	    offset = *posn - stloc;
	    if (offset > 0xFFFF) {
	      show_source();
	      fprintf(stderr, "Pair pos look-up table is too large!");
	      offset = 0;
	    }
	    subtable.pairSetOffset[ig1] = offset;
	    fg1 = g1l[ig1]->datum;
	    sort_value_pair_map(*fg1, &g2l, &ng2);
	    pst.pairValueCount = ng2;
	    pst.pairValueRecord = 0;
	    pvr_loc = &pst.pairValueRecord;
	    for (ig2 = 0; ig2 < ng2; ig2++) {
	      struct hybrid_pair_value_record *pvr;
	      pvr = g2l[ig2]->datum;
	      *pvr_loc = pvr;
	      pvr->_next = 0;
	      pvr_loc = &pvr->_next;
	    }
	    put_bine_hybrid_pairset_table(fp, &pst, &junk, posn);
	    free_hybrid_pairset_table(&pst);
	    free(g2l);
	  }
/* Rewrite the table with the offsets filled in */
	  gsbp_seek(fp, stloc);
	  put_bine_hybrid_pairpos_table(fp, &subtable, &junk, &stloc);
	  gsbp_seek(fp, *posn);
	  free_hybrid_pairpos_table(&subtable);
/* Calculate the coverage. */
	  for (ig1 = ng1 - 1; ig1 >= 0; ig1--) { /* In reverse may be faster */
	    sscanf(g1l[ig1]->key, "%d", &gno);
	    add_range(gno, gno, &local_coverage.members);
	  }
	  if (limiting_coverage) { /* Glimmer in my eye */
	    if (int_set_same(limiting_coverage->x.coverage_data->members,
			     local_coverage.members)) {
	      used_coverage = limiting_coverage->x.coverage_data;
	    } else {
	      show_source();
	      fprintf(stderr, "Explicit coverage does not match the coverage "
		      "of the pairs - using calculated coverage.\n");
	      used_coverage = &local_coverage;
	    }
	  } else {
	    used_coverage = &local_coverage;
	  }
	  free(g1l);
	} /* End of specific Format 1 processing */
	add_location(ld->type, st_posn, 2 + st_posn, 2, &used_coverage->lox);
	store_coverage("Pair pos coverage", ld->type, used_coverage);
	delete_int_set(&local_coverage.members);
	free(local_coverage.lox.loc);
      }
      if (!gonbad) std->stored = WRITTEN;
    }
    break;
  case GPOS_MARK2BASE:
  case GPOS_MARK2LIGATURE:
  case GPOS_MARK2MARK:
    fp = gpos;
    posn = &gpos_pos;
    st_posn = *posn;
    {
/* TODO:
     Device tables
*/
      int gonbad = 0;
      struct classification_data mark_classes;
      struct mark_map mark_table;
      struct anchor_set_map base_table;
#if 0 /* Yanked from here to provide map function */
      struct mark_data {
	int class;
	struct anchor_table at;
      };
#endif
      /* Index mark_table by glyph number (%5.5d) and get class and anchor.
	 Index base_table by base glyph number (%5.5d) and get map by mark
	 *class* (%5.5d) or ligature number (%5.5d).  The latter yields maps
	 by mark *class* (%5.5d).  The maps indexed by class yield anchors.

      */ 
      struct anchor_table cat;
      int current_class;
      int current_component;
      const int use_lig = GPOS_MARK2LIGATURE == ld->type;
      enum {INIT, MARK, BASE, COMP, ATTACH} state;
      enum {notoken, x, y, pt, mark, empty, base, comp, attach, end} tokid;
      struct glyph_group *gg;

      create_gclass_map(&mark_classes.classes);
      mark_classes.lox = no_locations;
      mark_classes.n_classes = 0;
      create_mark_map(&mark_table);
      create_anchor_set_map(&base_table);

      state = INIT;
      memset(&cat, 0, sizeof cat);
      cat.anchorFormat = 1;
      while (AN_IDENT == current.form) {
	if (!strcmp("mark", current.sval) && (tokid = mark) ||
	    !strcmp("base", current.sval) && (tokid = base) ||
	    !strcmp("empty", current.sval) && (tokid = empty) ||
	    !strcmp("component", current.sval) && (tokid = comp) && use_lig ||
	    !strcmp("attach", current.sval) && (tokid = attach) ||
	    !strcmp("end", current.sval) && (tokid = end) ) {
	  struct int_range *irge, single;
	  int ig;
          char sb[6], sl[6], sm[6];

	  switch (state) {
	  case INIT:
	    break;
	  case MARK:
	    for (irge = get_glyph_set(gg, &single); irge; irge = irge->next) {
	      for (ig = irge->start; ig <= irge->end; ig++) {
		struct mark_data *md;
		sprintf(sm, "%5.5d", ig);
		if (find_mark(mark_table, sm, &md)) {
		  show_source();
		  fprintf(stderr, "Mark class and anchor already defined for "
			  "mark glyph %s.\n", get_glyph_name(ig));
		  gonbad = 1;
		} else if (md = malloc(sizeof * md)) {
		  md->class = current_class;
		  md->at = cat;
		  add_mark(&mark_table, sm, md);
		} else {
		  fprintf(stderr, "Memory exhausted.\n");
		  exit(1);
		}
	      }
	    }
	    memset(&cat, 0, sizeof cat);
	    cat.anchorFormat = 1;
	    break;
	  case BASE:
	  case COMP:
	    break;
	  case ATTACH:
	    if (use_lig) sprintf(sl, "%5.5d", current_component);
	    sprintf(sm, "%5.5d", current_class);
	    for (irge = get_glyph_set(gg, &single); irge; irge = irge->next) {
	      for (ig = irge->start; ig <= irge->end; ig++) {
		struct anchor_set_map *flig;
		struct anchor_map *fcomp;
		struct anchor_table *nat;

		sprintf(sb, "%5.5d", ig);
		fcomp = provide_anchor_set(&base_table, sb, 0);
		if (use_lig) {
		  flig = (struct anchor_set_map*) fcomp;
		  fcomp = provide_anchor_set(flig, sl, 0);
		}
		if (find_anchor(*fcomp, sm, &nat)) {
		  show_source();
		  if (use_lig) {
		    fprintf(stderr, "Base anchor already defined for "
			    "base glyph %s ligature %d with mark class %s\n",
			    get_glyph_name(ig), current_component,
			    class_name_byno(&mark_classes, current_class));
		  } else {
		    fprintf(stderr, "Base anchor already defined for "
			    "base glyph %s with mark class %s\n",
			    get_glyph_name(ig),
			    class_name_byno(&mark_classes, current_class));
		  }
		  gonbad = 1;
		} else if (nat = malloc(sizeof * nat)) {
		  *nat = cat;
		  add_anchor(fcomp, sm, nat);
		} else {
		  fprintf(stderr, "Memory exhausted.\n");
		  exit(1);
		}
	      }
	    }
	    memset(&cat, 0, sizeof cat);
	    cat.anchorFormat = 1;
	    break;
	  default:
	    assert(0);
	  }
	  if (end == tokid) break;
	  switch (tokid) {
	  case mark:
	  case empty:
	    {
	      struct class_data *cld;
	      advance_token();
	      if (!check_ident()) {
		gonbad = 1;
		break;
	      }
	      if (find_gclass(mark_classes.classes, current.sval, &cld)) {
		current_class = cld->number;
	      } else if (cld = malloc(sizeof * cld)) {
		cld->members = create_int_set();
		cld->classification = &mark_classes;
		current_class = cld->number = mark_classes.n_classes++;
		add_gclass(&mark_classes.classes, current.sval, cld);
	      } else {
		fprintf(stderr, "Memory exhausted.\n");
		exit(1);
	      }
	      advance_token();
	      if (mark == tokid) {
		struct int_range single;

		state = MARK;
		if (!check_ident()) {
		  gonbad = 1;
		  break;
		}
		if (!lookup_in_map(glyph_names, current.sval, (void**)&gg)) {
		  gg = 0;
		  gonbad = 1;
		  show_source();
		  fprintf(stderr, "'%s' does not identify any glyphs.\n",
			  current.sval);
		}
		add_set(get_glyph_set(gg, &single), &cld->members);
		advance_token();
	      } else {
		state = INIT;
	      }
	    }
	    break;
	  case base:
	    advance_token();
	    if (!check_ident()) {
	      gonbad = 1;
	      break;
	    }
	    state = BASE;
	    if (!lookup_in_map(glyph_names, current.sval, (void**)&gg)) {
	      gg = 0;
	      gonbad = 1;
	      show_source();
	      fprintf(stderr, "'%s' does not identify any glyphs.\n",
		      current.sval);
	    }
	    advance_token();
	    break;
	  case comp:
	    assert(use_lig);
	    advance_token();
	    current_component = 0;
	    if (AN_INT == current.form) {
	      if (0 <= current.ival && current.ival <= 0xFFFF) {
		current_component = current.ival;
                state = COMP;
	      } else {
		gonbad = 1;
		show_source();
		fprintf(stderr, "Invalid component number.\n");
	      }
	    } else {
	      gonbad = 1;
	      show_source();
	      fprintf(stderr, "Expected an integer.\n");
	    }
	    advance_token();
	    break;
	  case attach:
	    {
	      struct class_data *cld;
	      advance_token();
	      if (!check_ident()) {
		gonbad = 1;
		show_source();
	      }
	      if (find_gclass(mark_classes.classes, current.sval, &cld)) {
		current_class = cld->number;
	      } else {
		gonbad = 1;
		show_source();
		fprintf(stderr, "'%s' does not identify a mark class for "
			"this subtable.\n", current.sval);
		current_class = 0;
	      }
	      state = ATTACH;
	      advance_token();
	    }
	    break;
	  default:
	    unwritten; /* internal error */
	  }
	} else if (!strcmp("x", current.sval)  && (tokid = x) ||
		   !strcmp("y", current.sval)  && (tokid = y) ||
		   !strcmp("pt", current.sval) && (tokid = pt) ) {
	  advance_token();
	  if (AN_INT == current.form) {
	    switch (state) {
	    case MARK:
	    case ATTACH:
	      switch (tokid) {
	      case x:
	      case y:
		if (-0x8000 <= current.ival && current.ival <= 0x7FFF) {
		  if (x == tokid) {
		    cat.xCoordinate = current.ival;
		  } else {
		    cat.yCoordinate = current.ival;
		  }
		} else {
		  gonbad = 1;
		  show_source();
		  fprintf(stdout, "%d is not a signed 16-bit number.\n",
			  current.ival);
		}
		break;
	      case pt:
		if (0 <= current.ival && current.ival <= 0xFFFF) {
		  switch (cat.anchorFormat) {
		  case 1:
		  case 2:
		    cat.anchorFormat = 2;
		    cat.anchorPoint = current.ival;
		    break;
		  case 3:
		    gonbad = 1;
		    show_source();
		    unwritten; /* Currently inaccessible */
		    break;
		  default:
		    unwritten; /* internal error */
		  }
		} else {
		  gonbad = 1;
		  show_source();
		  fprintf(stdout, "%d is not an unsigned 16-bit number.\n",
			  current.ival);
		}
		break;
	      default:
		unwritten; /* internal error */
	      }
	      break;
	    default:
	      show_source();
	      gonbad = 1;
	      fprintf(stderr, "Attachment point or mark not defined.\n");
	    }
	    advance_token();
	  } else {
	    gonbad = 1;
	    show_source();
	    fprintf(stderr, "Expected an integer.\n");
	    unwritten;
	  }
	} else {
	  gonbad = 1;
	  fprintf(stdout, "Keyword '%s' not recognised.\n", current.sval);
	  advance_token();
	}
      }
      if (AN_IDENT != current.form) {
	check_ident();
	gonbad = 1;
      }
      if (!gonbad) {
	struct mark_basepos_table mbt;
	struct coverage_data mark_coverage, base_coverage;
	const int stloc = *posn;
	int matloc, latloc;
	int loc;
	int offset;
	struct gclass_map_entry **cll;
	struct mark_map_entry **mdl;
	struct anchor_set_map_entry **ligl;
	int icl, ncl, junk, n_marks, imd, nmd, ilig, nlig;
	struct mark_array_table mat;
	struct mark_record *mr, **mr_loc;
	struct ligature_array_table lat;

	mark_coverage.members = create_int_set();
	mark_coverage.lox = no_locations;
	base_coverage.members = create_int_set();
	base_coverage.lox = no_locations;

/* Store first pass at mark to base table */	
	mbt.posFormat = 1;
	mbt.markCoverage = 0;
	add_location(ld->type, stloc, stloc + 2, 2, &mark_coverage.lox);
	mbt.baseCoverage = 0;
	add_location(ld->type, stloc, stloc + 4, 2, &base_coverage.lox);
	mbt.classCount = mark_classes.n_classes;
	mbt.markArray = 0;
	mbt.baseArray = 0;
	put_bine_mark_basepos_table(fp, &mbt, &junk, posn);
/* Store mark coverage */
	n_marks = 0;
	sort_gclass_map(mark_classes.classes, &cll, &ncl);
	assert(ncl == mark_classes.n_classes);
	for (icl = 0; icl < ncl; icl++) {
	  struct class_data *cld;
	  struct int_range *irge;

	  cld = cll[icl]->datum;
	  add_set(cld->members, &mark_coverage.members);
	  for (irge = cld->members; irge; irge = irge->next) {
	    n_marks += irge->end - irge->start + 1;
	  }
	}
/* Store mark array table - pass 1 */
	sort_mark_map(mark_table, &mdl, &nmd);
	assert(n_marks == nmd);
	mat.markCount = nmd;
	if (mat.markRecord = malloc(nmd * sizeof * mat.markRecord)) {
	  mr_loc = &mat.markRecord;
	  for (imd = 0, mr = mat.markRecord; imd < nmd; imd++, mr++) {
	    *mr_loc = mr;
	    mr_loc = &mr->_next;
	    *mr_loc = 0;
	  }
	} else {
	  fprintf(stderr, "Memory exhuasted.\n");
	  exit(1);
	}
	matloc = *posn;
	offset = matloc - stloc;
	if (offset > 0xFFFF) {
	  show_source();
	  fprintf(stderr, "Error - need to store mark array %d bytes beyond "
		  "mark to ... subtable.\n", offset);
	  gonbad = 1;
	  offset = 0;
	}
	mbt.markArray = offset; 
	put_bine_mark_array_table(fp, &mat, &junk, posn);
/* Store mark anchors and update the mark array table */
	for (imd = 0; imd < nmd; imd++) {
	  struct mark_data *md;
	  md = mdl[imd]->datum;
	  mat.markRecord[imd].class = md->class;
	  mat.markRecord[imd].markAnchor = *posn - matloc;
	  put_bine_anchor_table(fp, &md->at, &junk, posn);
	}
	gsbp_seek(fp, loc = matloc);
	put_bine_mark_array_table(fp, &mat, &junk, &loc);
	gsbp_seek(fp, *posn);
	if (use_lig) {
/* Ligature aray table - reserve space */
	  latloc = *posn;
	  offset = latloc - stloc;
	  if (offset > 0xFFFF) {
	    show_source();
	    fprintf(stderr, "Error - need to store ligature array table "
		    "%d bytes beyond mark to ... subtable.\n", offset);
	    gonbad = 1;
	    offset = 0;
	  }
	  mbt.baseArray = offset;
	  sort_anchor_set_map(base_table, &ligl, &nlig);
	  memset(&lat, 0, sizeof lat);
	  lat.ligatureCount = nlig;
	  if (lat.ligatureAttach = malloc(nlig * sizeof * lat.ligatureAttach)) {
	    for (ilig = 0; ilig < nlig; ilig++) lat.ligatureAttach[ilig] = 0;
	  } else {
	    fprintf(stderr, "Memory exhausted.\n");
	    exit(1);
	  }
	  put_bine_ligature_array_table(fp, &lat, &junk, posn);
	} else {
	  nlig = 1;
	}
/* Now generate first draft of ligature attach tables */
	for (ilig = 0; ilig < nlig; ilig++) {
	  struct anchor_set_map_entry **basel;
	  int ib, jb, ic, nb, ncomp;
	  struct base_array_table bat;
	  int batloc;

	  memset(&bat, 0, sizeof bat);
	  if (use_lig) {
	    struct anchor_set_map *flig;
	    int gno;

	    assert(1 == sscanf(ligl[ilig]->key, "%d", &gno));
	    add_range(gno, gno, &base_coverage.members);
	    flig = (struct anchor_set_map*) ligl[ilig]->datum;
	    sort_anchor_set_map(*flig, &basel, &nb);
	    assert(1 == sscanf(basel[nb-1]->key, "%d", &jb));
	    ncomp = jb + 1;
	  } else {
	    sort_anchor_set_map(base_table, &basel, &nb);
	    ncomp = nb;
	  }
	  bat.baseCount = ncomp;
	  if (bat.baseRecord = malloc(ncomp * sizeof * bat.baseRecord)) {
	    /* Link them up */
	    struct base_record *br, **br_loc;
	    br_loc = &bat.baseRecord;
	    for (ic = 0, br = bat.baseRecord; ic < ncomp; ic++, br++) {
	      *br_loc = br;
	      br_loc = &br->_next;
	      *br_loc = 0;
	      if (!(br->baseAnchor = malloc(ncl * sizeof * br->baseAnchor))) {
		fprintf(stderr, "Memory exhausted.\n");
		exit(1);
	      }
	    }
	  } else {
	    fprintf(stderr, "Memory exhausted.\n");
	    exit(1);
	  }
	  batloc = *posn;
	  if (use_lig) {
	    offset = batloc - latloc;
	    if (offset > 0xFFFF) {
	      show_source();
	      fprintf(stderr, "Error - need to store ligature array table "
		      "%d bytes beyond mark to ligature subtable.\n", offset);
	      gonbad = 1;
	      offset = 0;
	    }
	    lat.ligatureAttach[ilig] = offset;
	  } else if (!use_lig) {
	    offset = batloc - stloc;
	    if (offset > 0xFFFF) {
	      show_source();
	      fprintf(stderr, "Error - need to store base array table "
		      "%d bytes beyond mark to ... subtable.\n", offset);
	      gonbad = 1;
	      offset = 0;
	    }
	    mbt.baseArray = offset;
	  }
	  put_bine_base_array_table(fp, &bat, &junk, posn);
/* Now write base anchor tables, and record their locations */
	  for (ib = 0, ic = 0; ib < nb; ib++) {
	    struct anchor_map *fcomp = basel[ib]->datum;
	    char sm[6];
	    int im;
	    struct anchor_table *anchor;
	    
/* Bad code: assert(1 == sscanf(basel[nb-1]->key, "%d", &jb)); */
	    assert(1 == sscanf(basel[ib]->key, "%d", &jb));
	    if (use_lig) {
	      for (; ic < jb; ic++) {
		if (use_lig) {
		  int gno;
		  assert(1 == sscanf(ligl[ilig]->key, "%d", &gno));
		  fprintf(stderr, "WARNING: No base anchors for base %s, "
			  "component %d.\n", get_glyph_name(gno), ic);
		} else { /* This branch looks irrelevant, and rightfully dead */
		  gonbad = 1;
		  fprintf(stderr, "ERROR: No base anchors for %s.\n",
			  get_glyph_name(ic));
		}
		for (im = 0; im < ncl; im++) {
		  bat.baseRecord[ic].baseAnchor[im] = 0;
		}
	      }
	    } else {
	      int gno;
	      assert(1 == sscanf(basel[ib]->key, "%d", &gno));
	      add_range(gno, gno, &base_coverage.members);
	    }
	    for (im = 0; im < ncl; im++) {
	      sprintf(sm, "%5.5d", im);
	      if (find_anchor(*fcomp, sm, &anchor)) {
		offset = *posn - batloc;
		if (offset > 0xFFFF) {
		  show_source();
		  fprintf(stderr, "Offset too large.\n");
		  gonbad = 1;
		  offset = 0;
		}
		bat.baseRecord[ic].baseAnchor[im] = offset;
		put_bine_anchor_table(fp, anchor, &junk, posn);
	      } else {
		int gno;

		bat.baseRecord[ic].baseAnchor[im] = 0;
		if (use_lig) {
		  assert(1 == sscanf(ligl[ilig]->key, "%d", &gno));
		  fprintf(stderr, "WARNING: No base anchor for base %s "
			  "component %d and mark class %s.\n",
			  get_glyph_name(gno), ic,
			  class_name_byno(&mark_classes,im));
		} else {
		  assert(1 == sscanf(basel[ic]->key, "%d", &gno));
		  fprintf(stderr, "WARNING: No base anchor for base %s "
			  "and mark class %s.\n", get_glyph_name(gno),
			  class_name_byno(&mark_classes,im));
		}
	      }
	    }
	    ic++;
	  }
	  gsbp_seek(fp, loc = batloc);
	  put_bine_base_array_table(fp, &bat, &junk, &loc);
	  gsbp_seek(fp, *posn);
	}
/* Now rewrite the completed basic subtable */
	gsbp_seek(fp, loc = stloc);
#if 0
	put_bine_mark_basepos_table(fp, &mbt, &junk, posn);
#else
	put_bine_mark_basepos_table(fp, &mbt, &junk, &loc);
#endif
	if (use_lig) { /* Write ligature attachment table. */
	  gsbp_seek(fp, loc = latloc);
	  put_bine_ligature_array_table(fp, &lat, &junk, &loc);
	}
	gsbp_seek(fp, *posn);
/* In this unusual case, one could actually store the coverage offsets
   in the table before writing it out, but the writing interface does not
   support that, and store_coverage() could legitimately prompt for the
   earlier storage of other items.  */
	store_coverage(" mark coverage ", ld->type, &mark_coverage);
	store_coverage(" base coverage ", ld->type, &base_coverage);

	delete_int_set(&mark_coverage.members);
	free(mark_coverage.lox.loc);
	delete_int_set(&base_coverage.members);
	free(base_coverage.lox.loc);
      }
      delete_mark_map(&mark_table, 1);
      delete_anchor_set_map(&base_table, 0);
      delete_gclass_map(&mark_classes.classes, 0);
      if (!gonbad) std->stored = WRITTEN;
    }
    break;
  case GPOS_CURSIVE:
    fp = gpos;
    posn = &gpos_pos;
    /* Set st_posn and update std->stored when implemented! */
    fprintf(stderr, "Lookups of this type are not yet supported.\n");
    skip_to_end("lookup");
    return;
  case GSUB_EXTENSION:
  case GPOS_EXTENSION:
    fprintf(stderr, "Extension lookup types should not be used by the "
	    "compiler to classify lookups.\n");
    exit(1);
  default:
    fprintf(stderr, "Unified lookup type code %d not supported at "
	    "line %d of %s.\n", ld->type, __LINE__, __FILE__);
    exit(1);
  }
  {
    int loc;
    int offset, junk;

    offset = st_posn - std->ref.offset_ref;
    gsbp_seek(fp, loc = std->ref.offset_loc);
    if (4 == std->ref.offset_size) {
      struct UL_item fourbytes;

      fourbytes.val = offset;
      put_bine_UL_item(fp, &fourbytes, &junk, &loc);
    } else if (offset <= 0xFFFF) {
      struct UW_item twobytes;

      twobytes.val = offset;
      put_bine_UW_item(fp, &twobytes, &junk, &loc);
    } else {
      fprintf(stderr, "Definition of subtable %s is to be stored too far "
	      "from lookup.\n", get_subtable_name(std));
    }
  }
  gsbp_seek(fp, *posn);
  coda_end_syntax("lookup");
}

void check_all_subtables() {
  struct subtable_map_entry **stl;
  struct subtable_data *std;
  struct lookup_data *ld;
  struct UW_item twobytes;
  struct UL_item fourbytes;
  int i, n, junk;
  int loc;

  sort_subtable_map(subtables, &stl, &n);
  for (i = 0; i < n; i++) {
    std = stl[i]->datum;
    if (WRITTEN == std->stored) continue;
    fprintf(stderr, "Lookup subtable %s is missing.\n", stl[i]->key);
    ld = std->lookup;
    if (ld->is_extended) {
      fprintf(stderr, "Extension lookups are not yet supported.\n");
    }
    if (GSUB_1 <= ld->type && ld->type <= GSUB_9) {
      gsbp_seek(gsub, loc = std->ref.offset_loc);
      if (4 == std->ref.offset_size) {
	fourbytes.val = 0;
	put_bine_UL_item(gsub, &fourbytes, &junk, &loc);
      } else {
	twobytes.val = 0;
	put_bine_UW_item(gsub, &twobytes, &junk, &loc);
      }
      gsbp_seek(gsub, gsub_pos);
    } else if (GPOS_1 <= ld->type && ld->type <= GPOS_9) {
      gsbp_seek(gpos, loc = std->ref.offset_loc);
      if (4 == std->ref.offset_size) {
	fourbytes.val = 0;
	put_bine_UL_item(gsub, &fourbytes, &junk, &loc);
      } else {
	twobytes.val = 0;
	put_bine_UW_item(gpos, &twobytes, &junk, &loc);
      }
      gsbp_seek(gpos, gpos_pos);
    } else {
      fprintf(stderr, "Lookup subtable %s has unrecognised unified lookup "
	      "code %d.\n", stl[i]->key, ld->type);
    }
  }
  free(stl);
}

static struct coverage_data **glyph_sets;
static int ngs, mgs;
void init_glyph_sets() {
  glyph_sets = 0;
  ngs = mgs = 0;
}

static int add_glyph_set(struct coverage_data *cov) {
  int igs;
  for (igs = 0; igs < ngs; igs++) {
    if (cov == glyph_sets[igs]) {
      return igs;
    }
  }
  if (++ngs > mgs) {
    mgs++;
    if (!(glyph_sets = realloc(glyph_sets, mgs * sizeof * glyph_sets))) {
      fprintf(stderr, "Memory exhausted!\n");
      exit(1);
    }
  }
  glyph_sets[ngs-1] = cov;
  return ngs - 1;
}

int store_glyph_sets() {
  struct mark_glyph_sets_table index;
  int junk, posin, igs;
  char label[40];
  if (ngs) { /* Just set count if no glyph sets to record. */
    if (fseek(outfp, *out_pos, SEEK_SET)) {
      fprintf(stderr, "Error seeking within GDEF table.\n");
      exit(1);
    };
    index.format = 1;
    index.count = ngs;
    if (0 == (index.offsets = malloc(ngs * sizeof * index.offsets))) {
      fprintf(stderr, "Memory exhausted!\n");
      exit(1);
    }
    posin = *out_pos;
    put_bine_mark_glyph_sets_table(outfp, &index, &junk, out_pos);
    free_mark_glyph_sets_table(&index);
    for (igs = 0; igs < ngs; igs++) {
      sprintf(label, "mark glyph set %d", igs);
      add_location(GDEF_TGT, posin, posin + 4 + 4*igs, 4,
                   &glyph_sets[igs]->lox);
      store_coverage(label, GDEF_TGT, glyph_sets[igs]);
    }
  }
  return ngs;
}

void free_glyph_sets() {
  free(glyph_sets);
  ngs = mgs = 0;
}
