/*
Tokeniser for OpenType compiler
Copyright (C) 2005-2017      Richard Wordingham

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.
*/
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "tokeniser.h"
struct token previous, current;

/* Private declarations */
#include <stdio.h>
static FILE *src = 0;
static struct token next_token;
static int line_number = 0;

static void init_tokens() {
  memset(&previous,   0, sizeof(previous));
  memset(&current,    0, sizeof(current));
  memset(&next_token, 0, sizeof(next_token));
}

void open_source(const char *filename) {
  init_tokens();
  src = fopen(filename, "r");
  if (!src) {
    perror("Cannot open font definition file");
    exit(1);
  }
  line_number = 1;
  advance_token();
}

void backstep_token() {
  if (UNDEFINED != next_token.form) {
    show_source();
    fprintf(stderr, "Internal error!  Backstepped two tokens!\n");
    exit(1);
  }
  next_token = current;
  current = previous;
  previous.form = UNDEFINED;
}

void advance_token() {
  int onch, as_onch, onch2, namlen, capacity, in_comment;
  int isdig;
  char *name;
  #define NAMBUFSIZ 20
  char numbuf[NAMBUFSIZ + 1];

  if (A_STRING == previous.form) free(previous.sval);
  previous = current;
  in_comment = 0;
  if (UNDEFINED == next_token.form) {
    current.sval = 0;
    current.rval = 42.42;
    current.ival = 42;
    while (EOF != (onch = fgetc(src))) {
      as_onch = onch;
      if ('\r' == onch) as_onch = ' ';
      /* UTF-8 friendly, so long as no exotic whitespace. */
      if (0x80 & onch) as_onch = 'a';
      if ('.' == onch) as_onch = 'a';
      if ('_' == onch) as_onch = 'a';
      if (2 == in_comment) {
	if ('-' == onch) {
	  in_comment = 1;
	  continue;
	} else {
	  ungetc(onch, src);
	  current.form = PUNCTUATION;
	  current.ival = '-';
	  in_comment = 0;
	  break;
	}
      } else if ('\n' == as_onch) {
	line_number++;
	in_comment = 0;
      } else if (1 == in_comment) {
	/* Do nothing */
      } else if (isspace(as_onch)) {
	/* Continue */
      } else if (isalpha(as_onch)) {
	capacity = 32;
	name = malloc(capacity);
	if (!name) {
	  fprintf(stderr, "Memory exhausted processing identifier (1).\n");
	  show_source();
	  exit(1);
	}
	namlen = 1;
	name[0] = onch;
	while (EOF != (onch2 = fgetc(src))) {
	  as_onch = onch2;
	  if (0x80 & onch2) as_onch = 'b';
	  if ('.' == onch2) as_onch = 'c';
	  if ('_' == onch2) as_onch = 'd';
	  if (!isalnum(as_onch)) {
	    ungetc(onch2, src);
	    break;
	  }
	  if (namlen >= capacity - 1) {
	    capacity = 2 * capacity;
	    name = realloc(name, capacity);
	    if (!name) {
	      fprintf(stderr, "Memory exhausted processing identifier (2).\n");
	      show_source();
	      exit(1);
	    }
    	  }
	  name[namlen++] = onch2;
	}
	name[namlen] = 0;
	current.form = AN_IDENT;
	current.sval = name;
	break;
      } else if ('"' == onch) {
	int pending = 0;

	capacity = 32;
	name = malloc(capacity);
	if (!name) {
	  fprintf(stderr, "Memory exhausted processing identifier (1).\n");
	  show_source();
	  exit(1);
	}
	namlen = 0;
	while (EOF != (onch2 = fgetc(src))) {
	  if ('"' == onch2 && !pending) break;
/* More triggerable because of \ effects. */
	  if (namlen >= capacity - 2) {
	    capacity = 2 * capacity;
	    name = realloc(name, capacity);
	    if (!name) {
	      fprintf(stderr, "Memory exhausted processing identifier (2).\n");
	      show_source();
	      exit(1);
	    }
    	  }
	  if (pending) {
	    switch (onch2) {
	    case '\\':
	    case '"':
	      name[namlen++] = onch2;
	      break;
	    default:
	      name[namlen++] = '\\';
	      name[namlen++] = onch2;
	      break;
	    }
	    pending = 0;
	  } else if ('\\' == onch2) {
	    pending = 1;
	  } else {
	    name[namlen++] = onch2;
	  }
	}
	name[namlen] = 0;
	current.form = A_STRING;
	current.sval = name;
	break;
      } else if ((isdig = isdigit(as_onch))
		 || '+' == as_onch || '-' == as_onch) {
	enum {PLAIN, FRACTION, OCTAL, HEX} mode;
	int truncating = 0;
	char *unused;
	int is_ok, sign;

        namlen = 0;
	if (!isdig) {
	  numbuf[namlen++] = sign = as_onch;
	  onch = fgetc(src);
          if (onch & 0x80) {
	    isdig = 0;
	  } else {
	    isdig = isdigit(onch);
	  }
	}
	if (isdig) {
	  if ('0' == onch) {
	    mode = HEX;
	  } else {
	    mode = PLAIN;
	  }
	  numbuf[namlen++] = onch;
	  while (EOF != (onch2 = fgetc(src))) {
	    if (HEX == mode) {
	      is_ok = isxdigit(onch2);
	      is_ok |= 'X' == onch2;
	      is_ok |= 'x' == onch2;
	    } else {
	      is_ok = isdigit(onch2);
	    }
	    is_ok |= '.' == onch2;
	    if (is_ok) {
	      if (!truncating && namlen >= NAMBUFSIZ) {
		show_source();
		fprintf(stderr, "Number is too long.  Truncated.");
		truncating = 1;
	      }
	      if ('.' == onch2) {
		if (PLAIN == mode) {
		  mode = FRACTION;
		} else {
		  numbuf[namlen] = 0;
		  if (namlen == strspn(numbuf, "0")) {
		    mode = FRACTION;
		  } else {
		    break;
		  }
		}
	      }
	      if (!truncating) numbuf[namlen++] = onch2;
	    } else {
	      break;
	    }
	  }
	  ungetc(onch2, src);
	  numbuf[namlen] = 0;
	  if (FRACTION == mode) {
	    current.form = A_REAL;
	    current.rval = strtod(numbuf, &unused);
	  } else {
	    current.form = AN_INT;
	    current.ival = strtoll(numbuf, &unused, 0);
	  }
	  if (*unused) {
	    show_source();
	    fprintf(stderr, "Portion \"%s\" of \"%s\" was not interpreted.\n",
		    unused, numbuf);
	  }
	  break;
	} else {
	  if ('-' == sign && '-' == onch) {
	    in_comment = 1;
	  } else {
	    ungetc(onch, src);
	    current.form = PUNCTUATION;
	    current.ival = sign;
	    break;
	  }
	}
      } else if ('-' == onch) { /* Currently impossible */
	in_comment = 2;
      } else {
	current.form = PUNCTUATION;
	current.ival = onch;
	break;
      }
    }
    if (2 == in_comment) {
      current.form = PUNCTUATION;
      current.ival = '-';
    } else if (EOF == onch) {
      current.form = END_OF_FILE;
    }
    current.lino = line_number;
  } else {
    current = next_token;
    next_token.form = UNDEFINED;
    next_token.sval = 0;
  }
}

void show_source() {
  fprintf(stderr, "Processing line %d:\n", current.lino);
}

void got_what(FILE *fp, struct token curr) {
  fprintf(fp, "Got ");
  switch (curr.form) {
  case A_STRING:
    fprintf(fp, "string \"%s\"", curr.sval);
    break;
  case AN_IDENT:
    fprintf(fp, "identifier \"%s\"", curr.sval);
    break;
  case AN_INT:
    fprintf(fp, "integer %d (0x%x)", curr.ival, curr.ival);
    break;
  case A_REAL:
    fprintf(fp, "real %e", curr.rval);
    break;
  case PUNCTUATION:
    fprintf(fp, "punctuation '%c' (code 0x%x)",
	    curr.ival, curr.ival);
    break;
  case END_OF_FILE:
    fprintf(fp, "end of file");
    break;
  case UNDEFINED:
    fprintf(fp, "unclassified token");
    break;
  default:
    fprintf(fp, "unknown token type");
    break;
  }
  fprintf(fp, ".\n");
}
